Require Import Rdefinitions Raxioms RIneq Rbasic_fun Zwf.
Require Import Ranalysis1 Rsqrt_def.
Require Import IndefiniteDescription.
Require Import ssreflect ssrbool eqtype ssrnat choice seq ssrfun ssralg.
Require Import fintype bigop rat ssrnum poly.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section Rstruct.

Open Scope R_scope.


Lemma Req_EM_T : forall r1 r2:R, {r1 = r2} + {r1 <> r2}.
Proof.
  intros; elim (total_order_T r1 r2); intros;
    [ elim a; intro;
      [ right; red in |- *; intro; rewrite H in a0; elim (Rlt_irrefl r2 a0)
        | left; assumption ]
      | right; red in |- *; intro; rewrite H in b; elim (Rlt_irrefl r2 b) ].
Qed.


Definition eqr (r1 r2 : R) : bool := 
let H:=(Req_EM_T r1 r2) in
match H with 
|left _ => true
| right _ => false
end.

Lemma eqrP : Equality.axiom eqr.
Proof.
move=> r1 r2; apply: (iffP idP) => [|<-];
[rewrite /eqr; case (Req_EM_T r1 r2) =>//=|
rewrite /eqr; case (Req_EM_T r1 r1) =>//=].
Qed.


Canonical Structure real_eqMixin := EqMixin eqrP.
Canonical Structure real_eqType := Eval hnf in EqType R real_eqMixin.

Lemma R_irrelevance : forall (x y : R) (E E' : x = y), E = E'.
Proof. exact: eq_irrelevance. Qed.

Parameter find_real : pred R -> nat -> option R.

Axiom correct_real : forall P n x, find_real P n = Some x -> P x.
Axiom complete_real: forall (P : pred R), (exists x, P x) ->
  exists n, find_real P n.
Axiom ext_real : forall P Q, P =1 Q -> find_real P =1 find_real Q.


Definition real_choiceMixin := Choice.Mixin correct_real complete_real ext_real.
Canonical Structure real_choiceType := 
  Eval hnf in ChoiceType R real_choiceMixin.

(* Definition real_choiceMixin := ChoiceMixin  sreal_xchooseP  eq_sreal_xchoose. *)
(* Canonical Structure real_choiceType := Eval hnf in ChoiceType R real_choiceMixin. *)

(*we define a commutative ring structure for the real numbers, 
so that we may use the matrices in ssreflect*)
 
Lemma RplusA : associative (Rplus).
by move=> *; ring_simplify. Qed.

Lemma RplusC : commutative (Rplus).
by move=> *; ring_simplify.
Qed.
Lemma RplusR0 : left_id R0 Rplus.
by move=> *; ring_simplify.
Qed.
Lemma R0Rplus : right_id R0 Rplus.
by move=> *; ring_simplify.
Qed.
Lemma RplusN : left_inverse R0 Ropp Rplus.
by move=> *; ring_simplify.
Qed.


Definition real_zmodMixin := ZmodMixin RplusA RplusC RplusR0 RplusN.
Canonical Structure real_zmodType := Eval hnf in ZmodType R real_zmodMixin.

Lemma RmultA : associative (Rmult).
by move=> *; ring_simplify.
Qed.
Lemma left_idRmult: left_id R1 Rmult.
by move=> *; ring_simplify.
Qed.
Lemma right_idRmult : right_id R1 Rmult.
by move=> *; ring_simplify.
Qed.
Lemma left_distr_Rmult_Rplus : left_distributive Rmult Rplus.
by move=> *; ring_simplify.
Qed.
Lemma right_distr_Rmult_Rplus : right_distributive Rmult Rplus.
by move=> *; ring_simplify.
Qed.
Lemma R1_neq_0 : R1 != R0.
apply/eqP.
by exact: R1_neq_R0.
Qed.


Definition real_ringMixin := RingMixin RmultA left_idRmult 
  right_idRmult left_distr_Rmult_Rplus right_distr_Rmult_Rplus R1_neq_0.
Canonical Structure real_ringType := Eval hnf in RingType R real_ringMixin.

Lemma RmultC:  commutative Rmult.
by move=> *; ring_simplify.
Qed.

(*Definition real_comringMixin := ComRingMixin RmultC.

RmultA RmultC left_idRmult 
   left_distr_Rmult_Rplus  R1_neq_0 .*)

Canonical Structure real_comringType := Eval hnf in ComRingType R RmultC.


(*why do i need all the lemmas in the mixin for the com ring?; it should be infered 
from the can struct of ring for R *)

Definition Rinvx r:= if (r != 0) then (Rinv r) else r.

Definition unit_R : pred R := fun r => r != 0.

(*Definition invmx A := if unit_mx A then (\det A)^-1 *m: \adj A else A.*)

Lemma RmultRinvx : {in unit_R, left_inverse 1 Rinvx Rmult}.
move=> r; rewrite -topredE /= => n0r.
rewrite /Rinvx /=.
rewrite /unit_R in n0r.
case C: (r != 0).
by rewrite <-Rinv_l_sym; [| apply/eqP].
move/eqP : C => C.
move/eqP : n0r => n0r.
elim (n0r C).
Qed.


Lemma RinvxRmult : {in unit_R, right_inverse 1 Rinvx Rmult}.
move=> r.
 rewrite -topredE /= => n0r.
rewrite /Rinvx /=.
rewrite /unit_R in n0r.
case C: (r != 0).
by rewrite <-Rinv_r_sym; [| apply/eqP].
move/eqP : C => C.
move/eqP : n0r => n0r.
elim (n0r C).
Qed.

Import GRing.Theory.

Lemma intro_unit_R : forall a b : R , b * a = 1 /\ a * b = 1 -> unit_R a.
move=> a b [Hab Hba].
rewrite /unit_R.
apply/eqP.
move=> Hc.
rewrite Hc in Hab.
ring_simplify in Hab.
have Hc2:= R1_neq_R0.
by apply Hc2; rewrite Hab.
Qed.

Lemma Rinvx_out : {in predC unit_R, Rinvx =1 id}.
Proof.
by move=> A; rewrite inE /=/Rinvx   -if_neg => ->.
Qed.

Definition real_unitRingMixin :=
  UnitRingMixin RmultRinvx RinvxRmult intro_unit_R Rinvx_out.
Canonical Structure real_unitRing :=
  Eval hnf in UnitRingType R real_unitRingMixin.

Canonical Structure real_comUnitRingType := 
  Eval hnf in [comUnitRingType of R]. 

Lemma neq_norP (b1 b2: bool): reflect (~ (~~ b1 /\ ~~ b2)) ((b1 || b2)).
Proof.
by case b1; case b2; constructor; auto; case; auto. 
Qed.

Lemma real_idomainMixin : forall p q, p * q = 0 -> (p == 0) || (q == 0).
Proof.
move=> p q Hpq.
apply/neq_norP.
move => [Hc1 Hc2].
move/eqP: Hc1 => Hc1.
move/eqP: Hc2 => Hc2.
have Hc3: p*q <>0.
by apply Rmult_integral_contrapositive_currified.
by elim Hc3.
Qed.

Canonical Structure real_idomainType :=
   Eval hnf in IdomainType R real_idomainMixin.

Lemma unit_n0 : forall x, x != 0 -> x \is a GRing.unit.
Proof.
by move=> x Hx; rewrite /GRing.unit /= /unit_R.
Qed.

Lemma Rinvx0: Rinvx 0 = 0.
Proof.
rewrite /Rinvx.
by case C: (0!=0); move/eqP: C => C; elim C.
Qed.


Definition real_fieldMixin : 
 GRing.Field.mixin_of [unitRingType of R].
Proof.
 move=> x nzx.
 rewrite /GRing.unit /= /unit_R.
apply: nzx.
Defined.

Definition real_fieldIdomainMixin := FieldIdomainMixin real_fieldMixin.

(* Canonical Structure real_field := @FieldType (IdomainType R real_fieldIdomainMixin) 
  real_fieldMixin.*)
Canonical Structure real_field := Eval hnf in FieldType R real_fieldMixin.

(* reflect the order on the reals to bool *)

Definition Rleb r1 r2 := 
 match (Rle_dec r1 r2) with
  |left _ => true
  |right _ => false
end.

Definition Rltb r1 r2 := Rleb r1 r2 && (r1 != r2).
Definition Rgeb r1 r2 := Rleb r2 r1.
Definition Rgtb r1 r2 := Rltb r2 r1.

Lemma RlebP: forall r1 r2, reflect (r1 <= r2) (Rleb r1 r2).
Proof.
move=> r1 r2; rewrite /Rleb; apply: (iffP idP);
by elim (Rle_dec r1 r2) => H.
Qed.

Lemma RltbP: forall r1 r2, reflect (r1 < r2) (Rltb r1 r2).
Proof.
move=> r1 r2; rewrite /Rltb /Rleb; apply: (iffP idP).
elim (Rle_dec r1 r2) => H; move/andP=> [tf Hdif].
by move/eqP : Hdif => Hdif; elim (Rle_lt_or_eq_dec r1 r2 H) => H1.
done.
elim (Rle_dec r1 r2) => H => Hineq.
by rewrite /=; apply/eqP; apply Rlt_not_eq.
by elim H; left.
Qed.

Lemma RgeP: forall r1 r2, reflect (r1 >= r2) (Rgeb r1 r2).
Proof.
move=> r1 r2; rewrite /Rleb; apply: (iffP idP) => H.
by apply Rle_ge;  apply/RlebP.
by apply/RlebP; apply Rge_le.
Qed.

Lemma RgtP: forall r1 r2, reflect (r1 > r2) (Rgtb r1 r2).
Proof.
move=> r1 r2; rewrite /Rleb; apply: (iffP idP) => H.
by apply Rlt_gt;  apply/RltbP.
by apply/RltbP; apply Rgt_lt.
Qed.

End Rstruct.

Ltac toR := rewrite /GRing.add /GRing.opp /GRing.zero /GRing.mul /GRing.inv
  /GRing.one //=. 

(**************************************************************************)

Section ssreal_struct.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Local Open Scope R_scope.

Lemma Rleb_norm_add x y : Rleb (Rabs (x + y)) (Rabs x + Rabs y).
Proof. by apply/RlebP/Rabs_triang. Qed.

Lemma addr_Rgtb0 x y : Rltb 0 x -> Rltb 0 y ->  Rltb 0 (x + y).
Proof. by move/RltbP=> Hx /RltbP Hy; apply/RltbP/Rplus_lt_0_compat. Qed.

Lemma Rnorm0_eq0 x : Rabs x = 0 -> x = 0.
Proof. by move=> H; case: (x == 0) /eqP=> // /Rabs_no_R0. Qed.

Lemma Rleb_leVge x y : Rleb 0 x -> Rleb 0 y -> (Rleb x y) || (Rleb y x).
Proof.
move/RlebP=> Hx /RlebP Hy; case: (Rlt_le_dec x y).
  by move/Rlt_le/RlebP=> ->.
by move/RlebP=> ->; rewrite orbT.
Qed.

Lemma RnormM : {morph Rabs : x y / x * y}.
exact: Rabs_mult. Qed.

Lemma Rleb_def x y : (Rleb x y) = (Rabs (y - x) == y - x).
apply/(sameP (RlebP x y))/(iffP idP)=> [/eqP H| /Rle_minus H].
  apply: Rminus_le; rewrite -Ropp_minus_distr.
  apply/Rge_le/Ropp_0_le_ge_contravar.
  by rewrite -H; apply: Rabs_pos.
apply/eqP/Rabs_pos_eq.
rewrite -Ropp_minus_distr.
by apply/Ropp_0_ge_le_contravar/Rle_ge.
Qed.

Lemma Rltb_def x y : (Rltb x y) = (y != x) && (Rleb x y).
apply/(sameP (RltbP x y))/(iffP idP).
  case/andP=> /eqP H /RlebP/Rle_not_gt H2.
  by case: (Rtotal_order x y)=> // [][] // /esym.
move=> H; apply/andP; split; [apply/eqP|apply/RlebP].
  exact: Rgt_not_eq.
exact: Rlt_le.
Qed.

Definition realNumMixin := NumMixin Rleb_norm_add addr_Rgtb0 Rnorm0_eq0 
  Rleb_leVge RnormM Rleb_def Rltb_def.
Canonical Structure realNumDomainType := NumDomainType R realNumMixin.

Lemma RleP : forall x y, reflect (Rle x y) (Num.le x y).
Proof. exact: RlebP. Qed.

Lemma RltP : forall x y, reflect (Rlt x y) (Num.lt x y).
Proof. exact: RltbP. Qed.

Canonical Structure realNumFieldType := [numFieldType of R].

Lemma Rreal_axiom (x : R) : (Rleb 0 x) || (Rleb x 0).
Proof.
case: (Rle_dec 0 x)=> [/RlebP ->|] //.
by move/Rnot_le_lt/Rlt_le/RlebP=> ->; rewrite orbT.
Qed.

Canonical Structure realRealDomainType := RealDomainType R Rreal_axiom.

Canonical Structure realRealFieldType := [realFieldType of R].

Lemma Rarchimedean_axiom : Num.archimedean_axiom realRealFieldType.
Proof.
move=> x; exists (Zabs_nat (up x) + 2)%N.
have [Hx1 Hx2]:= (archimed x).
have Hz (z : Z): z = (z - 1 + 1)%Z by rewrite Zplus_comm Zplus_minus.
have Zabs_nat_Zopp z : Zabs_nat (- z)%Z = Zabs_nat z by case: z.
apply/RltbP/Rabs_def1.
  apply: (Rlt_trans _ ((Zabs_nat (up x))%:R)%R); last first.
    rewrite -[((Zabs_nat _)%:R)%R]Rplus_0_r mulrnDr.
    by apply/Rplus_lt_compat_l/Rlt_0_2.   
  apply: (Rlt_le_trans _ (IZR (up x)))=> //.
  elim/(well_founded_ind (Zwf_well_founded 0)): (up x) => z IHz.
  case: (Z_lt_le_dec 0 z) => [zp | zn].
    rewrite [z]Hz plus_IZR Zabs_nat_Zplus //; last exact: Zlt_0_le_0_pred.
    rewrite plusE mulrnDr.
    apply/Rplus_le_compat_r/IHz; split; first exact: Zlt_le_weak.
    exact: Zlt_pred.
  apply: (Rle_trans _ (IZR 0)); first exact: IZR_le.
  by apply/RlebP/(ler0n realNumDomainType (Zabs_nat z)). 
apply: (Rlt_le_trans _ (IZR (up x) - 1)).
  apply: Ropp_lt_cancel; rewrite Ropp_involutive. 
  rewrite Ropp_minus_distr /Rminus -opp_IZR -{2}(Zopp_involutive (up x)).
  elim/(well_founded_ind (Zwf_well_founded 0)): (- up x)%Z => z IHz .
  case: (Z_lt_le_dec 0 z) => [zp | zn]. 
    rewrite [z]Hz Zabs_nat_Zopp plus_IZR.
    rewrite Zabs_nat_Zplus //; last exact: Zlt_0_le_0_pred. 
    rewrite plusE -Rplus_assoc -addnA [(_ + 2)%N]addnC addnA mulrnDr.
    apply: Rplus_lt_compat_r; rewrite -Zabs_nat_Zopp.
    apply: IHz; split; first exact: Zlt_le_weak.
    exact: Zlt_pred.
  apply: (Rle_lt_trans _ 1).
    rewrite -{2}[1]Rplus_0_r; apply: Rplus_le_compat_l.
    by rewrite -/(IZR 0); apply: IZR_le.
  rewrite mulrnDr; apply: (Rlt_le_trans _ 2).
    by rewrite -{1}[1]Rplus_0_r; apply/Rplus_lt_compat_l/Rlt_0_1. 
  rewrite -[2]Rplus_0_l; apply: Rplus_le_compat_r.
  by apply/RlebP/(ler0n realNumDomainType (Zabs_nat _)). 
apply: Rminus_le.
rewrite /Rminus Rplus_assoc [- _ + _]Rplus_comm -Rplus_assoc -!/(Rminus _ _).
exact: Rle_minus.
Qed.

Canonical Structure realArchiFieldType := ArchiFieldType R Rarchimedean_axiom.


(*********************************************************************)
Lemma continuity_eq f g : f =1 g -> continuity f -> continuity g.
Proof.
move=> Hfg Hf x eps Heps.
have [y [Hy1 Hy2]]:= Hf x eps Heps.
by exists y; split=> // z; rewrite -!Hfg; exact: Hy2.
Qed.

Lemma continuity_sum (I : finType) F (P : pred I): 
(forall i, P i -> continuity (F i)) ->
  continuity (fun x => (\sum_(i | P i) ((F i) x)))%R.
Proof.
move=> H; elim: (index_enum I)=> [|a l IHl].
  set f:= fun _ => _.
  have Hf: (fun x=> 0) =1 f by move=> x; rewrite /f big_nil.
  by apply: (continuity_eq Hf); exact: continuity_const.
set f := fun _ => _.
case Hpa: (P a).  
  have Hf: (fun x => F a x + \sum_(i <- l | P i) F i x)%R =1 f.
  by move=> x; rewrite /f big_cons Hpa.
  apply: (continuity_eq Hf); apply: continuity_plus=> //.
  exact: H.
have Hf: (fun x => \sum_(i <- l | P i) F i x)%R =1 f.
  by move=> x; rewrite /f big_cons Hpa.
exact: (continuity_eq Hf).
Qed.

Lemma continuity_exp f n: continuity f -> continuity (fun x => (f x)^+ n)%R.
Proof.
move=> Hf; elim: n=> [|n IHn]; first exact: continuity_const.
set g:= fun _ => _.
have Hg: (fun x=> f x * f x ^+ n)%R =1 g.
  by move=> x; rewrite /g exprS. 
by apply: (continuity_eq Hg); exact: continuity_mult.
Qed.
(*******************************************************************)

Lemma Rreal_closed_axiom : Num.real_closed_axiom realArchiFieldType.
Proof.
move=> p a b; rewrite !ler_eqVlt. 
case Hpa: ((p.[a])%R == 0%R).
  by move=> ? _ ; exists a=> //; rewrite lerr ler_eqVlt.
case Hpb: ((p.[b])%R == 0%R).
  by move=> ? _; exists b=> //; rewrite lerr ler_eqVlt andbT.
case Hab: (a == b).
  by move=> _; rewrite (eqP Hab) eq_sym Hpb (ltrNge 0) /=; case/andP=> /ltrW ->.
rewrite eq_sym Hpb /=; clear=> /RltbP Hab /andP [] /RltbP Hpa /RltbP Hpb.
suff Hcp: continuity (fun x => (p.[x])%R). 
  have [z [[Hza Hzb] /eqP Hz2]]:= IVT _ a b Hcp Hab Hpa Hpb.
  by exists z=> //; apply/andP; split; apply/RlebP.
(************************************************)
rewrite -[p]coefK poly_def.
set f := fun _ => _.
have Hf: (fun (x : R) => \sum_(i < size p) (p`_i * x^+i))%R =1 f.
   move=> x; rewrite /f horner_sum.
   by apply: eq_bigr=> i _; rewrite hornerZ hornerXn.
apply: (continuity_eq Hf); apply: continuity_sum=> i _.
apply:continuity_scal; apply: continuity_exp=> x esp Hesp.
by exists esp; split=> // y [].
Qed.

Canonical Structure realRcfType := RcfType R Rreal_closed_axiom.

End ssreal_struct.


