Require Import Reals Fourier Rdefinitions Raxioms RIneq Rbasic_fun.
Require Import Rstruct min_max interval_R (* real_matrix *).
Require Import ssrmatching ssreflect ssrbool ssrnat eqtype ssrfun seq.
Require Import fintype finfun bigop ssralg poly matrix mxpoly complex.
Require Import perm ssrnum mxalgebra polyrcf polyorder realalg zmodp. 
Require Import closed_poly mxorder mxtop struct_top outils.
Require Import perron_frobenius.



(*
abstract the notion of eigenvalues and spectral radius,
admit results: Perron-Frobenius, properties of Rayleigh quotients
*)

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope R_scope.
Delimit Scope R_scope with Re. 


Open Scope ring_scope.
Delimit Scope ring_scope with Ri.

Import GRing.Theory Num.Theory.

Section eigenDef.

(*need to talk about eigenvalues, spectral radius *)
(* the type here was : comUnitRingType *)
Variable T: fieldType.
   
(* Variable n': nat. *)
Variable n : nat.
(* Notation n := n'.+1. *)

Variable A: matrix T n n.

Definition eigenv := root (char_poly A).
(* roots defined for unitRing, char_poly for comRing *)

End eigenDef.

Section detProp.

Variable n: nat.
(*Notation n := n'.+1.*)

Variable T: fieldType.
   

Lemma det0P_cv_n :
forall A: 'M[T]_n,  reflect (forall x: 'cV_n, A *m x = 0 -> x = 0)  (\det A != 0).
Proof.
move=> A.
have H0:= @det0P T n (trmx A).
case C: (\det A != 0); [apply: ReflectT |apply: ReflectF].
move=> x Hx.
replace  x  with ( (invmx A) *m (A *m x)).
by rewrite Hx mulmx0.
rewrite mulmxA mulVmx.
by rewrite mul1mx.
rewrite /unitmx -topredE /=.
rewrite unitfE.
apply: C.
move/negP: C.
move/negP => C.
rewrite -det_tr in C.
elim: (elimT H0 C) => x Hx1 Hx2 Hc.
have Hcx:= Hc (x^T).
have Htx: A *m x^T = 0.
by rewrite -(trmxK A) -trmx_mul Hx2 trmx_const /GRing.zero /=.
move/eqP: Hx1 => Hx1.
elim Hx1.
rewrite /GRing.zero /= -trmx_const -(trmxK x) (Hcx Htx); f_equal.
Qed.

Lemma det0P_cv:
forall A: 'M[T]_n, reflect (exists2 v : 'cV_n, v != 0 & A *m v = 0) (\det A == 0).
Proof.
move=> A.
have H0:= @det0P T n (trmx A).
case C: (\det A == 0); [apply: ReflectT |apply: ReflectF].
rewrite -det_tr in C.
elim: (elimT H0 C) => x Hx1 Hx2 .
exists (x^T).
rewrite /GRing.zero /=  -trmx_const .
apply/eqP => Htx.
move/eqP: Hx1=> Hx1; elim: Hx1.
apply/matrixP => i j; rewrite mxE; move/matrixP: Htx => Htx.
by have:= Htx j i; rewrite /trmx !mxE.
by rewrite -(trmxK A) -trmx_mul Hx2 trmx_const /GRing.zero /=.
rewrite -det_tr in C.
move/eqP: C => Hdet [x Hex1 Hex2].
have H: (exists2 v : 'rV_n, v != 0 & v *m A^T = 0).
exists (x^T).
apply/eqP => Htx.
move/eqP: Hex1=> Hx1; elim: Hx1.
 apply/matrixP => i j; rewrite mxE; move/matrixP: Htx => Htx.
by have:= Htx j i; rewrite /trmx !mxE.
by rewrite -trmx_mul Hex2 trmx_const /GRing.zero /=.
by elim: Hdet; apply/eqP; apply/H0.
Qed.

End detProp.


Section eigenProp.

Variable n: nat.
(* Notation n := n'.+1. *)

Variable T: fieldType.
   

Lemma lam_equ:
forall (A: 'M[T]_n) lam, lam \in eigenv A ->
    exists2 x: 'cV_n, (x != 0) & (A *m x = lam *: x).
Proof.
move=> A lam Hlam.
rewrite /eigenv /root /char_poly in Hlam.
have H:= det0P_cv (lam%:M - A).
have Heq: (\det ('X%:M - map_mx (@polyC T) A)).[lam] = \det (lam%:M - A).
 rewrite /determinant horner_sum.
 apply eq_bigr => s _ .
 rewrite hornerM.
 congr (_ * _ ).
  case (odd_perm s); rewrite /=.
   by rewrite !expr1 hornerN; f_equal; rewrite hornerC.
  by rewrite !expr0 hornerC.
 elim: index_enum.
  by rewrite !big_nil hornerC.
 move=> i t IH.
 rewrite !big_cons -IH hornerE.
 congr (_ * _).
 rewrite !mxE  hornerE hornerE hornerC.
 congr (_ + _).
 by case (i == s i); rewrite /=; [rewrite !mulr1n hornerX| rewrite !mulr0n hornerC].
move: Hlam; rewrite -topredE /= => Hlam.
rewrite Heq in Hlam.
have [x Hx1 Hx2]: (exists2 v : 'cV_n, v != 0 & (lam%:M - A) *m v = 0) by apply/H.
exists x; first by done.
rewrite mulmxBl  -scalemx1 -scalemxAl mul1mx in Hx2.
by move/eqP: Hx2; rewrite subr_eq0; move/eqP=> ->.
Qed.

End eigenProp.

Section eigenR_def.

Variable n: nat.
(* Notation n := n'.+1. *)

(* Check rootsR. *)
Definition eigens (A:'M[R]_n) := rootsR (char_poly A).
(* Check eigens. *)
(* Variable eigens: matrix R n n -> seq R. *)

Lemma Heig A lam : lam \in (eigenv A) = (lam \in (eigens A)).
Proof.
have cAn0 : char_poly A != 0 by rewrite -size_poly_eq0 size_char_poly.
exact: (roots_onP (roots_on_rootsR cAn0)).
Qed.
(*
Hypothesis Heig: forall A lam, lam \in (eigenv A) = (lam \in (eigens A)).
(* all elements in eigens A, satisfy the predicate, i.e. are eigenvalues of A *)
*)

(* spectral radius: we can define it with big[Rmax] on eigens *) 

Definition specr A := \big[Rmax/Rabs (last 0 (eigens A))]_(lam <- (eigens A)) Rabs lam.

End eigenR_def.

Section eigenR.

Variable n: nat.
Lemma lam_le_specr:
  forall (A : 'M_n) lam, lam \in eigens A -> `|lam| <= specr A.
Proof.
rewrite /specr => A lam.
elim: (eigens A) => // t s IH Hlam.
rewrite big_cons.
rewrite in_cons in Hlam.
move/orP: Hlam=> Hlam; elim Hlam; move/eqP => Hlam1.
  by rewrite Hlam1; apply/RleP/RmaxLess1.
apply: (ler_trans (_ : _ <= (\big[Rmax/Rabs (last 0 (t :: s))]_(j <- s) Rabs j))).
  have Hs: s != [::].
    apply/negP; move/eqP => H.
    have H1 := in_nil lam.
    by move/eqP: Hlam1; rewrite H H1.
  rewrite -(min_max.last_cons t Hs).
  by apply: IH; apply/eqP ; apply Hlam1.
apply/RleP/RmaxLess2.
Qed.

(*Lemma lam_le_specr_val:
  forall A lam, lam \in eigenv A -> Rabs lam <= specr A.
Proof.
move=> A lam Hlam.
have H:= @lam_le_specr A lam.
have He:= Heig A.
move/allP : He => He.
have Hq := He lam.
apply H.



Hypothesis def_specr:
  forall A lam, eigenv A lam -> Rabs lam <= specr A.
*)


Definition lam_min (A : 'M_n) :=
    \big[Rmin/ (last 0 (eigens A))]_(i <- eigens A) i.
Definition lam_max (A : 'M_n) :=
    \big[Rmax/(last 0 (eigens A))]_(i <- eigens A) i .

Variable mnorm : can_norm realNumDomainType.

Lemma lam_equ_norm (A : 'M[R]_n) (lam : R) : eigenv A lam -> 
  exists2 x : 'cV_n,  || x: mnorm | = 1 & A *m x = lam *: x .
Proof.
move=> Hlam.
have [x Hx1 Hx2] := @lam_equ n real_field A lam Hlam.
have Hdifx: ||x : mnorm| != 0.
 apply/eqP; apply Rgt_not_eq.
 by apply/RltbP; move : Hx1; rewrite -(cnorm_lt0 mnorm).
have Hunit: (|| x : mnorm | : real_field) \is a GRing.unit.
 by rewrite unitfE.
exists (1 / || x : mnorm | *: x).
 rewrite hom_pos_norm ger0_norm.
  by rewrite -mulrA  mulVr; first by rewrite mul1r.
 by rewrite mulr_ge0 ?ler01 // invr_ge0 pos_norm.
by rewrite scalerA (mulrC lam _) -(scalerA _ lam) -Hx2 -scalemxAr.
Qed.

Lemma eigenv_le_norm (A : 'M[R]_n) (lam : R) :
   eigenv A lam -> `|lam| <= || A: mnorm |.
Proof.
move=> Hlam.
have [x Hxnorm Hxeq] := lam_equ_norm Hlam.
rewrite -(mulr1 `|lam|) -Hxnorm -hom_pos_norm -Hxeq.
apply ler_trans  with (|| A : mnorm | * || x : mnorm |)%Ri.
  exact: prod_ineq_norm.
by rewrite Hxnorm mulr1.
Qed.

Corollary specr_le_norm (A : 'M[R]_n) : specr A <= || A : mnorm |.
Proof.
rewrite /specr.
case C: (eigens A != [::]).
have H1 : (forall i : real_eqType, i \in eigens A -> 0 <= Rabs i)%Re .
move=>*; apply Rabs_pos.
elim: ( @max_exists_seq2 _ Rabs C) => k [Hkex Hkeq].
  by rewrite Hkeq; apply eigenv_le_norm; rewrite -Heig in Hkex.
move/eqP: C => ->.
 rewrite big_nil /last Rabs_R0; exact: pos_norm.
Qed.

Lemma pos_def_eigen_pos:
forall (A: 'M[R]_n) (lam: R), eigenv A lam -> 
 pos_def_gen A -> 0 < lam.
Proof.
move=> A lam Hlam HA.
rewrite /pos_def_gen in HA.
elim: (@lam_equ n real_field A lam Hlam) => x Hx0 Heq.
have HAx:= HA x Hx0.
rewrite -mulmxA Heq -scalemxAr in HAx.
have HAx0 := HAx ord0 ord0.
have/vec_prod_pos := Hx0 => /(_ 0 0).
rewrite !mxE in HAx0 * => Hx0'.
by rewrite -(ltr_pmul2r Hx0') mul0r.
Qed.

Lemma eigenv_tr (A : 'M[R]_n) a :
  a \in eigenv A^T -> a \in eigenv A.
Proof.
rewrite /eigenv /char_poly /char_poly_mx.
by rewrite -{1}tr_scalar_mx -map_trmx -raddfB det_tr.
Qed.

Lemma eigenv_char (A : 'M[R]_n) a (x : 'cV_n) :
  x != 0 -> A *m x = a *: x -> a \in eigenv A.
Proof.
move => xnz xsc; apply/eigenv_tr; rewrite /eigenv.
rewrite unfold_in -[_ == 0]/(root (char_poly A^T) a) -eigenvalue_root_char.
apply/eigenvalueP; exists (x^T).
  by rewrite -mul_mx_scalar -tr_scalar_mx -!trmx_mul mul_scalar_mx xsc.
by apply/negP => /eqP abs; case/negP: xnz; rewrite -[x]trmxK abs trmx0.
Qed.

Lemma specr_scalemx a (A : 'M[R]_n) : specr (a *: A) = `|a| * specr A.
Proof.
rewrite /specr.
case ca : (a == 0).
  rewrite (eqP ca) scale0r normr0 mul0r; case h: (eigens (0 : 'M_n) != [::]).
    move: h; case/(big_max_exists Rabs) => l [le lm]; rewrite lm.
    move:le; rewrite -Heig; case/lam_equ => x [xnz]. 
    move/eqP; rewrite mul0mx eq_sym scaler_eq0 (negbTE xnz) orbF => /eqP ->.
    by rewrite Rabs_R0.
  by rewrite (eqP (negbFE h)) big_nil /= Rabs_R0.
case cl: (eigens A == [::]).
  rewrite (eqP cl) big_nil /= Rabs_R0 mulr0.
  case cl' : (eigens (a *: A)) => [| l e].
    by rewrite big_nil /= Rabs_R0.
  suff : l/a \in [::] by rewrite in_nil.
  rewrite -(eqP cl) -Heig.
  have : l \in (eigens (a *: A)) by rewrite cl' mem_head.
  rewrite -Heig; case/lam_equ => x xnz xsc.
  apply: (eigenv_char xnz).
  by rewrite mulrC -scalerA -xsc -scalemxAl scalerA mulVf ?negbT ?scale1r.
have : eigens A != [::] by exact: negbT.
case/(big_max_exists Rabs) => l [le lm].
have : eigens (a *: A) != [::].
  apply/eqP => abs; suff : (a * l \in eigenv (a *: A)).
    by rewrite Heig abs in_nil.
  move: le; rewrite -Heig; case/lam_equ => x [xnz xsc].
  apply: (eigenv_char xnz).
  by rewrite -scalerA -xsc scalemxAl.
case/(big_max_exists Rabs) => l' [l'e l'm].
rewrite l'm lm; apply/eqP; rewrite eqr_le; apply/andP.
move: l'm lm le l'e.
rewrite -/(specr (a *: A)) -/(specr A) -!Heig  => l'm lm le l'e.
split.
  have nap : 0 < `|a| by rewrite normr_gt0 negbT.
  rewrite -lter_pdivr_mull // -normfV //.
  rewrite -normrM -lm lam_le_specr // -Heig.
  move: l'e; case/lam_equ => x [xnz xsc].
  apply: (eigenv_char xnz).
  by rewrite -scalerA -xsc -scalemxAl scalerA mulVf ?scale1r ?negbT.
rewrite -l'm -normrM lam_le_specr ?andbT //.
move: le; case/lam_equ => x [xnz xsc]; rewrite -Heig; apply: (eigenv_char xnz).
by rewrite -scalerA -xsc scalemxAl.
Qed.

Lemma corPF1 (m : nat) (A : 'M[R]_m): 0 <=m: A -> specr A = \rho A.
Proof. admit. Qed.

Corollary corPF:
forall a A, Mle 0 A ->
   (exists u: 'cV_n,  [/\ 0 <=m: u, u <> 0 & (a *: u) <=m: A *m u]) ->
  a <= specr A.
move => a; case: (lerP a 0).
  move => aneg A _ _; rewrite /specr (ler_trans aneg) //.
  apply big_ind; first exact: normr_ge0.
    by move => x y /RleP xp /RleP yp; apply/RleP/Rmax_ge.
  move => i _; exact: normr_ge0.
wlog : a / a = 1.
  move => wl agt0 A Age0 [u [uge0 unz umul]].
  have aVgt0 : 0 < a ^-1 by rewrite invr_gt0.
  have anz : a != 0 by rewrite neqr_lt agt0 orbT.
  have A'ge0 : 0 <=m: a^-1 *: A.
    move => i j; rewrite !mxE mulr_ge0 ?(@ltrW _ _ (a^-1)) ?invr_gt0 //.
    by move: (Age0 i j); rewrite !mxE.
  have umul' : 1 *: u <=m: (a^-1 *: A) *m u.
    rewrite -scalemxAl => i j; rewrite !mxE.
    rewrite -[1](mulVf anz) -mulrA ler_pmul2l //.
    by move: (umul i j); rewrite !mxE.
  have := wl 1 (erefl _) ltr01 _ A'ge0 (ex_intro _ u (And3 uge0 unz umul')).
  rewrite specr_scalemx.
  have := ltrW aVgt0; rewrite -eqr_norm_id => /eqP ->.
  by move/(ler_wpmul2l (ltrW agt0)); rewrite mulr1 mulrA mulfV // mul1r.
case : n.
  by move => -> _ A Age0 [z [zge0 znz]]; case: znz; apply/matrixP => [] [].
move => n' -> _ A Age0 [z [zge0 znz]]; rewrite scale1r => zsc.
have HAltz k : z <=m: A^+(k + 1) *m z.
  have Htr m : A^+m *m z <=m: A^+( m + 1) *m z.
    elim: m=>[|m IHm]; first by rewrite mul1mx.
    rewrite 2!exprS -!mulmxA; apply: (Mle_wpmul2l _ IHm).
    by move=> i j; apply: (Age0 i j).
  by elim: k=> // k IHk; apply/(Mle_trans IHk); rewrite addn1.
rewrite corPF1 //.
case: (lerP 1 \rho(A)) => // /mx_cvg0P/(cvg_shift _ _ 1) cvg0.
have/continue_cvg Hcvgm := (continu_mull (m:= n'.+1) (A:= z)).
have := Hcvgm _ _ cvg0; rewrite mul0mx => cvg0'.
have := lemx_cvg_lemx HAltz cvg0' => zle0.
case: znz; apply/matrixP => i j; apply/eqP.
by rewrite eqr_le zge0 zle0.
Qed.

Lemma Rayleigh_min:
forall (A:'M_n) (x: 'cV_n), x != 0 ->  
  (lam_min (A^T *m A))%:M <=m:  (x^T *m (A^T *m A) *m x) / (x^T *m x).
Admitted.

Lemma Rayleigh_max:
forall (A:'M_n) (x: 'cV_n), x != 0 -> 
   (x^T *m (A^T *m A) *m x) / (x^T *m x) <=m: (lam_max (A^T *m A))%:M  .
Admitted.
 
End eigenR.

Implicit Arguments eigenv[n T].
Implicit Arguments specr[n].




