Require Import Fourier Rdefinitions Raxioms RIneq Rbasic_fun Rfunctions 
  R_sqrt.
Require Import  ssreflect ssrbool ssrnat eqtype ssrfun seq fintype finfun 
   finset matrix bigop ssralg perm ssrnum zmodp.
Require Import Rstruct min_max (* topology *) interval_R 
  (* real_matrix *) eigen mxorder intMatrix struct_top.
Require Import set_theory.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope R_scope.
Delimit Scope R_scope with Re.

Open Scope ring_scope.
Delimit Scope ring_scope with Ri.

Import GRing.Theory.
Import Num.Theory.

Section MatrixIntSys.

Variable m n: nat.
Variable A: matrix closedInt m n.
Variable b: matrix closedInt m 1.

(* need to coerce interval matrices and vectors to sets (preds) 
  of matrices and vectors of reals or do it manually?*)

Theorem Beeck1:
  forall x, sigma_sol A b x <-> exists t , t \ins (inter (inSetm (mmuls_i A x)) (inSetm b)).
Proof.
move=> x; split=> [Hsol|Hinter].
  elim Hsol=> vA [Ha [vB [Hb Hx]]].
  exists vB;  split; last by done.
  by rewrite -Hx; apply mmuls_i_correct.
elim Hinter => vb [Hvb1 Hvb2].
rewrite /sigma_sol.
elim ( mvmuls_i_compl Hvb1) => vA [Ha1 Ha2] .
exists vA; split; first by done.
exists vb; split; first by done.
by rewrite Ha2.
Qed.

Theorem Beeck2:
 forall x, sigma_sol A b x <-> inSetm (madd_i (mmuls_i A x) (mopp_i b)) 0 .
Proof.
move=> x; rewrite Beeck1; split => Hs.
  elim Hs => t [HtA Htb].
  replace (GRing.zero (matrix_zmodType real_zmodType m (S O)))
    with ( t + (- t)); [|rewrite -matrixP=> i j; rewrite !mxE; toR; ring].
  apply madd_i_correct; first by done.
  by apply mopp_i_correct.
  elim (madd_i_compl Hs) => [av  [Hav1 [Hav2 Havplus]]].
  exists av.1; split; first by done.
  have Hop: av.2 = - av.1.
  move: Havplus;  rewrite -!matrixP => Havplus i j; have hav:= Havplus i j.
  rewrite !mxE; rewrite !mxE in hav.
  replace (@GRing.zero real_zmodType) with (av.1 i j + (-av.1 i j))%Re in hav; 
    last by (toR; ring).
  by symmetry; apply Rplus_eq_reg_l with (av.1 i j).
rewrite Hop in Hav2.
replace b with (mopp_i (mopp_i b)).
  replace (av.1) with (- (- av.1)).
    by apply mopp_i_correct.
  by rewrite  -matrixP => i j; rewrite !mxE; rewrite ->Ropp_involutive.
rewrite /mopp_i -matrixP=> i j; rewrite !mxE /opp_i eq_intervalP //=.
by split; rewrite Ropp_involutive.
Qed.
   
Corollary OettliPrager:
 forall x, sigma_sol A b x <-> 
   Mabs (mmid A *m x - mmid b) <=m: mrad A *m Mabs x + mrad b.
Proof.
move=> x.
rewrite Beeck1.
have Ha: forall u v, 
  (exists t :  'cV_m, t \ins (inter (inSetm u) (inSetm v))) <->
  Mabs (mmid u - mmid v) <=m: mrad u + mrad v.
move=> u v; split => H1.
elim H1=> t [Htu Htv].
move=> i j; rewrite !mxE // .
    have Heq: (mid (u i j) - mid (v i j))%Ri = 
      ((mid (u i j) - t i j) + (t i j - mid (v i j)))%Ri by toR; ring.
    rewrite Heq (ler_trans (ler_norm_add _ _)) ?ler_add //.
      by rewrite distrC; apply/RleP/in_mid_rad .
    by apply/RleP/in_mid_rad.
  have H3:  forall i, Rabs (mid (u i ord0) - mid (v i ord0)) 
     <= rad (u i ord0) + rad (v i ord0) ->
     {r : R | r \in (u i ord0) /\ r \in (v i ord0)}.
    by move=>*; apply/intersect_int/RleP.
have H4:  forall i : 'I_m,
     Rabs (mid (u i ord0) - mid (v i ord0)) <= rad (u i ord0) + rad (v i ord0).
move=> i; have H11:= H1 i ord0.
by rewrite !mxE in H11.
have H5: forall i, {r : R | r \in u i ord0 /\ r \in v i ord0}.
by move=> i; apply H3.
exists (\col_i sval (H5 i)).
by split; move=> i j; rewrite mxE ; case: (H5 i) => [r [Hr1 Hr2]]; 
  rewrite //= ord1.
set a:= (mmuls_i A x).
have Hmid: mmid a = mmid A *m x.
rewrite /a -matrixP=> i k; rewrite !mxE mid_big_add.
apply eq_bigr=> j _.
by rewrite /mmid mxE -mid_mul.
have Hrad: mrad a = mrad A *m Mabs x.
rewrite /a -matrixP=> i k; rewrite !mxE rad_big_add.
apply eq_bigr=> j _.
by rewrite /mrad !mxE  -rad_mul.
rewrite -Hmid -Hrad.
apply: Ha.
Qed.

End MatrixIntSys.


Section basicRegCrit.

Variable n': nat.
Notation n := n'.+1. 
Variable A: matrix closedInt n n.


Lemma reg_all_sol0:
  regular A <-> forall (x: 'cV_n) a, inSetm A a ->  a *m x = 0 -> x = 0.
Proof.
split => H; [move=> x a Ha Hax|].
have H0:= H _ Ha.
have H1: (invmx a) *m  (a *m x) =  (invmx a) *m 0.
by congr (_ *m _).
rewrite mulmxA in H1.
have H2: (invmx a *m a) = 1%:M.
rewrite mulVmx; toR.
rewrite H2 mul1mx in H1.
by rewrite H1 mulmx0.
move=> a Ha.
have H1 := fun x => H x a Ha.
have Hdet:= det0P_cv_n (a: matrix real_field _ _).
by apply/Hdet.
Qed.


Lemma reg_int_sol0:
  regular A <-> (forall x: 'cV_n, inSetm (mmuls_i A x) 0 -> x = 0).
 rewrite reg_all_sol0.
split => H x.
  move=> Hin; elim : (mvmuls_i_compl Hin) => a [Ha1 Ha2].
  by apply: (H _ a); [| rewrite Ha2].
by move=> a Ha Hax; apply: H; rewrite -Hax; apply: mmuls_i_correct.
Qed.


Lemma reg_sig_sol0:
regular A <-> 
(forall x, sigma_sol A (\col_i (R0:closedInt)) x <-> x = 0).
Proof.
rewrite reg_all_sol0.
split=> H x.
  split=> Hss.
    elim: Hss => a [Ha [b [Hb Hab]]].
    apply: (H _ a); first by done.
    rewrite Hab -matrixP => i k; have Hbi := Hb i k.
    move: Hbi; rewrite in_int !mxE //= => Hbi.
    by apply Rle_antisym; elim Hbi.
  exists (minf A); split.
    by move=> i j; rewrite mxE //=; apply: inf_in_int.
  exists (\col_i R0); split.
    by move=> i k; rewrite !mxE in_int //=; split; right.
  rewrite Hss mulmx0.
  by apply/colP => i; rewrite !mxE.
move=> a Ha Hax; rewrite -H; exists a; split; first by done.
exists 0; split; last by done.
by move=> i k; rewrite !mxE in_int; split; right.
Qed.

Lemma mid0 : mid 0%Re = 0.
Proof. by rewrite /mid /= Rplus_0_l /Rdiv Rmult_0_l. Qed.

Lemma rad0 : rad 0%Re = 0.
Proof. by rewrite /rad /= /Rminus Rplus_opp_r /Rdiv Rmult_0_l. Qed.


Lemma reg_ineq_mid_rad:
  regular A <-> 
  (forall x: 'cV_n, Mabs (mmid A *m x) <=m:  mrad A *m Mabs x -> x =  0).
Proof.
rewrite reg_sig_sol0; split=> H x; have Hx:= H x.
  move: Hx; rewrite OettliPrager => Hx.
  move=> Hv; rewrite -Hx => i k; rewrite !mxE /Rminus.
  rewrite (ler_trans (ler_norm_add _ _)) ?ler_add //.
    by have Hvi:= (Hv i k); rewrite !mxE in Hvi.
  by rewrite ler_eqVlt normrN mid0 rad0 normr0 eqxx.
split.
  rewrite OettliPrager => Hop; apply Hx=> i k; rewrite !mxE.
  have Hopi:= (Hop i k); rewrite !mxE in Hopi.
  apply: (ler_trans (_ : _ <= `|\sum_j mmid A i j * x j k - mid R0|)).
    by rewrite mid0 subr0.
  by rewrite (ler_trans Hopi) // rad0 addr0.
move=> Hx0.
exists (minf A); split.
  by move=> i j; rewrite mxE //=; apply: inf_in_int.
exists (\col_i R0); split.
  by move=> i k; rewrite !mxE in_int //=; split; right.
rewrite Hx0 mulmx0.
by apply /colP => i; rewrite !mxE.
Qed.


(* classicaly equivalent to previous lemma *)
Require Import Classical.

Lemma sing_ineq_mid_rad:
singular A <-> 
  (exists x: 'cV_n, Mabs (mmid A *m x) <=m: (mrad A *m Mabs x) /\ x <> 0).
Proof.
rewrite iff_not -reg_nsing_eq  reg_ineq_mid_rad.
split => H.
apply all_not_not_ex=> vA.
apply or_not_and.
apply imply_to_or => Ha Heq; elim Heq.
by apply H.
apply: not_ex_not_all => H1; elim H1=> x Hx.
apply H; exists x.
by apply imply_to_and.
Qed.

Lemma sing_ineq_mid_rad_norm:
singular A <-> 
  (exists x: 'cV_n, Mabs (mmid A *m x) <=m: (mrad A *m Mabs x) /\ x^T *m x = 1).
Proof.
rewrite sing_ineq_mid_rad; split; move=> [x [Hineq Hx]].
  exists ((1/ sqrt ((x^T *m x) ord0 ord0)) *: x).
  split.
    rewrite Mabs_scalemx -!scalemxAr Mabs_scalemx -!mul_scalar_mx.
    apply: Mle_wpmul2l => // i j.
    by rewrite !mxE ; case C: (i == j); rewrite //= mulr1n.
  rewrite [_ ^T]linearZ -scalemxAr -scalemxAl scalerA.
  have/eqP/vec_prod_pos := Hx => / (_ ord0 ord0); rewrite mxE => Hx'.
  have /RltP/sqrt_lt_R0 Hsqrt := Hx'.
  have Heq: ((1 / sqrt ((x^T *m x) ord0 ord0) *
     (1 / sqrt ((x^T *m x) ord0 ord0))))
     = (1/ (x^T *m x) ord0 ord0).
    toR; ring_simplify.
    have Csqrt: (sqrt ((x^T *m x) ord0 ord0) != 0).
      by apply/eqP; apply Rgt_not_eq.
    have Cx: (x^T *m x) ord0 ord0 != 0.
      by rewrite neqr_lt Hx' orbT.
    rewrite /Rinvx Csqrt Cx -Rinv_pow; last by apply Rgt_not_eq. 
    by rewrite /pow Rmult_1_r sqrt_sqrt //; apply /RleP; rewrite ltrW.
  rewrite Heq.
  apply/matrixP => i j; rewrite !ord1 mxE -mulrA mulVr.
    by rewrite  mxE /= mulr1n mulr1 .
  by rewrite unitfE neqr_lt Hx' orbT.
exists x; split; first by done.
apply/eqP; apply/vec_prod_pos; rewrite Hx => i j.
by rewrite !mxE !ord1 /= mulr1n ltr01.
Qed.

Lemma oettli_prager_sing:
  {x: 'cV_n |  Mabs ( mmid A *m x) <=m:  mrad A *m Mabs x /\ 
  x <> 0} ->   {vA | inSetm A vA /\ \det vA == 0}.
Proof.
move=>[x [Hre Hneq]].
set y := (\col_i if ( (mrad A *m Mabs x) i ord0 != 0) then 
                                (mmid A *m x) i ord0 / ((mrad A) *m (Mabs x)) i ord0
                            else 1).
pose z := \col_(i<n) (if (Rleb 0 (x i ord0)) then R1 else (-R1)).
pose vA := \matrix_(i, j) ((mmid A) i j - y i ord0 * z j ord0 *  (mrad A) i j).
exists vA; split.
  move=> i j; rewrite in_mid_rad.
  rewrite /vA mxE.
  replace ( (mmid A i j - y i ord0 * z j ord0 * mrad A i j)%Ri - mid (A i j))%Re
  with
      ((- y i ord0 * z j ord0 * rad (A i j))%Re). 
    rewrite Rabs_mult -{2}(Rmult_1_l (rad (A i j))) (Rabs_right (rad _));
      [| apply Rle_ge; apply rad_pos].
    apply Rmult_le_compat_r; first by apply rad_pos .
    rewrite /z /y mxE; 
      rewrite (mxE _  ( fun i0 _=> (if Rleb 0 (x i0 ord0) then 1 else -1)) j ).
    case Cle : (Rleb 0 (x j ord0)); rewrite Rabs_mult.
      rewrite (Rabs_right 1); [|fourier].
      rewrite  Rmult_1_r.
      case Cle2: ( (mrad A *m Mabs x) i ord0 != 0); 
         [|rewrite Rabs_Ropp Rabs_right; toR; fourier].
      rewrite Rabs_Ropp /Rdiv Rabs_mult.
      apply Rmult_le_reg_l with (Rabs ((mrad A *m Mabs x) i ord0 )).
         by apply Rabs_pos_lt; move/eqP: Cle2.
      rewrite Rmult_comm Rmult_assoc -Rabs_mult.
      have Htor:= @GRing.mulVr real_unitRing ( (mrad A *m Mabs x) i ord0).
      rewrite /GRing.mul /GRing.inv /cancel /= in Htor; toR; rewrite Htor;
      last by apply Cle2.
      rewrite Rabs_R1 !Rmult_1_r.
      have Hrei:= Hre i ord0.
      rewrite mxE in Hrei; apply/RleP/(ler_trans Hrei).
      rewrite Rabs_right // mxE.
      apply big_ind; first by right.
        by move=> a b Hz Ht; apply Rle_ge; apply: Rplus_le_le_0_compat; 
        apply Rge_le.
      move=> k _; rewrite mxE mxE; apply Rle_ge; apply Rmult_pos;
      [apply rad_pos| apply Rabs_pos].
    rewrite !Rabs_Ropp Rabs_R1 Rmult_1_r.
    case Cle2: ( (mrad A *m Mabs x) i ord0  != 0);
         [|rewrite Rabs_R1; right; ring].
    rewrite /Rdiv Rabs_mult.
    apply Rmult_le_reg_l with (Rabs ((mrad A *m Mabs x) i ord0)).
      by apply Rabs_pos_lt; move/eqP: Cle2.
    rewrite Rmult_comm Rmult_assoc -Rabs_mult.
    have Htor:= @GRing.mulVr real_unitRing ( (mrad A *m Mabs x) i ord0).
    rewrite /GRing.mul /GRing.inv /cancel /= in Htor; toR; rewrite Htor;
      last by apply Cle2.
    rewrite Rabs_R1 !Rmult_1_r.
    have /RleP Hrei:= Hre i ord0.
    rewrite mxE in Hrei; apply Rle_trans with (1:=Hrei).
    rewrite Rabs_right; first by right.
    rewrite mxE.
    apply big_ind; first by right.
      by move=> a b Hz Ht; apply Rle_ge; apply: Rplus_le_le_0_compat; 
        apply Rge_le.
    move=> k _; rewrite mxE mxE; apply Rle_ge; apply Rmult_pos;
      [apply rad_pos| apply Rabs_pos].
  rewrite /mmid /mrad !mxE; toR; ring.
have Hdet:= det0P_cv_n (vA: matrix real_field _ _).
have Heq:  (\det vA == 0) =  (~~ (\det vA != 0)).
  case ((\det vA == 0)); by rewrite negbK.
rewrite Heq.
apply/Hdet => H1.
apply Hneq.
apply H1.
apply/matrixP => i k; rewrite !mxE /vA .
transitivity (\sum_j (mmid A i j - y i ord0 * z j ord0 * mrad A i j)* x j ord0 ).
  by apply eq_bigr => j _; rewrite mxE !ord1.
transitivity 
  (\sum_j ((mmid A i j * x j ord0 ) - (y i ord0 * (z j ord0 * mrad A i j * x j ord0)) )).
  apply eq_bigr => j _; by toR; ring.
rewrite GRing.sumrB -big_distrr /=.
rewrite /y mxE;  case Cle: ((mrad A *m Mabs x) i ord0 != 0).
  rewrite !mxE.
  have Heq1:  (\sum_j mrad A i j * Mabs x j ord0) = 
  (\sum_(i0 < n) z i0 ord0 * mrad A i i0 * x i0 ord0) .
    apply eq_bigr => j _; rewrite (GRing.mulrC (z j ord0)) -GRing.mulrA.
    apply Rmult_eq_compat_l.
    rewrite /Mabs  /z !mxE /Rabs.
  case Cle2: (Rleb 0 (x j ord0)); move/RlebP/RleP: Cle2 => Cle2.
      by apply/eqP; rewrite mul1r eqr_norm_id.
    by apply/eqP; rewrite mulNr mul1r eqr_normN ltrW // ltrNge.
  rewrite -Heq1 -mulrA mulrC  mulVr; 
  last by (rewrite mxE in Cle; apply Cle).
  rewrite mul1r; toR; ring.
move/eqP: Cle => Cle.
rewrite mxE in Cle.
have Heq1:  (\sum_j mrad A i j * Mabs x j ord0) = 
  (\sum_(i0 < n) z i0 ord0 * mrad A i i0 * x i0 ord0) .
  apply eq_bigr => j _; rewrite (GRing.mulrC (z j ord0)) -GRing.mulrA.
  apply Rmult_eq_compat_l.
  rewrite /Mabs  /z !mxE /Rabs.
  case Cle2: (Rleb 0 (x j ord0)); move/RlebP/RleP: Cle2 => Cle2.
    by apply/eqP; rewrite mul1r eqr_norm_id.
  by apply/eqP; rewrite mulNr mul1r eqr_normN ltrW // ltrNge.
rewrite -Heq1 Cle mulr0 oppr0 addr0.
by have := Hre i ord0; rewrite !mxE !Cle normr_le0 => /eqP.
Qed.


End basicRegCrit.


Section midInv_RegCrit.

Variable n': nat.
Notation n:= n'.+1.

Variable A: matrix closedInt n n.

Lemma Mabs_neg0 : forall n p (x : 'M[R]_(n,p)), x <> 0 -> Mabs x <> 0.
move=> n p x nzx; move/matrixP => mx0; apply/nzx/matrixP => i j.
by have := (mx0 i j); rewrite !mxE; apply: Rnorm0_eq0.
Qed.

(* uses proof by contradiction *)
Theorem thm31_midinva_specr:
forall X, 
specr (Mabs (1 - X * mmid A) + Mabs X * mrad A )  < 1 -> regular A.
Proof.
move=> M Hsr; change ('M[R]_n) in M.
apply nsing_reg; rewrite sing_ineq_mid_rad => Hsing.
elim Hsing => x [Hle Hdif0] {Hsing}.

have Hinq: Mabs x <=m:  
  (Mabs (1 - M * mmid A) + Mabs M * mrad A )  *m (Mabs x).
  have H1:  x =   (1 - M * mmid A) *m x +  (M * mmid A) *m x.
    rewrite mulmxBl mul1mx -matrixP => i k ; rewrite !mxE; toR; ring.
  rewrite {1}H1.
  apply: (Mle_trans (Mabs_triang _ _)).
  apply: (Mle_trans (_ : _ <=m: ( Mabs (1 - M * mmid A) *m (Mabs  x) +  
   (Mabs M ) *m (Mabs (mmid A *m x))))).
    apply: Mle_add; first by apply: MabsM.
    by rewrite -mulmxA; apply MabsM.
  rewrite mulmxDl; apply Mle_add.
    by move=> i j.
  rewrite -mulmxA; apply Mle_wpmul2l; last by [].
  by move=> i j; rewrite !mxE; apply normr_ge0.
have Heq:  1 *: (Mabs x ) = Mabs x.
  by apply/matrixP=> i j; rewrite mxE; toR; ring.
move: Hinq; rewrite -{1}Heq => Hinq.
pose a := (Mabs (1 - M * mmid A) + Mabs M * mrad A) .
move: Hsr Hinq; rewrite -/a => Hsr Hinq.
move: Hsr; apply/negP; rewrite -lerNgt; apply:corPF.
  rewrite /a; move => i j; rewrite !mxE.
  apply/addr_ge0; first exact: normr_ge0.
  apply big_ind;  [by []| apply addr_ge0|].
  move=> k _; rewrite !mxE; apply: mulr_ge0; [apply: normr_ge0|].
  by apply/RleP/ rad_pos.
exists (Mabs x); split => //; last exact: Mabs_neg0.
by move=> i j; rewrite !mxE; apply /RleP/Rabs_pos.
Qed.

Corollary cor32_midinv_specr:
  \det (mmid A) != 0 -> 
  specr (Mabs (mmid A)^-1 * mrad A )  < 1 -> regular A.
Proof.
move=> Hma Hsr.
apply (@thm31_midinva_specr ( (mmid A)^-1)).
rewrite mulVr; last by [].
rewrite addrN.
have Heq : @Mabs realNumDomainType n n 0 = 0.
  by rewrite /Mabs ; apply/matrixP => i j; rewrite !mxE; apply Rabs_R0.
by rewrite Heq add0r.
Qed.

Theorem thm33:
(exists Rm , exists j,  
   \col_i (1 + Mabs ( 1 - mmid A * Rm))%Ri i j <=m:  
   \col_i (mrad A * (Mabs Rm))%Ri i j)
 -> singular A.
Proof.
move=> [Rm [j Hle]].
have Hin: Mabs (mmid A *m (\col_i Rm i j)) <=m: 
  mrad A *m Mabs (\col_i Rm i j).
have -> : ( Mabs (mmid A *m \col_i Rm i j)) =
   (\col_i (Mabs ((mmid A) *  Rm) i j)).
    apply/matrixP => i k; rewrite !mxE /=.
    by congr (`| _ |); apply eq_bigr=> r_; rewrite mxE.
  have -> : \col_i Mabs (mmid A * Rm) i j =
            \col_i Mabs (1 - (1 - (mmid A) * Rm)) i j.
    by apply/matrixP=> i k; rewrite !mxE opprB addrCA addrN addr0.
  move=> i j0; apply: (ler_trans (_ : _ <= (1 + (Mabs (1 - mmid A * Rm))) i j)).
    rewrite !mxE (ler_trans (ler_norm_add _ _)) //.
    apply ler_add.
      by case: (i == j); rewrite ler_eqVlt eqr_norm_id ?ler01 ?lerr.
    by rewrite normrN.
  have := (Hle i ord0); rewrite mxE => Hlei; apply: (ler_trans Hlei). 
  rewrite !mxE ler_eqVlt; apply /orP; left; apply/eqP.
  by apply eq_bigr => k _; rewrite !mxE .
have Hsa: {vA| inSetm A vA /\ \det vA == 0}.
  apply oettli_prager_sing.
  exists (\col_i (Rm i j)).
  split; first by exact Hin.
  apply/matrixP => Hct.
  have Hlej:= Hle j ord0.
  move: Hlej; rewrite !mxE (eq_refl j).
  apply/negP; rewrite -ltrNge.
  apply: (ler_lt_trans (_ : _ <= 0)).
    rewrite ler_eqVlt; apply/orP; left; apply/eqP.
    apply: big1 => i _ ; rewrite !mxE .
    have Hcti := Hct i ord0.
    by rewrite !mxE in Hcti ; rewrite Hcti normr0 mulr0.
  apply: (ltr_le_trans (_ : _ < 1)); first exact: ltr01.
  by rewrite -{1}(addr0 1) ler_add2l normr_ge0.
elim Hsa => [va [H1 H2]].
by exists va; split; [|apply/eqP].
Qed.


Corollary cor34:
  \det (mmid A) != 0 -> 1 <=
  \big[Rmax/0]_(j< n) (mrad A * Mabs (mmid A)^-1) j j  
  -> singular A.
Proof.
move=> Hn0 Hle.
apply  thm33.
exists ((mmid A)^-1).
have Hpn : (0 < n)%Dn by [].
elim: (@max_exists n Hpn (fun j =>(mrad A * Mabs (mmid A)^-1) j j)).
  move=> j Heq.
  exists j.
  move=> i k.
  rewrite  mulrV.
    rewrite addrN.
    case C: (i==j).
      have C1: i=j by move/eqP: C.
      rewrite C1 !ord1 !mxE .
      rewrite !mxE in Heq.
      rewrite  -Heq.
      apply: ler_trans Hle.
      by rewrite normr0 addr0 eq_refl.
    rewrite !mxE C normr0 addr0 /=.
  apply big_ind; first by [].
      by move=> x y Hx Hy; apply addr_ge0.
    move=> k0 _; rewrite !mxE.
    by apply mulr_ge0;[apply/RleP; apply rad_pos|apply normr_ge0].
  by apply Hn0.
move=> i; rewrite !mxE; apply big_ind; first by right.
  by move => x y Hx Hy; apply Rplus_le_le_0_compat.
by move=> k _; apply Rmult_le_pos; rewrite !mxE; [apply rad_pos|apply Rabs_pos].
Qed.

End midInv_RegCrit.


Section eigen_RegCrit.

Variable n': nat.
Notation n:= n'.+1.

Variable A: matrix closedInt n n.


Theorem thm41_lam_minmax:
 lam_max ((mrad A) ^T *m mrad A) < 
  lam_min ((mmid A)^T *m mmid A) ->
regular A.
Proof.
move=> Hlam.
apply nsing_reg; rewrite sing_ineq_mid_rad => Hsing.
elim Hsing => x [Hle Hdif0].
move/matrixP : (Hdif0) => Hdif0'.
move/eqP: Hdif0 => Hdif0.
have [i0 H] : exists i0, x i0 ord0 != 0.
  have Hdif2' : ~~[forall i, x i 0 == (0 : 'cV_n) i 0].
    apply/negP => abs; case: Hdif0' => u v; apply/eqP.
    by rewrite ord1; have/forallP -> := abs.
  move: Hdif2'; rewrite negb_forall => /existsP; case => i'; rewrite mxE.
  by exists i'.
move: Hlam; rewrite ltrNge =>/negP; case.
have Hxtx: 0 <=m: (x^T *m x)^-1.
  have /vec_prod_pos xtxn0 : x != 0.
    by apply/eqP/matrixP => abs; case/negP: H;apply /eqP; rewrite abs mxE.
  rewrite (mx11_scalar (x^T *m x)) [Z in _ <=m: Z]invmx_scalar.
  move => i j; rewrite !ord1 !mxE eqxx.
  rewrite /GRing.natmul /= /GRing.inv /= /Rinvx /=.
  set w := \sum_(k < _) _.
  move:  (xtxn0 ord0 ord0); rewrite !mxE /w => xt'. 
  have -> : (w != 0) = true by rewrite neqr_lt xt' orbT.
  by apply/RleP; apply Rlt_le, Rinv_0_lt_compat; apply/RltP.  
have Hm: @scalar_mx _ 1 (lam_min ((mmid A)^T * mmid A)) <=m: 
    (lam_max ((mrad A)^T *m mrad A))%:M.
  have HR := Rayleigh_min (mrad A) Hdif0.
  apply Mle_trans with (1:= Rayleigh_min (mmid A) Hdif0).
  rewrite mulmxA -trmx_mul.
  apply Mle_trans with 
    (Mabs ((mmid A *m x)^T) *m Mabs (mmid A *m x)  / (x^T *m x) ).
    apply: Mle_wpmul2r; first by [].
    apply: Mle_trans (MabsM _ _).
    by rewrite mulmxA; move=> i j; rewrite !mxE;  apply/RleP/RRle_abs.
  apply Mle_trans with 
    ( (mrad A *m  Mabs  x)^T *m  (mrad A *m  Mabs x) / (x^T *m x) ).
    apply: Mle_wpmul2r; first by [].
  apply: Mle_wpmul; try exact: Mabs_ge0.
      by rewrite /trmx=> i j; move: (Hle j i); rewrite !mxE.
    exact: Hle.
  rewrite trmx_mul .
  apply Mle_trans with 
    ( ((Mabs x)^T *m ((mrad A)^T *m mrad A) *m Mabs x) / (x^T *m x) ).
    by move=> i j; rewrite !mulmxA.
  have Heqx: x^T *m x = (Mabs x) ^T *m Mabs x.
    apply/matrixP=> i j; rewrite !mxE !ord1.
    apply: eq_bigr => k _; rewrite /trmx !mxE.
    by rewrite -!expr2 real_normK // num_real.
  rewrite Heqx; apply: Rayleigh_max.
  by rewrite Mabs_eq0.
have Hm1 := mx11_scalar ((lam_min ((mmid A)^T * mmid A))%:M).
rewrite Hm1  in Hm.
have Hm2 := mx11_scalar ((lam_max ((mrad A)^T * mrad A))%:M).
rewrite Hm2 in Hm.
rewrite !mxE in Hm.
have Hm0:= Hm ord0 ord0.
by rewrite !mxE /= !GRing.mulr1n in Hm0.
Qed.

End eigen_RegCrit.

Section posDef_RegCrit.

Variable n' : nat.
Notation n := n'.+1.

Variable A : matrix closedInt n n.

Variable mnorm : can_norm realNumDomainType.

Theorem thm51_posdef:
pos_def_gen ((mmid A)^T * mmid A - || (mrad A) ^T * mrad A : mnorm| *: 1 ) ->
  regular A.
Proof.
move=> Hpos.
apply nsing_reg; rewrite sing_ineq_mid_rad_norm => Hsing.
elim Hsing => x [Hle Hnorm1].
have Hdif0 : x != 0.
  apply/vec_prod_pos; rewrite Hnorm1 => i j.
  by rewrite !mxE !ord1 //= mulr1n ltr01.
have/eqP/matrixP Hdif0' := Hdif0.
rewrite /pos_def_gen in Hpos.
have [i0 H] : exists i0, x i0 ord0 != 0.
  have Hdif2' :
     ~~[forall i, x i ord0 == (0 : 'cV[R]_n) i ord0].
    apply/negP => abs; case: Hdif0' => u v; apply/eqP.
    by rewrite ord1; have/forallP -> := abs.
  move: Hdif2'; rewrite negb_forall => /existsP; case => i'; rewrite mxE.
  by exists i'.
have Hposx := Hpos x Hdif0.
have Hc:  x^T *m ((mmid A)^T * mmid A - || (mrad A)^T * mrad A : mnorm| *: 1) *m x
 <=m: 0.
  rewrite mulmxBr mulmxBl.
  have H2 : x^T *m ((mmid A)^T * mmid A) *m x <=m:
  x^T *m (|| (mrad A)^T * mrad A : mnorm| *: 1) *m x.
    rewrite mulmxA -trmx_mul.
    apply Mle_trans with (Mabs ((mmid A *m x)^T) *m Mabs (mmid A *m x)).
      apply: Mle_trans (MabsM _ _).
      rewrite mulmxA; move=> i j; rewrite !mxE;  apply/RleP/RRle_abs.
    apply Mle_trans with ((mrad A *m  Mabs  x)^T *m  (mrad A *m  Mabs x)).
      apply: Mle_wpmul; try exact: Mabs_ge0.
        by rewrite /trmx=> i j; move: (Hle j i); rewrite !mxE.
      exact: Hle.
    rewrite trmx_mul .
    apply Mle_trans with 
      ( ((Mabs x)^T *m ((mrad A)^T *m mrad A) *m Mabs x) ).
      by move=> i j; rewrite !mulmxA.
    apply Mle_trans with 
      ((Mabs x)^T *m ((mrad A)^T *m mrad A) *m Mabs x / ((Mabs x)^T *m Mabs x)).
      have Heqx: x^T *m x = (Mabs x) ^T *m Mabs x.
        apply/matrixP=> i j; rewrite !mxE !ord1.
        apply: eq_bigr => k _; rewrite /trmx !mxE.
        by rewrite -!expr2 real_normK // num_real.
      by rewrite -Heqx Hnorm1 invr1 mulr1 => i j.
    apply Mle_trans with (lam_max (((mrad A)^T *m mrad A)))%:M.
      apply: Rayleigh_max.
      apply/eqP=> Habs; move/eqP: Hdif0=> Hdif0 ; 
          apply: Hdif0;  apply/matrixP => i j; rewrite mxE.
      move: Habs; rewrite -matrixP=> Habs.
      have Hij := Habs i j.
      rewrite !mxE in Hij.
      by apply/eqP; rewrite -normr_eq0 Hij.
    apply Mle_trans with (specr ((mrad A)^T *m mrad A))%:M.
      move=> i j; rewrite !mxE (ord1_0 i) (ord1_0 j) //= !mulr1n /lam_max /specr.
      elim ( (eigens ((mrad A)^T *m mrad A))).
        by rewrite !big_nil normr_ge0.
      move=> t s IH; rewrite !big_cons.
      case Hs: (s != [::]).
        apply/RleP/rmax4; first by apply RRle_abs.
        by apply /RleP; rewrite -(last_cons t Hs).
      move/eqP: Hs => ->.
      by rewrite !big_nil; apply/RleP/rmax4; apply: RRle_abs.
    apply Mle_trans with (||(mrad A)^T * mrad A : mnorm|)%:M.
      move=> i j; rewrite !mxE  (ord1_0 i) (ord1_0 j) //= !mulr1n.
      exact: specr_le_norm.
    by rewrite -scalemxAr mulmx1 -scalemxAl Hnorm1 scalemx1 => i j.
  move=> i j; have h3 := H2 i j.
  rewrite addrC mxE.
  rewrite -(ler_add2l (( x^T *m (||(mrad A)^T * mrad A : mnorm| *: 1) *m x) i j)).
  rewrite addrA !mxE addrN add0r addr0.
  by move: h3; rewrite !mxE.
by have/andP := conj (Hposx 0 0) (Hc 0 0); rewrite ltr_le_asym.
Qed.

End posDef_RegCrit.
