Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq fintype finfun zmodp.
Require Import finset ssralg  matrix bigop perm ssrnum rat realalg complex.
Require Import outils.
 
(* Require Import Reals. *)
(* Require Import Fourier. *)
  
(* Require Import ClassicalChoice. *)

(* Require Import Rstruct min_max. *)



(*
- specific definitions for real matrices: absolute value, order component wise;
- definition for a canonical norm structure
- properites of the norm
- definition for limit and convergence of matrix series - componentwise
- properties for limit, convergence ...
- link between norm and determinant 
*)

(* Reserved Notation "\maxr_ i F" *)
(*   (at level 41, F at level 41, i at level 0, *)
(*            format "'[' \maxr_ i '/ '  F ']'"). *)
(* Reserved Notation "\minr_ ( i <- r | P ) F" *)
(*   (at level 41, F at level 41, i, r at level 50, *)
(*            format "'[' \minr_ ( i  <-  r  |  P ) '/ '  F ']'"). *)
(* Reserved Notation "\minr_ ( i <- r ) F" *)
(*   (at level 41, F at level 41, i, r at level 50, *)
(*            format "'[' \minr_ ( i  <-  r ) '/ '  F ']'"). *)
(* Reserved Notation "\maxr_ ( m <= i < n | P ) F" *)
(*   (at level 41, F at level 41, i, m, n at level 50, *)
(*            format "'[' \maxr_ ( m  <=  i  <  n  |  P ) '/ '  F ']'"). *)
(* Reserved Notation "\maxr_ ( m <= i < n ) F" *)
(*   (at level 41, F at level 41, i, m, n at level 50, *)
(*            format "'[' \maxr_ ( m  <=  i  <  n ) '/ '  F ']'"). *)
(* Reserved Notation "\maxr_ ( i | P ) F" *)
(*   (at level 41, F at level 41, i at level 50, *)
(*            format "'[' \maxr_ ( i  |  P ) '/ '  F ']'"). *)
(* Reserved Notation "\maxr_ ( i : t | P ) F" *)
(*   (at level 41, F at level 41, i at level 50, *)
(*            only parsing). *)
(* Reserved Notation "\maxr_ ( i : t ) F" *)
(*   (at level 41, F at level 41, i at level 50, *)
(*            only parsing). *)
(* Reserved Notation "\maxr_ ( i < n | P ) F" *)
(*   (at level 41, F at level 41, i, n at level 50, *)
(*            format "'[' \maxr_ ( i  <  n  |  P ) '/ '  F ']'"). *)
(* Reserved Notation "\maxr_ ( i < n ) F" *)
(*   (at level 41, F at level 41, i, n at level 50, *)
(*            format "'[' \maxr_ ( i  <  n )  F ']'"). *)
(* Reserved Notation "\maxr_ ( i 'in' A | P ) F" *)
(*   (at level 41, F at level 41, i, A at level 50, *)
(*            format "'[' \maxr_ ( i  'in'  A  |  P ) '/ '  F ']'"). *)
(* Reserved Notation "\maxr_ ( i 'in' A ) F" *)
(*   (at level 41, F at level 41, i, A at level 50, *)
(*            format "'[' \maxr_ ( i  'in'  A ) '/ '  F ']'"). *)


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* Open Scope R_scope. *)
Local Open Scope ring_scope.

(* Delimit Scope R_scope with Re. *)
(* Delimit Scope ring_scope with Ri. *)


Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Section matRDef.

Variable R : numDomainType.
Variable p q: nat.

(*Definitions for absolute value of a matrix and componentwise order *)

Definition Mabs (A: 'M[R]_(p, q)) := map_mx normr A.

(*Definition Mabs (A: 'M_(p, q)) := \matrix_(i, j) Rabs (A i j).*)

Definition Mle (A B: 'M[R]_(p, q)) := forall i j, A i j <= B i j. 

Definition Mlt (A B: 'M[R]_(p, q)) := forall i j, A i j < B i j.

Definition one_M: matrix R p q := \matrix_(i, j) 1.

End matRDef.

Implicit Arguments one_M[p q].

Infix "<m:" := Mlt (at level 70): ring_scope.
Infix "<=m:" := Mle (at level 70): ring_scope.


Section matRProp.

Variable R : numDomainType.
Variable n p q: nat.

Open Scope ring_scope.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

(* Lemma Rmat_neq: *)
(* forall (A B:'M[R]_(n, p)), A != B -> exists i, exists j,  A i j != B i j. *)
(* Proof. *)
(* move=> A B; move/eqP=> Hab. *)
(* have H: ~ forall i j, A i j = B i j. *)
(* by move=> H1; apply: Hab; apply/matrixP. *)
(* elim (not_all_ex_not _ _ H) => i H2. *)
(* exists i. *)
(* elim (not_all_ex_not _ _ H2) => j H3. *)
(* exists j. *)
(* by apply/eqP. *)
(* Qed. *)

Lemma Mle_trans (A B C : 'M[R]_(p, q)) : A <=m: B -> B <=m: C -> A <=m: C.
Proof. by move=> HAB HBC i j; rewrite (ler_trans (HAB i j)). Qed.

Lemma Mle_refl (A : 'M[R]_(p, q)) : A <=m: A.
Proof. by []. Qed.

Lemma Mlt_trans (A B C : 'M[R]_(p, q)) : A <m: B -> B <m: C -> A <m: C.
Proof. by move=> HAB HBC i j; rewrite (ltr_trans (HAB i j)). Qed.

Lemma Mlt_le_trans (A B C : 'M[R]_(p, q))  : A <m: B -> B <=m: C -> A <m: C.
Proof. by move=> HAB HBC i j; rewrite (ltr_le_trans (HAB i j)). Qed.

Lemma MltW (A B : 'M[R]_(p, q)) : A <m: B -> A <=m: B.
Proof. by move=> HAB i j; rewrite ltrW. Qed.

Lemma Mabs_eq0 (A :  'M[R]_(p, q)) : ((Mabs A) == 0) = (A == 0).
Proof. 
apply/idP/idP=> [/eqP/matrixP H|/eqP ->]; apply/eqP/matrixP=> i j; rewrite !mxE.
  apply/eqP; rewrite -normr_eq0.
  by have/(_ i j) := H; rewrite !mxE=> ->.
by rewrite normr0.
Qed.

Lemma Mabs_ge0 (A :  'M[R]_(p, q)) : 0 <=m: Mabs A.
Proof. by move=> i j; rewrite !mxE normr_ge0. Qed.

Lemma Mabs_triang (A B : 'M[R]_(p,q)) : (Mabs (A + B)) <=m: (Mabs A + Mabs B).
Proof. by move=> i k; rewrite !mxE ler_norm_add. Qed.

Lemma Mle_add:
forall (a b c d: 'M[R]_(p,q)), Mle a b -> Mle c d -> Mle (a + c) (b + d).
Proof. by move=> a b c d Hab Hcd i k; rewrite !mxE ler_add. Qed.

Lemma MabsM (M : 'M[R]_(n, p)) (N :'M_(p, q)) :
  Mabs (M *m N) <=m: (Mabs M) *m (Mabs N).
Proof.
move=> i k; rewrite !mxE (ler_trans (ler_norm_sum _ _ _)) //.
by apply: ler_sum=> j _; rewrite normrM !mxE. 
Qed.

Lemma Mabs_scalemx  a (M : 'M[R]_(n, p)) : Mabs (a *: M) = `|a| *: Mabs M.
Proof. by apply/matrixP=> i j; rewrite !mxE normrM. Qed.

Lemma eq_Mabs_id (M : 'M[R]_(n, p)) : 0 <=m: M -> Mabs M = M.
Proof. 
move=> Hge0; apply/matrixP=> i j; rewrite mxE.
apply/eqP; rewrite eqr_norm_id.
by have/(_ i j) := Hge0; rewrite mxE.
Qed.

Lemma subr_Mge0 (A B : 'M[R]_(p,q)) : B <=m: A <-> 0 <=m: (A - B).
Proof. 
split=> HAB i j; rewrite ?mxE ?subr_ge0 //.
by have/(_ i j) := HAB; rewrite !mxE subr_ge0.
Qed.

Lemma subr_Mgt (A B C : 'M[R]_(p,q)) : B + A <m: C <-> A <m: C - B.
Proof. by split=> H i j; have/(_ i j):= H; rewrite !mxE ltr_subr_addl. Qed.

Lemma mulmx_gt0 (A: 'M[R]_(p,q)) (x : 'cV_q) : 0 <m: A -> 
   0 <=m: x -> x != 0 -> 0 <m: A *m x.
Proof.
case: q A x=>[A x|m A x].
  by have ->: x == 0 by apply/eqP/matrixP=> [][][].  
move => Apos xpos xneq0 i j; rewrite !mxE.
pose k := odflt ord0 [pick k | x k j != 0].
rewrite (bigD1 k) //= ltr_spaddl // ?mulr_gt0 // ?[_ < x _ _]ltr_def.
    by have/(_ i k) := Apos; rewrite mxE. 
  have/(_ k j):= xpos; rewrite mxE => ->.
  rewrite andbT /k; case: pickP=> //= predx.
  rewrite predx -(negbTE xneq0).
  apply/eqP/matrixP=> h l; rewrite mxE; apply/eqP.
  by rewrite /is_true -[true]negbK [X in ~~ X]/= -(predx h) !ord1 negbK.
apply: sumr_ge0=> l _; rewrite mulr_ge0 //.
  by have/(_ i l):= Apos; rewrite mxE; apply: ltrW.
by have/(_ l j):= xpos; rewrite mxE.
Qed.

Lemma Mlt_scaler (A B : 'M[R]_(p,q)) a : 0 < a -> A <m: B -> a *: A <m: a *: B.
Proof. by move=> Hagt0 HAltB i j; rewrite !mxE ltr_pmul2l. Qed.

Lemma ltr_sum (I : eqType) (r : seq I) (P : pred I) (F G : I -> R) (i : I) :
 (forall i, P i -> F i < G i) -> i \in r -> P i ->
 \sum_(i <- r | P i) F i < \sum_(i <- r | P i) G i.
Proof.
move=> Hlt.
elim: r=> // a l IHl.
rewrite in_cons=> /orP []=> [/eqP -> Ha|Hi HPi].
  rewrite !big_cons Ha ltr_le_add // ?Hlt // ler_sum // => j Hj.
  by rewrite ltrW // Hlt //.
rewrite !big_cons; case H: (P a); last exact: IHl.
by rewrite ltr_add ?Hlt // IHl. 
Qed.

Lemma Mlt_pmul2r (A B : 'M[R]_(n, p.+1)) (C :'M_(p.+1, q)) : 
  0 <m: C ->  A <m: B -> A *m C <m: B *m C. 
Proof.
move=> HCgt0 HAltB i j; rewrite !mxE.
apply: (ltr_sum (i:= ord0))=> // [k _|]; last by rewrite mem_index_enum.
have/(_ k j):= HCgt0; rewrite mxE=> HC.
have/(_ i k) HA:= HAltB.
by rewrite ltr_pmul2r. 
Qed.

Lemma Mlt_pmul2l (C : 'M[R]_(n, p.+1)) (A B :'M_(p.+1, q)) : 
  0 <m: C ->  A <m: B -> C *m A <m: C *m B. 
Proof.
move=> HCgt0 HAltB i j; rewrite !mxE.
apply: (ltr_sum (i:= ord0))=> // [k _|]; last by rewrite mem_index_enum.
have/(_ i k):= HCgt0; rewrite mxE=> HC.
have/(_ k j) HA:= HAltB.
by rewrite ltr_pmul2l. 
Qed.

Lemma Mle_wpmul2l (C : 'M[R]_(n, p)) (A B :'M_(p, q)) : 
  0 <=m: C ->  A <=m: B -> C *m A <=m: C *m B. 
Proof.
move=> HCgt0 HAltB i j; rewrite !mxE.
apply: (ler_sum )=> // k _.
have/(_ i k):= HCgt0; rewrite mxE=> HC.
have/(_ k j) HA:= HAltB.
by rewrite ler_wpmul2l. 
Qed.

Lemma Mle_wpmul2r (C : 'M[R]_(p, n)) (A B :'M_(q, p)) : 
  0 <=m: C ->  A <=m: B -> A *m C <=m: B *m C. 
Proof.
move=> HCge0 HAleB i j; rewrite !mxE.
apply: (ler_sum )=> // k _.
have/(_ k j):= HCge0; rewrite mxE=> HC.
have/(_ i k) HA:= HAleB.
by rewrite ler_wpmul2r.
Qed.

(* Lemma Mmul_mle_compat_l: *)
(* forall (a: 'M_(n,p)) (u v: 'M_(p, q)), u <=m: v -> Mle 0 a -> a *m u <=m: a *m v. *)
(* move=> a u v Huv Ha i k; rewrite !mxE. *)
(* apply: big_ind2; [by right|apply Rplus_le_compat |]. *)
(* move=> j _; apply Rmult_le_compat_l; last by apply Huv. *)
(* by move : (Ha i j); rewrite mxE. *)
(* Qed. *)

(* Lemma Mmul_mle_compat_r: *)
(* forall (a: 'M_(q,n)) (u v: 'M_(p, q)), u <=m: v -> Mle 0 a -> u *m a <=m: v *m a. *)
(* move=> a u v Huv Ha i k; rewrite !mxE. *)
(* apply: big_ind2; [by right|apply Rplus_le_compat |]. *)
(* move=> j _; apply Rmult_le_compat_r; last by apply Huv. *)
(* by move : (Ha j k); rewrite mxE. *)
(* Qed. *)

(* Lemma Mmul_mle_compat: *)
(* forall (a b: 'M_(q,n)) (u v: 'M_(p, q)),  *)
(*   u <=m: v -> a <=m: b -> Mle 0 a -> Mle 0 u -> u *m a <=m: v *m b. *)
(* move=> a b u v Huv Hab Ha Hu i k; rewrite !mxE. *)
(* apply: big_ind2; [by right|apply Rplus_le_compat |]. *)
(* move=> j _; apply Rmult_le_compat; [| |apply Huv | apply Hab]. *)
(* by move : (Hu i j); rewrite mxE. *)
(* by move : (Ha j k); rewrite mxE. *)
(* Qed. *)

(* (* *)
(* Lemma Mlt_not_mle: *)
(* forall (a b: 'M_(q,n)), a <=m:  b <-> ~ b <m: a. *)
(* Proof. *)
(* move=> a b; split => H. *)
(* *) *)

End matRProp.

Section matRProp2.
Variable R : rcfType.
Variable p q n : nat.

Lemma Mle_wpmul :
forall (a b: 'M[R]_(q,n)) (u v: 'M_(p, q)),
  u <=m: v -> a <=m: b -> Mle 0 a -> Mle 0 u -> u *m a <=m: v *m b.
move=> a b u v Huv Hab Ha Hu.
have Hb := Mle_trans Ha Hab.
apply: Mle_trans (Mle_wpmul2r Hb Huv).
exact: Mle_wpmul2l.
Qed.

Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).

(* Local Notation "A ^f" := (map_mx (fun x => x%:C) A). *)
(*   (map_mx (ComplexField.real_complex_rmorphism realRcfType) A). *)

Lemma Mabs_real_complex (A : 'M[R]_(p,q)) : Mabs A^%:C = (Mabs A)^%:C. 
Proof.
apply/matrixP=> i j; rewrite !mxE normcE sqrtc_sqrtr ?mulcJ_ge0 //.
by rewrite conjc_real -rmorphM -sqrtr_sqr.
Qed.


End matRProp2.

Section normDef.

Variable R : numDomainType.

Structure can_norm : Type := MNorm {
  anorm : forall p q, matrix R p q -> R;
  pos_norm : forall p q (A : 'M_(p, q)), 0 <= anorm A;
  hom_pos_norm : forall p q (A : 'M_(p, q)) (r : R),
    anorm (r *: A) = `|r| * anorm  A;
  trin_ineq_norm : forall p q (A B : 'M_(p, q)), 
    anorm (A + B) <= anorm A + anorm B;
  prod_ineq_norm : forall p q r (A : 'M_(p, q)) (B : matrix _ q r),
    anorm (A *m B) <= anorm A * anorm B;
  canonical_norm : forall p q (A : 'M_(p, q) ) i j, `|A i j| <= anorm A;
  canonical_norm2 : forall p q (A B: 'M_(p, q)), 
    (forall i j, `|A i j| <= `|B i j|) -> anorm A <= anorm B
}.

(*Coercion anorm : can_norm  >-> Funclass.*)

End normDef.

(* Implicit Arguments anorm[p q]. *)

Notation "|| A : N |" := (anorm N A)
  (at level 0, A, N at level 99, format "|| A  :  N |") : ring_scope.


Section theory.

Variable R : numFieldType.

Lemma Mle_eps m (u v : 'cV[R]_m.+1) :
    let F i := ((u i 0) / 2%:R)/ (v i 0) in  
    let eps := min_fT F in
    0 <m: u -> 0 <m: v -> eps *: v <m: u.  
Proof.
move=> F eps Hu Hv i j; rewrite ord1 !mxE.
have HumxE l : 0 < u l 0 by have/(_ l 0):= Hu; rewrite mxE. 
have HvmxE l : 0 < v l 0 by have/(_ l 0):= Hv; rewrite mxE. 
have [k Hk] := ex_min_fT F ord0.
have unitv l : v l 0 \is a GRing.unit.
  by have:= Hv l 0; rewrite mxE lt0r unitfE => /andP [].  
apply: (ler_lt_trans (y:= u i 0 / 2%:R)).
  rewrite /eps Hk /F -[u i _ / _]mulr1 -[X in _ <= _ * X](mulVr (unitv i)).
  rewrite mulrA ler_pmul ?mulr_ge0 ?invr_ge0 ?[0 <= _]ltrW ?ltr0n //.
  rewrite /F in Hk; rewrite -Hk  min_fT_le // => l.
  by rewrite gtr0_real // ?mulr_gt0 // invr_gt0 // ltr0n.
by rewrite mulrC gtr_pmull // invr_lt1 ?ltr0n ?ltr1n // unitfE pnatr_eq0.
Qed.

End theory.


Section Defnorm.

Variable R : numDomainType.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.


Definition norm8 (m n : nat) (A : 'M[R]_(m,n)) :=
   \maxr_i \sum_j `|A i j|.

Lemma norm8_pos (m n : nat) (A : 'M[R]_(m,n)) : 0 <= norm8 A. 
Proof. by apply: bigmax_pos=> i; apply: sumr_ge0. Qed.

Lemma norm8_hom p q (A : 'M_(p, q)) (r : R) : norm8 (r *: A) = `|r| * norm8 A.
Proof.
rewrite /norm8 (eq_bigr (fun i => `|r| * \sum_j `|A i j|)) ?bigmax_distrl //.
move=> i _; rewrite big_distrr /=.
by apply: eq_bigr=> j _; rewrite mxE normrM.
Qed.

Lemma norm8_tri p q (A B : 'M_(p, q)) : norm8 (A + B) <= norm8 A + norm8 B.
Proof.
have Hpos i: 0 <= \sum_j `|(A + B) i j| by apply: sumr_ge0.
have Hle i: \sum_j `|(A + B) i j| <= \sum_j `|A i j| + \sum_j `|B i j|.
  by rewrite -big_split /= ler_sum // => j _; rewrite mxE ler_norm_add.
apply: (ler_trans (bigmax_ler Hpos Hle)).
by apply: bigmax_ler_sum=> i; apply: sumr_ge0.
Qed.

Lemma norm8M p q r (A : 'M_(p, q)) (B : matrix _ q r) :
  norm8 (A *m B) <= norm8 A * norm8 B.
Proof.
have HABge0 i : 0 <= \sum_j `|(A *m B) i j| by apply: sumr_ge0. 
have HAge0 i : 0 <= \sum_j `|A i j| by apply: sumr_ge0.
have HBge0 i : 0 <= \sum_j `|B i j| by apply: sumr_ge0.
have H: norm8 (A *m B) <= \maxr_i \sum_k `|A i k| * \sum_j `|B k j|.
  apply: (bigmax_ler HABge0)=> i.
  rewrite (eq_bigr (fun k => \sum_ j `|A i k * B k j|))=>[|j _]. 
    by rewrite exchange_big /= ler_sum // => j _; rewrite mxE ler_norm_sum //.
  by rewrite big_distrr; apply: eq_bigr=> k _; rewrite normrM.
apply: (ler_trans H).
apply: ler_trans (bigmax_le_prod2 HAge0 HBge0). 
apply: bigmax_ler=> i; first by apply: sumr_ge0=> j _; rewrite mulr_ge0. 
rewrite -bigmax_distrl // big_distrl /= ler_sum // => j _. 
by rewrite ler_pmul // (bigmax_ger HBge0) // mem_index_enum.
Qed.

Lemma norm8_can1 p q (A : 'M_(p, q) ) i j : `|A i j| <= norm8 A.
Proof.
have Hpos k : 0 <= \sum_j `|A k j| by apply: sumr_ge0.
apply: ler_trans (bigmax_ger Hpos (mem_index_enum i)).
by rewrite (bigD1 j) //= ler_addl sumr_ge0.
Qed.

Lemma norm8_can2 p q (A B: 'M_(p, q)) : 
  (forall i j, `|A i j| <= `|B i j|) -> norm8 A <= norm8 B.
Proof.
move=> Hle.
have Hpos k : 0 <= \sum_j `|A k j| by apply: sumr_ge0.
by apply: (bigmax_ler Hpos)=> i; apply: ler_sum.
Qed.


Canonical Structure can_norm8 := MNorm norm8_pos norm8_hom norm8_tri 
  norm8M norm8_can1 norm8_can2. 
 
End Defnorm.

Section normProp.

Variable R : numDomainType.
Variable p q: nat.

Variable N : can_norm R.

Open Scope ring_scope.


Lemma cnorm0 : || (0: 'M_(p, q)) : N | = 0.
Proof.
set z := (0 : 'M_(p,q)).
have: || z : N | = || (2%:R *: z) : N | by rewrite scaler0.
by rewrite hom_pos_norm normr_nat mulr_natl -[X in X = _]addr0=> /addrI <-.
Qed.

Lemma cnorm_eq0 (A : 'M[R]_(p,q)) : (|| A : N | == 0) = (A == 0).
Proof.
apply/idP/idP=> [/eqP Heq0|/eqP ->].
  apply/eqP/matrixP=> i j; rewrite mxE.
  by apply/eqP; rewrite -normr_eq0 eqr_le -{1}Heq0 canonical_norm /=.
by rewrite cnorm0.
Qed.
 
Lemma cnorm_lt0  (A: 'M_(p, q)) : (0 < ||A : N |) = (A != 0).
Proof. by rewrite ltr_neqAle pos_norm -cnorm_eq0 eq_sym andbT. Qed.

Lemma cnormN  (A: 'M_(p, q)) : || -A : N| = || A : N|.
Proof. by rewrite -[A]scale1mx -scaleNr !hom_pos_norm normrN. Qed.


(* Lemma RabsM_ineq: *)
(* forall (A: 'M_(p, q)) eps, 0 <= eps -> (forall i j, Rabs (A i j) < eps) -> *)
(*  || A : mynorm | <= eps * || (one_M:'M_(p, q)) : mynorm |. *)
(* Proof. *)
(* move=> A e He H. *)
(* have H1: forall i j, Rabs (A i j) < (e *: one_M) i j. *)
(* move=> i j; rewrite !mxE; rewrite -> Rmult_1_r; apply H. *)
(* have Hepos: forall (i:'I_p) (j:'I_q),  *)
(*   Rabs ((e *: one_M) i j) = (e *: one_M) i j. *)
(* move=> i j;  rewrite !mxE;  rewrite -> Rabs_right; first by done. *)
(* apply Rle_ge; apply Rle_trans with (Rabs (A i j)) ; [apply Rabs_pos|  *)
(*   left; rewrite -> Rmult_1_r; apply H] . *)
(* apply Rle_trans with (|| (e *: (@one_M p q)) : mynorm|). *)
(* apply canonical_norm2. *)
(* move=> i j; rewrite Hepos; left; apply H1. *)
(* rewrite -{2}(Rabs_right e). *)
(* right; apply hom_pos_norm. *)
(* by apply Rle_ge. *)
(* Qed. *)

(* Import GRing.Theory. *)

(* Lemma tri_minus_norm: *)
(* forall (a b: 'M_(p, q)), || a : mynorm | - || b: mynorm | <= || (a - b)%Ri : mynorm |. *)
(* Proof. *)
(* move=> a b. *)
(* apply (Rplus_le_reg_l (||b :mynorm |)). *)
(* ring_simplify. *)
(* pattern a at 1; replace a with (b +( a - b))%Ri. *)
(* apply: trin_ineq_norm. *)
(* rewrite addrCA addrA. *)
(* by rewrite addrK. *)
(* Qed. *)

(* Lemma eq_norm_min: *)
(* forall (a b: 'M_(p, q)), || (a - b)%Ri : mynorm | = || (b - a)%Ri : mynorm |. *)
(* Proof. *)
(* move=> a b. *)
(* replace (a - b)%Ri with (-1 *: (b - a)%Ri). *)
(* by rewrite hom_pos_norm Rabs_Ropp Rabs_R1; toR; rewrite Rmult_1_l. *)
(* toR; apply/matrixP=> i j; rewrite !mxE ; toR; ring. *)
(* Qed. *)

(* Lemma tri_min2: *)
(* forall (a b: 'M_(p, q)), Rabs ( ||a : mynorm| - ||b : mynorm|) <= || (a - b)%Ri : mynorm |. *)
(* Proof. *)
(* move=> a b. *)
(* rewrite /Rabs; case (Rcase_abs (|| a : mynorm | - || b : mynorm |) )=> H; *)
(* [|apply tri_minus_norm] . *)
(* rewrite Ropp_minus_distr eq_norm_min; apply tri_minus_norm. *)
(* Qed. *)

(* Lemma mat_norm_sum: forall (A : nat -> 'M_(p, q)) n, *)
(*    || (\sum_(i < n.+1) A i) : mynorm | <=  \sum_(i < n.+1)  || A i : mynorm |. *)
(* Proof. *)
(* move=> A n. *)
(* apply: big_morph_rel. *)
(* by move=> x; right. *)
(* by move=> *; apply Rplus_le_compat. *)
(* by move=>b *; apply Rle_trans with b. *)
(* apply mynorm0. *)
(* by move=> a v; apply: trin_ineq_norm. *)
(* Qed. *)


 End normProp.



(* Section limitMatDef. *)

(* Variable p q: nat. *)



(* Definition limit_m (An: nat -> 'M_(p, q)) (La: 'M_(p, q)):= *)
(* forall i j, Un_cv (fun n => An n i j) (La i j). *)

(* Definition cv_mat_ser (Ak:nat->'M_(p, q)) (A:'M_(p, q)) :=  *)
(* forall i j, Un_cv (fun N => sum_f_R0 (fun n => Ak n i j) N) (A i j). *)

(* Definition cv_mat_ser2 (Ak:nat->'M_(p, q)) (A: 'M_(p, q)) := *)
(* limit_m (fun N => \sum_(i<N.+1) Ak i) A. *)



(* End limitMatDef. *)


(* Section limitMatProp. *)

(* Variable p q r: nat. *)
(* Variable mynorm: can_norm. *)

(* Lemma lim_conv_low: *)
(* forall (un:nat->R) (a l:R), *)
(* (forall n, a<=un n)-> *)
(* Un_cv un l -> *)
(* a<=l. *)
(* Proof. *)
(* intros. *)
(* cut (~(l<a)). *)
(* intros. *)
(* apply Rnot_gt_le; auto. *)
(* unfold not; intros. *)
(* unfold Un_cv in H0. *)
(* assert (a-l>0). *)
(* toR; fourier. *)
(* assert (Hcon: exists N, forall n, (n>=N)%coq_nat ->R_dist (un n) l <a-l). *)
(* apply (@H0 (a-l) H2). *)
(* elim Hcon; intros N Hlim. *)
(* assert (Hdist:R_dist (un N) l<a-l). *)
(* apply (Hlim N). *)
(* omega. *)
(* unfold R_dist in Hdist. *)
(* assert (Ha: a<=un N). *)
(* apply (H N). *)
(* rewrite Rabs_right in Hdist; [|fourier]. *)
(* assert (un N<a); fourier. *)
(* Qed. *)

(* Lemma lim_seq_monot: *)
(* forall (xn yn:nat->R) (x y:R),  *)
(* (forall n, xn n<= yn n)-> *)
(* (Un_cv xn x)-> *)
(* (Un_cv yn y)-> *)
(* x<=y. *)
(* intros. *)
(* assert (0<=y-x). *)
(*   assert (forall n, 0<=yn n- xn n). *)
(*     intros. *)
(*     elim (H n);intros; toR; fourier. *)
(*   assert (Un_cv (fun i:nat => yn i - xn i) (y - x)). *)
(*     toR; apply: CV_minus; assumption. *)
(*   apply lim_conv_low with (fun  i:nat => yn i - xn i); *)
(* assumption. *)
(* rewrite /GRing.zero //= in H2; fourier. *)
(* Qed. *)


(* Lemma el_eq_seq_eq: *)
(* forall xn yn x , (forall n, xn n = yn n )-> Un_cv xn x-> Un_cv yn x. *)
(* Proof. *)
(* by move=> xn yn x Heq Hcv eps He; elim (Hcv eps He) => N HN; *)
(* exists N=> n Hn; rewrite -Heq; apply HN. *)
(* Qed. *)

(* Lemma el_eq_mat_seq_cv: *)
(* forall xn yn (x: 'M_(p, q)), (forall n, xn n = yn n) -> limit_m xn x -> limit_m yn x. *)
(* Proof. *)
(* move=> xn yn x Hxy Hx i j. *)
(* apply el_eq_seq_eq with (fun n => xn n i j). *)
(* by move=> n; rewrite Hxy. *)
(* apply Hx. *)
(* Qed. *)

(* Lemma echiv_mat_ser: *)
(* forall Ak (A: 'M_(p, q)), cv_mat_ser Ak A <-> cv_mat_ser2 Ak A. *)
(* Proof. *)
(* move=> Ak A; split=> H i j; have H1:= H i j. *)
(* have H2: forall N,  sum_f_R0 (fun k : nat => Ak k i j) N =  *)
(* (\sum_(i0 < N.+1) Ak i0) i j. *)
(* move=> N; rewrite echiv_sums. *)
(* elim: index_enum. *)
(* by rewrite !big_nil !mxE. *)
(* by move => k s Ih; rewrite !big_cons Ih !mxE. *)
(* apply (@el_eq_seq_eq _ _ _ H2 H1). *)
(* have H2: forall N,  (\sum_(i0 < N.+1) Ak i0) i j *)
(* =sum_f_R0 (fun k : nat => Ak k i j) N . *)
(* move=> N; rewrite echiv_sums. *)
(* elim: index_enum. *)
(* by rewrite !big_nil !mxE. *)
(* by move => k s Ih; rewrite !big_cons !mxE Ih. *)
(* apply (@el_eq_seq_eq _ _ _ H2 H1). *)
(* Qed. *)

(* Lemma limit_m_mul: *)
(* forall (A: nat -> 'M_(p, q)) (R: 'M_(r, p)) (S: 'M_(p, q)), limit_m A S ->  *)
(*   limit_m (fun n => R *m (A n))%Ri (R *m S)%Ri. *)
(* Proof. *)
(* move=> A R S Hl i j eps He. *)
(* red in Hl; red in Hl. *)
(* have Hl1: forall eps, eps > R0 -> forall i j, *)
(*      exists N : nat, *)
(*        forall n : nat, (n >= N)%coq_nat -> R_dist (A n i j) (S i j) < eps. *)
(*   by move=>*; apply Hl. *)
(* have H0 :  R0 <= \big[Rmax/R0]_j Rabs (R i j) . *)
(*   elim: index_enum; [by rewrite big_nil; right|]. *)
(*   move=> t s IH; rewrite big_cons. *)
(*   by apply Rmax_ge; [apply Rabs_pos|]. *)
(* case H0 => H. *)
(*   have Hrw : \big[Rmax/0]_j Rabs (R i j) != 0. *)
(*     by apply/eqP; apply: Rgt_not_eq. *)
(*   have Hp: (0<p)%coq_nat. *)
(*     have hbig0: forall   f , p = 0%Dn ->  *)
(*   \big[Rmax/0]_(i<p) f i =0. *)
(*       move=> f Hp; move: f; rewrite Hp; apply big_ord0. *)
(*     have HR := hbig0 (fun j => Rabs (R i j)). *)
(*     case Cp: (p == O). *)
(*       move/eqP: Cp => Cp. *)
(*       have Hcp:= HR Cp. *)
(*       rewrite Hcp in H. *)
(*       elim (Rlt_irrefl _  H). *)
(*     move/eqP: Cp => Cp;  omega. *)
(*   have H01: 0 < \big[Rmax/R0]_i \big[Rmax/0]_j Rabs (R i j). *)
(*     apply Rlt_le_trans with (\big[Rmax/0]_j Rabs (R i j)); [done|]. *)
(*     apply (@max_max r (fun i0 =>\big[Rmax/0]_j0 Rabs (R i0 j0))) => i0. *)
(*     elim: index_enum; [by rewrite big_nil; right|]. *)
(*     move=> t s IH; rewrite big_cons. *)
(*     by apply Rmax_ge; [apply Rabs_pos|]. *)
(*   have H1: (eps / (\big[Rmax/0]_i \big[Rmax/R0]_i0 Rabs (R i i0))) /INR p > 0. *)
(*     rewrite /Rdiv; apply Rmult_gt_0_compat. *)
(*        by apply Rmult_gt_0_compat; [| apply: Rinv_0_lt_compat] . *)
(*     apply: Rinv_0_lt_compat; apply: lt_INR_0. *)
(*     done. *)
(*   have Hl2:=  *)
(*   (Hl1 ((eps/(\big[Rmax/R0]_i \big[Rmax/R0]_i0 (Rabs (R i i0))))/INR p)%Re H1). *)
(*   have Hc1:  forall i , exists c1: 'I_q -> nat, *)
(*         forall j n, *)
(*         (n >= (c1 j))%coq_nat -> *)
(*         R_dist (A n i j) (S i j) < *)
(*         eps / \big[Rmax/0]_i0 \big[Rmax/0]_i1 Rabs (R i0 i1) / INR p. *)
(*     move=> i0; apply (choice _ (Hl2 i0)). *)
(*   case (choice _ Hc1) => ch Hch. *)
(*   exists ( \max_(i<p)  \max_(j<q) (ch i j)) => n Hn. *)
(*   rewrite !mxE /R_dist. *)
(*   apply Rle_lt_trans with (Rabs (\sum_j0 (R i j0 * A n j0 j)%Ri - \sum_j0 (R i j0 * S j0 j)%Ri)%Ri). *)
(*     right; f_equal. *)
(*   rewrite -GRing.sumrB. *)
(*   apply Rle_lt_trans with (Rabs (\sum_i0 (R i i0 * (A n i0 j -  S i0 j)%Ri))). *)
(*     right; f_equal; apply eq_bigr => i0 _; toR; ring. *)
(*   apply Rle_lt_trans with  (\sum_i0 Rabs (R i i0 * (A n i0 j - S i0 j)%Ri)). *)
(*     rewrite //=; apply abs_sum_le_sum_abs with (f:= (fun i0 => R i i0 * (A n i0 j - S i0 j))) . *)
(*   apply Rle_lt_trans with (\sum_i0 (\big[Rmax/0]_i0 Rabs (R i i0) * Rabs (A n i0 j - S i0 j)%Ri)). *)
(*     apply (sum_le_f_le) => i0. *)
(*     rewrite Rabs_mult. *)
(*     apply Rmult_le_compat. *)
(*           apply Rabs_pos . *)
(*         apply Rabs_pos. *)
(*       apply (@max_max p (fun j=> Rabs (R i j)))=> i1; apply Rabs_pos. *)
(*     by right. *)
(*   rewrite -big_distrr //= GRing.mulrC. *)
(*   replace eps with (  (eps / \big[Rmax/0]_i0 Rabs (R i i0)) * \big[Rmax/0]_(i0 < p) Rabs (R i i0) )%R. *)
(*     apply Rmult_lt_compat_r . *)
(*       done. *)
(*     apply: (ineq_sum_el  ). *)
(*         by move/leP: Hp. *)
(*       left; apply: Rmult_gt_0_compat=> //.  *)
(*       by rewrite /GRing.inv /= /Rinvx Hrw; apply: Rinv_0_lt_compat. *)
(*     move=> x. *)
(*     apply Rlt_le_trans with (eps / \big[Rmax/0]_i0 \big[Rmax/0]_i1 Rabs (R i0 i1) / INR p)%Re. *)
(*       rewrite /R_dist in Hch. *)
(*       apply Hch. *)
(*       apply: (le_trans _  (\max_(i < p) \max_(j < q) ch i j)%nat _);[|apply Hn]. *)
(*       apply le_trans with (\max_(j0 < q) ch x j0 ). *)
(*         apply/leP; apply (@leq_bigmax (ordinal_finType q) (fun j0 => ch x j0) j ). *)
(*       apply/leP; apply (@leq_bigmax _ (fun i0 =>\max_(j0 < q) ch i0 j0) ) . *)
(*     apply: Rmult_le_compat_r. *)
(*       by left; apply: Rinv_0_lt_compat; apply lt_INR_0. *)
(*     apply: Rmult_le_compat_l; [by left|]. *)
(*     toR; rewrite /Rinvx Hrw.     *)
(*     apply: Rle_Rinv; try done. *)
(*     apply (@max_max r (fun i0 => \big[Rmax/0]_i1 Rabs (R i0 i1))) => i0. *)
(*     have Hp1: (0<p)%Dn. *)
(*       by move/leP: Hp. *)
(*     case (@max_exists p Hp1 (fun j =>  Rabs (R i0 j)))=> i1; *)
(*  [apply Rabs_pos|  *)
(* move=> Hx; rewrite Hx; apply Rabs_pos]. *)
(*   toR; rewrite /Rinvx Hrw /Rdiv  Rmult_assoc Rinv_l; [ring|]. *)
(*   by apply Rgt_not_eq. *)
(* exists O => n Hn. *)
(* rewrite !mxE /R_dist. *)
(* apply Rle_lt_trans with (Rabs (\sum_j0 (R i j0 * A n j0 j)%Ri - \sum_j0 (R i j0 * S j0 j)%Ri)%Ri). *)
(*   right; f_equal. *)
(* rewrite -GRing.sumrB. *)
(* apply Rle_lt_trans with (Rabs (\sum_i0 (R i i0 * (A n i0 j -  S i0 j)%Ri))). *)
(*    right; f_equal; apply eq_bigr => i0 _; toR; ring. *)
(* apply Rle_lt_trans with  (\sum_i0 Rabs (R i i0 * (A n i0 j - S i0 j)%Ri)). *)
(*   rewrite //=; apply abs_sum_le_sum_abs with (f:= (fun i0 => R i i0 * (A n i0 j - S i0 j))) . *)
(* apply Rle_lt_trans with (\sum_i0 (\big[Rmax/0]_i0 Rabs (R i i0) * Rabs (A n i0 j - S i0 j)%Ri)). *)
(*   apply (sum_le_f_le ) => i0. *)
(*   rewrite Rabs_mult. *)
(*   apply Rmult_le_compat. *)
(*         apply Rabs_pos . *)
(*       apply Rabs_pos. *)
(*     apply (@max_max p (fun j=> Rabs (R i j)))=> i1; apply Rabs_pos. *)
(*   by right. *)
(* rewrite -big_distrr //= GRing.mulrC. *)
(* by rewrite -H GRing.mulr0. *)
(* Qed. *)



(* Lemma limit_m_add: *)
(* forall A B (x y: 'M_(p, q)), limit_m A x -> limit_m B y -> limit_m (fun n => A n + B n) (x + y). *)
(* Proof. *)
(* move=>  a b x y Ha Hb i j. *)
(* rewrite !mxE. *)
(* apply el_eq_seq_eq with (fun n => a n i j + b n i j); [by move=> n; rewrite mxE |]. *)
(* apply CV_plus;[apply Ha|apply Hb]. *)
(* Qed. *)

(* Lemma limit_m_opp_0: *)
(* forall (a: nat -> 'M_(p,q)) , limit_m a 0 -> limit_m (fun n=> (- a n)%Ri) 0. *)
(* Proof. *)
(* move=> a Hcv. *)
(* move=> i j eps He; case (Hcv  i j eps He) => N HN; exists N => n Hn. *)
(* have HN1 := HN n Hn. *)
(* rewrite !mxE /R_dist !Rminus_0_r in HN1 |- *. *)
(* by toR; rewrite Rabs_Ropp. *)
(* Qed. *)

(* Lemma limit_m_Sn: *)
(* forall a (x: 'M_(p, q)), limit_m a x -> limit_m (fun n => a (n.+1)) x. *)
(* Proof. *)
(* move=> a x Hcv i j eps He ; case (Hcv i j eps He) => N HN;  *)
(* exists (N.+1) => n Hn; apply HN; omega. *)
(* Qed. *)

(* Lemma limit_m_uniq: *)
(* forall a (x y: 'M_(p, q)), limit_m a x -> limit_m a y -> x = y. *)
(* Proof. *)
(* move=> a x y  Hx Hy; apply/ matrixP => i j. *)
(* red in Hx. *)
(* by apply UL_sequence with (fun n => a n i j) . *)
(* Qed.  *)




(* Lemma th1_246: *)
(* forall Ak,  (exists S:'M_(p,q), cv_mat_ser2 Ak S) -> limit_m Ak 0. *)
(* Proof. *)
(* move=> Ak [S Hcv]. *)
(* move: Hcv; rewrite -echiv_mat_ser => Hcv. *)
(* move=> i j eps He. *)
(* have Hc2:= Hcv i j. *)
(* have Hcau: Cauchy_crit (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N). *)
(* by apply CV_Cauchy; exists (S i j). *)
(* elim (Hcau eps He) => N HN. *)
(* exists (N.+1)%nat => n Hn. *)
(* have H1: (minus n  1 >=N)%coq_nat; [omega|]. *)
(* have H2: ((minus n 1).+1 >= N)%coq_nat;[omega|]. *)
(* have HN2:= HN _ _ H1 H2. *)
(* rewrite /R_dist //=  Rabs_minus_sym Rplus_comm  *)
(* /Rminus Rplus_assoc Rplus_opp_r Rplus_0_r in HN2. *)
(* rewrite /R_dist mxE Rminus_0_r. *)
(* replace n with ((minus n 1).+1);[done|omega]. *)
(* Qed. *)



(* Lemma lem_cv1: *)
(* forall (Ak: nat -> 'M_(p, q)) (A: 'M_(p, q)),  *)
(*   limit_m Ak A <-> Un_cv (fun k => || (A - (Ak k)) : mynorm |) R0. *)
(* Proof. *)
(* move=> Ak A; split; move=> H. *)
(* red in H; red in H. *)
(* have H1: forall eps, eps>0->forall i j,  exists N : nat, *)
(*       forall n : nat, (n >= N)%coq_nat -> R_dist (Ak n i j) (A i j) < eps. *)
(* by move=>*; apply H. *)

(* have H2: forall eps, eps>0 -> forall i , exists c2: 'I_q->nat,  *)
(* forall j,   forall n : nat, (n >= (c2 j))%coq_nat -> R_dist (Ak n i j) (A i j) < eps. *)
(* move=> eps He i; *)
(*  apply choice with (A:= 'I_q) (B:= nat)  *)
(* (R:=fun j N  =>  forall n : nat, (n >= N)%coq_nat -> R_dist (Ak n i j) (A i j) < eps). *)
(* by move=> *; apply H1. *)
(* have H3: forall eps, eps>0-> exists c1: 'I_p->'I_q->nat, *)
(* forall i j n,  (n >= (c1 i j))%coq_nat -> R_dist (Ak n i j) (A i j) < eps. *)
(* move=> eps He ; *)
(* apply (choice _ (H2 eps He)). *)
(* move=> eps He /=. *)
(* elim (Req_dec (||@one_M p q :mynorm|) R0) => Hdec. *)
(* elim (H3 eps He)=> ch Hch. *)
(* exists (\max_(i<p) \max_(j<q) ch i j)=> n Hn. *)
(*  apply Rle_lt_trans with R0. *)
(* rewrite /R_dist Rminus_0_r Rabs_right. *)
(* replace R0 with (eps * ||@one_M p q:mynorm|). *)
(* apply RabsM_ineq; [left; apply: He|].  *)
(* move=> i j; rewrite !mxE Rabs_minus_sym. *)
(* apply:Hch. *)
(* rewrite /ge; apply le_trans with  (\max_(i < p) \max_(j < q) ch i j)%nat. *)
(* apply le_trans with  (\max_(j < q) ch i j)%nat. *)
(* apply/leP. *)
(* apply: leq_bigmax. *)
(* apply/leP. *)
(* apply (@leq_bigmax _ (fun i => \max_(j0 < q) ch i j0 ) i). *)
(*  omega. *)
(* rewrite Hdec. *)
(* by rewrite -> Rmult_0_r. *)
(* apply Rle_ge;  apply: pos_norm. *)
(* done. *)

(* have Hen: (1/2* eps / ||@one_M p q:mynorm|)%Ri > 0. *)
(* apply: Rmult_gt_0_compat. *)
(* apply Rmult_gt_0_compat; rewrite /GRing.mul /GRing.inv //= /Rinvx . *)
(*   have t2: 2!= 0. *)
(*     apply /eqP; rewrite /GRing.zero //=; apply Rgt_not_eq; fourier. *)
(* rewrite t2; rewrite /GRing.mul /GRing.inv /GRing.one /GRing.zero //=; fourier. *)
(* elim (pos_norm mynorm  (@one_M p q)) => Hen2. *)
(* rewrite /GRing.inv /=. *)
(* case cn: (||@one_M p q : mynorm| == 0); *)
(* rewrite /Rinvx cn //=. *)
(* by apply Rinv_0_lt_compat. *)
(* by elim Hdec. *)
(* elim (H3 (1/2*eps/(||one_M : mynorm |)) Hen)=> ch Hch. *)
(* exists (\max_(i<p) \max_(j< q ) ch i j)=> n Hn. *)
(* rewrite /R_dist Rminus_0_r Rabs_right;[|apply Rle_ge; apply:pos_norm]. *)
(* apply Rle_lt_trans with ((1/2*eps/ || (@one_M p q) : mynorm|) *  || (@one_M p q) : mynorm|). *)
(* have Hrd: forall i j, Rabs (Ak n i j - A i j) < (1 / 2 * eps / || @one_M p q : mynorm |)%Ri. *)
(* move => i j ; apply:Hch. *)
(* rewrite /ge; apply le_trans with  (\max_(i < p) \max_(j < q) ch i j)%nat. *)
(* apply le_trans with  (\max_(j < q) ch i j)%nat. *)
(* apply/leP; apply: leq_bigmax.  *)
(* apply/leP; apply: (@leq_bigmax _ (fun i => \max_(j0 < q) ch i j0 ) i). *)
(* omega. *)
(* apply RabsM_ineq. *)
(* by left; apply Rgt_lt. *)
(* by move=> i j; rewrite !mxE Rabs_minus_sym. *)
(* toR; rewrite /Rinvx. *)
(*   have t2: 2!= 0. *)
(*     apply /eqP; rewrite /GRing.zero //=; apply Rgt_not_eq; fourier. *)
(*   have tn: || @one_M p q : mynorm | != 0. *)
(*   by apply /eqP. *)
(* rewrite t2 tn /Rdiv Rmult_assoc Rinv_l; last by done. *)
(* rewrite Rmult_1_r ; fourier. *)

(* move=> i j eps He. *)
(* elim (H eps He) => N HN. *)
(* exists N => n Hn. *)
(* rewrite /R_dist in HN |-*. *)
(* apply Rle_lt_trans with (|| (A - Ak n) : mynorm |). *)
(* rewrite Rabs_minus_sym. *)
(* have Heq: (A i j - Ak n i j) = ((A - Ak n) i j); *)
(* [by rewrite !mxE|]. *)
(* move: Heq;  toR =>  Heq; rewrite /Rminus. *)
(* rewrite Heq. *)
(* apply canonical_norm. *)
(* have H1:= HN n Hn. *)
(* rewrite Rminus_0_r in H1. *)
(* rewrite -(Rabs_right ( || (A - Ak n) : mynorm |));[done|apply Rle_ge; apply:pos_norm]. *)
(* Qed. *)


(* Lemma abs_cv_cv: *)
(* forall xn xs, Un_cv (fun n => sum_f_R0 (fun k => Rabs (xn k)) n) xs -> *)
(* exists l, Un_cv (fun n=> sum_f_R0 xn n) l. *)
(* Proof. *)
(* move=> xn xs Hcva. *)
(* have H1: Cauchy_crit (fun n : nat => sum_f_R0 (fun k : nat => Rabs (xn k)) n). *)
(* apply CV_Cauchy; by exists xs. *)
(* have H2: {l| Un_cv (fun n : nat => sum_f_R0 xn n) l}. *)
(* apply R_complete => eps He. *)
(* elim (H1 eps He) => N HN. *)
(* exists N => n m Hn Hm; have H2:= HN n m Hn Hm. *)
(* rewrite /R_dist in H2 |-*. *)
(* rewrite !echiv_sums in H2 |-*. *)
(* elim (le_lt_dec m n) => Hdec. *)
(* elim (le_lt_eq_dec _ _ Hdec) => Hd2. *)
(* move/ltP:Hd2=>Hd2. *)
(* have H10:= (@sum_dif xn m n Hd2 ). *)
(* (* rewrite /GRing.opp /GRing.add /= in H10. *) *)
(* rewrite  H10. *)
(* have H3:= (@big_addn _ _ _ (O%nat) (n.+1)%nat (m.+1) (fun _ => true) xn). *)
(* rewrite H3 big_mkord. *)
(* apply Rle_lt_trans with (\sum_(i < n.+1 - m.+1) Rabs (xn (i + m.+1)%Dn)). *)
(* apply (abs_sum_le_sum_abs). *)
(* (* (fun i =>xn (i + m.+1)%Dn) (n.+1 - m.+1)) .*) *)
(* have H4:= (@big_addn _ _ _ (O%nat) (n.+1)%nat (m.+1)  *)
(*    (fun _ => true) (fun i => Rabs (xn i))). *)
(* have h11:= (@sum_dif ( fun i => Rabs (xn i)) _ _ Hd2) . *)
(* rewrite h11  H4 big_mkord in H2. *)
(* apply Rle_lt_trans with (Rabs (\sum_(i < n.+1 - m.+1) Rabs (xn (i + m.+1)%Dn)) ). *)
(*  apply RRle_abs . *)
(* done. *)
(* by rewrite Hd2 /Rminus Rplus_opp_r Rabs_R0. *)
(* rewrite Rabs_minus_sym . *)
(* move/ltP:Hdec => Hdec. *)
(* have h12:= (@sum_dif _ _ _ Hdec). *)
(* rewrite h12. *)
(* have H3:= (@big_addn _ _ _ (O%nat) (m.+1)%nat (n.+1) (fun _ => true) xn). *)
(* rewrite H3 big_mkord. *)
(* apply Rle_lt_trans with (\sum_(i < m.+1 - n.+1) Rabs (xn (i + n.+1)%Dn)). *)
(* apply (abs_sum_le_sum_abs). *)
(* have H4:= (@big_addn _ _ _ (O%nat) (m.+1)%nat (n.+1)  *)
(*    (fun _ => true) (fun i => Rabs (xn i))). *)
(* rewrite Rabs_minus_sym in H2. *)
(* have H5:=  (sum_dif ( fun i => Rabs (xn i)) Hdec). *)
(* rewrite H5 in H2. *)
(* rewrite H4 big_mkord in H2. *)
(* apply Rle_lt_trans with (Rabs (\sum_(i < m.+1 - n.+1) Rabs (xn (i + n.+1)%Dn)) ). *)
(*  apply RRle_abs . *)
(* done. *)
(* by elim H2 => l Hl; exists l. *)
(* Qed. *)


(* Lemma th2_246: *)
(* forall (Ak: nat -> 'M_(p, q)),  *)
(*   (exists S,  cv_mat_ser (fun n=> Mabs (Ak n)) S) -> exists S', *)
(* cv_mat_ser Ak S'. *)
(* Proof. *)
(* move=> Ak  [S Hcv]. *)
(*  red in Hcv. *)
(* rewrite /cv_mat_ser . *)
(* have H1:  forall i j, exists l,  Un_cv  *)
(*   (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) l. *)
(* move=> i j. *)
(* apply abs_cv_cv with (S i j). *)
(* apply el_eq_seq_eq with (fun N : nat => sum_f_R0 (fun n : nat => Mabs (Ak n) i j) N). *)
(* rewrite /Mabs. *)
(* induction n => //=; [by rewrite !mxE|by rewrite IHn !mxE]. *)
(* apply Hcv. *)
(* have Hc: forall i, exists c1:'I_q->R, forall j, *)
(*  Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) (c1 j). *)
(* by move=> i; apply choice with (R:= fun j l=>  *)
(*  Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) l) . *)
(* have Hc2: exists c2:'I_p->'I_q->R, forall i j,  *)
(* Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) (c2 i j). *)
(* apply (choice _ (Hc)). *)
(* elim Hc2 => ch Hch. *)
(* by exists (\matrix_(i, j) ch i j) => i j; rewrite !mxE. *)
(* Qed. *)

(* Lemma thm2_246: *)
(* forall (Ak: nat -> 'M_(p, q)) , (exists S, cv_mat_ser2 (fun n=> Mabs (Ak n)) S) -> exists S', *)
(* cv_mat_ser2 Ak S'. *)
(* Proof. *)
(* move=> Ak  [S Hs]. *)
(* have Hs1: exists S : matrix real_field p q, *)
(*        cv_mat_ser (fun n : nat => Mabs (Ak n)) S. *)
(* by exists S; move: Hs; rewrite -echiv_mat_ser => Hs. *)
(* elim (th2_246  Hs1) => S' HS'. *)
(* by exists S'; rewrite -echiv_mat_ser. *)
(* Qed. *)

(* Lemma th3_247: *)
(* forall Ak S, Un_cv  *)
(*   (fun N:nat => sum_f_R0 (fun n => || (Ak n: 'M_(p, q)) : mynorm |)  N) S ->  *)
(* exists S', cv_mat_ser (fun n=> Mabs (Ak n)) S'. *)
(* Proof. *)
(* move=> Ak S Hcv. *)
(* rewrite /cv_mat_ser. *)
(* have H: forall i j, {l |   *)
(*  Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Mabs (Ak n) i j) N) l}. *)
(* move=> i j. *)
(* apply Rseries_CV_comp with  *)
(*  (fun n : nat => || Ak n : mynorm |). *)
(* move=> n; rewrite  !mxE; split; [apply Rabs_pos|apply canonical_norm]. *)
(* by exists S. *)
(* have H1: forall i j, exists l, *)
(*  Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Mabs (Ak n) i j) N) l. *)
(* by move=> i j; elim (H i j) => l  Hl; exists l. *)
(* have H2: forall i, exists S1, *)
(*   forall j : 'I_q, *)
(*   Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Mabs (Ak n) i j) N) *)
(*     (S1  j). *)
(* move=> i; apply (choice _ (H1 i)). *)
(* have H3: exists S1, *)
(*   forall i j, *)
(*   Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Mabs (Ak n) i j) N) *)
(*     (S1 i j). *)
(* apply (choice _ H2). *)
(* by elim H3 => S1  Hs1;  exists (\matrix_(i, j) S1 i j)=> i j; rewrite !mxE. *)
(* Qed. *)


(* End limitMatProp. *)


(* Section limitsqrMat. *)

(* Variable p': nat. *)
(* Notation p := p'.+1. *)

(* Lemma Hp: (0 < p)%Dn. *)
(* by done. Qed. *)

(* Variable mynorm: can_norm. *)

(* Hypothesis Hyp_norm1: *)
(* || (1 : 'M_p) : mynorm | = 1.  *)

(* Lemma matr_prod_ineq: *)
(* forall (X: 'M_p) k, || (X ^+k): mynorm | <= (|| X : mynorm |)^+k. *)
(* Proof. *)
(* move=> X k; elim k. *)
(* by right ; rewrite /GRing.exp. *)
(* move=> n IHn; rewrite !GRing.exprS. *)
(* have H1:= prod_ineq_norm mynorm X (X^+n). *)
(* rewrite /(@GRing.mul) //=. *)
(* apply Rle_trans with (|| X : mynorm | * || X^+n : mynorm |). *)
(* done. *)
(* apply Rmult_le_compat_l. *)
(* apply pos_norm. *)
(* done. *)
(* Qed. *)



(* Lemma th4_247: *)
(* forall (X: 'M_p) , || X : mynorm| < R1 ->  exists S,  cv_mat_ser *)
(*           (fun n =>  X^+n) S. *)
(* Proof. *)
(* move=> X Hx. *)
(* apply th2_246. *)
(* rewrite /cv_mat_ser. *)
(* have H1: forall i j, {S| *)
(*   Un_cv (fun N : nat => sum_f_R0 (fun n : nat =>Mabs  (X ^+n) i j) N) S }. *)
(* move=> i j. *)
(* apply Rseries_CV_comp with (fun n => (|| X : mynorm |)^+n). *)
(* move=> n; rewrite !mxE; split; [apply Rabs_pos|]. *)
(* apply Rle_trans with ( || X^+n : mynorm |). *)
(* apply canonical_norm. *)
(* apply matr_prod_ineq. *)
(* have H3: Rabs (|| X : mynorm |) < 1. *)
(* rewrite Rabs_right;[done|apply Rle_ge; apply: pos_norm]. *)

(* have H2:= GP_infinite (|| X : mynorm |) H3 . *)
(* exists (/(1 - || X : mynorm |)). *)
(* move=> eps He; case (H2 eps He)=> N Hn0; exists N => n Hn. *)
(* have H4:= Hn0 n Hn. *)
(* rewrite /R_dist in H4 |-*. *)
(* apply Rle_lt_trans with  *)
(*  (Rabs (sum_f_R0 (fun n : nat => 1 * || X : mynorm | ^ n ) n - / (1 - || X : mynorm |)) ). *)
(* right. *)
(* congr (Rabs _). *)
(* congr ( _ - _). *)
(* elim n => //=. *)
(* by rewrite !expr0 mulr1. *)
(*  move=> n0 Ihn;  rewrite Ihn. *)
(* congr ( _ + _ ). *)
(* rewrite exprS mul1r. *)
(*  congr ( _ * _). *)
(* elim n0 => //=; move=> n1 Ihn1; rewrite exprS /GRing.mul /=. *)
(* by  congr ( _ * _ ). *)
(* done. *)
(* have Ha: forall i j : 'I_p, *)
(*      exists S,  *)
(*      (Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Mabs (X ^+n) i j) N) S). *)
(* by move=> i j; case (H1 i j) => S Hs; exists S. *)
(* have Hz:  forall i : 'I_p, *)
(*      exists S, forall j,  *)
(*      (Un_cv (fun N : nat =>  *)
(*               sum_f_R0 (fun n : nat => Mabs (X ^+n) i j) N) (S j)). *)
(* move=> i; case (choice _ (Ha i)) => ch1 Hch1; by exists ch1. *)
(* case (choice _ Hz) => ch Hch. *)
(* by exists (\matrix_(i, j) ch i j) => i j; rewrite !mxE. *)
(* Qed. *)
 

(* Lemma sum_prod_id: *)
(* forall (X: matrix R p p) k , ((1 - X)*(\sum_(0<=i < k.+1) (X^+i))  = 1 - X^+(k.+1))%Ri. *)
(* Proof. *)
(* move=> X k. *)
(* replace (1 - X)%Ri with (- (-1 + X))%Ri;[|by rewrite opprD opprK]. *)
(* rewrite mulNr  mulrDl  mulN1r opprD opprK. *)
(* rewrite big_distrr  /=. *)
(* rewrite  -sumrB //=. *)
(* induction k => //=. *)
(* by rewrite big_mkord big_ord_recl big_ord0 expr0 expr1 addr0 mulr1. *)
(* rewrite !big_mkord in IHk |-*.  *)
(* rewrite big_ord_recr //= IHk . *)
(* by rewrite !exprS addrA  mulrA addrC -addrA addNr addr0 addrC. *)
(* Qed. *)


(* Lemma mat_to_zero: *)
(* forall (X: 'M_p) , || X : mynorm | < 1 -> limit_m (fun n => X^+n) 0. *)
(* Proof. *)
(* move=> X Hx. *)
(* apply th1_246. *)
(* have H1: exists S0 : matrix real_field p p, cv_mat_ser (fun n : nat => X ^+n) S0. *)
(* by apply th4_247. *)
(* by case H1 => S Hs; exists S; *)
(* apply/echiv_mat_ser. *)
(* Qed. *)


(* Lemma th5_aj1: *)
(* forall (X: 'M_p) S, || X : mynorm | < 1 -> cv_mat_ser2 (fun n => X^+n) S -> (1 - X) * S = 1. *)
(* Proof. *)
(* move=> X S Hmn1 Hcv. *)
(* red in Hcv. *)
(* have Hcv2:=  (@limit_m_mul _ _ _  _ (1 - X)%Ri _ Hcv). *)
(* have Hcv3: limit_m (fun n : nat => ((1 - X) * (\sum_(i < n.+1) X ^+i))%Ri) *)
(*          (1)%Ri. *)
(* apply el_eq_mat_seq_cv with (fun n => ((1 - X^+n.+1))%Ri). *)
(* by move=> n; rewrite  -sum_prod_id big_mkord. *)
(* replace (1:matrix_ringType real_ringType p' ) with ((1 + 0):matrix real_field p p). *)

(* apply: limit_m_add. *)
(* by move=> i j eps He; exists O=> n Hn; rewrite /R_dist !mxE addr0 /Rminus Rplus_opp_r *)
(* Rabs_R0. *)
(* apply limit_m_opp_0; apply limit_m_Sn ; rewrite //=. *)
(* by apply: mat_to_zero. *)
(* by apply/matrixP => i j; rewrite !mxE addr0. *)
(* by apply limit_m_uniq with (fun n =>  ((1 - X) * (\sum_(i < n.+1) X ^+i))%Ri); [|done] . *)
(* Qed. *)


(* Lemma th5_248: *)
(* forall X: 'M_p, || X : mynorm | < R1 -> \det (1 - X) <> 0. *)
(* Proof. *)
(* move=> X H. *)
(* case (@th4_247 X H) => S HS. *)
(* move: HS; rewrite echiv_mat_ser => HS. *)
(* have H1:= @th5_aj1 X S H HS. *)
(* have H0: (\det (1 - X) * \det S)%Ri <> 0. *)
(* rewrite -(@detM _  p' ((1 - X))  S) H1 det1 . *)
(* rewrite /1 /0 /=;  auto with real.  *)
(* rewrite /GRing.mul //= in H0 . *)
(* by case (RIneq.Rmult_neq_0_reg (\det (1 - X)) (\det S) H0). *)
(* Qed. *)


(* Lemma cor1_th5: *)
(* forall X: 'M_p,  || (1 - X) : mynorm | < R1 -> \det X <> 0. *)
(* Proof. *)
(* move => X Hn. *)
(* have Heq: X = 1 - (1 - X). *)
(* apply/matrixP => i j; rewrite !mxE; toR. *)
(* by rewrite Ropp_minus_distr'  Rplus_comm Rplus_assoc Rplus_opp_l Rplus_0_r. *)
(* by rewrite Heq; apply th5_248. *)
(* Qed. *)


(* Lemma cor2_th5: *)
(* forall X : 'M_p, || X : mynorm | < R1 ->  cv_mat_ser2 (fun n => X^+n) ((1 - X) ^-1). *)
(* Proof. *)
(* move=> X H. *)
(* rewrite -echiv_mat_ser. *)
(* case (@th4_247 X H) => S HS. *)
(* move: HS; rewrite echiv_mat_ser => HS. *)
(* have Hm:= @th5_aj1 X S H HS.  *)
(* have Heq2: (1 - X) * ((1 - X)^-1) = 1. *)
(* by apply: mulmxV ; apply /eqP; apply th5_248. *)
(* have Heq : S = ((1 - X)^-1). *)
(* have Heq:  (1 - X) * S =  (1 - X) * (1 - X)^-1. *)
(* by rewrite Hm Heq2. *)
(* have H1: \det (1 - X) != 0. *)
(* by apply/eqP; apply th5_248. *)
(* apply:  (mulrI _ Heq); rewrite /GRing.unit //=. *)
(* by rewrite -Heq echiv_mat_ser. *)
(* Qed. *)


(* Lemma norm_inv_ser0: *)
(* forall (a: 'M_p), || a : mynorm | < R1 -> || (1: 'M_p) : mynorm | = 1 -> *)
(*  || ( 1 -  a)^-1 : mynorm | <= 1 / (1 - || a : mynorm | ). *)
(* Proof. *)
(* move=> X H H1. *)
(* have H2 := @cor2_th5 X H. *)
(* red in H2; move: H2; rewrite (@lem_cv1 p p mynorm) => H2. *)
(* have H3: Un_cv (fun k : nat =>  || \sum_(i < k.+1) X ^+i : mynorm | )   *)
(*  ( || (1 - X)^-1 : mynorm |) . *)
(* move=> eps He. *)
(* case (H2 eps He) => N HN; exists N => n Hn. *)
(* rewrite /R_dist in HN |-*. *)
(* have HN' := HN n Hn. *)
(* rewrite Rminus_0_r in HN'. *)
(* apply Rle_lt_trans with ( || (1 - X)^-1 - \sum_(i < n.+1) X ^+i : mynorm | ). *)
(* have Heq:    || (1 - X)^-1 - \sum_(i < n.+1) X ^+i : mynorm | = *)
(*  || \sum_(i < n.+1) X ^+i - (1 - X)^-1 : mynorm|. *)

(* have Hb:= eq_norm_min mynorm (\sum_(i < n.+1) X ^+i)  ((1 - X)^-1). *)
(* (*rewrite !dif_mult_sc_1 in Hb |-*.*) *)
(* rewrite  /GRing.opp /GRing.one //= in Hb. *)
(* rewrite Heq. *)
(* have H7:= tri_min2 mynorm (\sum_(i < n.+1) X ^+i)  ((1 - X)^-1). *)
(* (*rewrite !dif_mult_sc_1 in H7 |-*.*) *)
(* rewrite  /GRing.opp /GRing.one //= in H7. *)
(* rewrite -(Rabs_right ( || (1 - X)^-1 - \sum_(i < n.+1) X ^+i : mynorm | ) ); *)
(* [done| apply Rle_ge ; apply pos_norm]. *)
(* apply lim_seq_monot with ((fun k : nat => || \sum_(i < k.+1) X ^+i : mynorm |)) *)
(*  ((fun k : nat => (\sum_(i < k.+1) (|| X : mynorm |) ^+i))). *)
(* move=> n. *)
(* apply Rle_trans with (\sum_(i < n.+1) || X ^+i : mynorm |). *)
(* apply: mat_norm_sum. *)
(* apply: big_ind2. *)
(* by right. *)
(* by move=>*; apply Rplus_le_compat. *)
(* by move=> *; apply matr_prod_ineq. *)
(* (* add this statement as a new lemma: norm_sum_ineq *)
(* mynorm p q  (\sum_(i <= n) X ^+i) <= \sum_(i <= n) mynorm p q  X ^+i *)
(* proof using big_op_rel; give generalisation fu: R1->R2; *)
(* R1:= matrix R p p; R2:= R *)


(* unlock reducebig. *)
(* elim ((index_enum I_(n.+1))) => //=. *)
(* have Heq:  (mynorm p q  0) =  (mynorm p q  (0 *s \1)). *)
(* f_equal; apply /matrixP=> i j; rewrite !mxE; simp_R; ring. *)
(* rewrite Heq hom_pos_norm Rabs_R0 Rmult_0_l; by right. *)
(* move=> N s IHs. *)
(* apply Rle_trans with ( mynorm p q  *)
(*   (X ^+N) + mynorm p q  ( *)
(*    foldr (fun (i : ordinal n.+1) (x : matrix real_ring p p) => X ^+i + x) 0 s)%Ri ). *)
(* apply: trin_ineq_norm. *)
(* apply Rplus_le_compat. *)
(* apply matr_prod_ineq. *)
(* done. *)
(* *) *)
(* done. *)
(* have H4: Rabs || X : mynorm | < 1. *)
(* rewrite Rabs_right; [done| apply Rle_ge; apply pos_norm]. *)
(* have Hpser:=  GP_infinite (|| X : mynorm |) H4. *)
(* red in Hpser; red in Hpser. *)

(* move=> eps He; case (Hpser eps He)=> N Hn0; exists N => n Hn. *)
(* have H5:= Hn0 n Hn. *)
(* rewrite /R_dist in H5 |-*. *)
(* apply Rle_lt_trans with  *)
(*  (Rabs (sum_f_R0 (fun n : nat => 1 * || X : mynorm | ^ n) n - / (1 - || X : mynorm |)) ). *)
(* right. *)
(* congr (Rabs _). *)
(* congr ( _ - _). *)
(* rewrite -echiv_sums //=. *)
(* elim n => //=. *)
(* rewrite expr0; toR; ring. *)
(* move=> n0 Ihn; rewrite Ihn. *)
(* rewrite exprS; toR. *)
(* congr ( _ + _ ); rewrite Rmult_1_l. *)
(*  congr ( _ * _). *)
(* elim n0 => //=; move=> n1 Ihn1. *)
(* rewrite exprS; by congr ( _ * _ ). *)
(* by rewrite /Rdiv Rmult_1_l. *)
(* done. *)
(* Qed. *)

(* End limitsqrMat. *)

Section posdefMat.

Variable R : numDomainType.
Variable p : nat.

(* Definition sym_mx (T: eqType) (A:'M[T]_p):= forall i j, A i j = A j i. *)

(* (* the notion of positive definite matrix is given for a symetric matrix*) *)
(* Definition pos_def (A: 'M_p) := *)
(*   sym_mx A /\ forall (x: 'cV_p), x != 0 ->  0 <m: (x^T *m A *m x). *)

(* (* is there a reason why it should not be defined in general? *) *)
Definition pos_def_gen (A: 'M[R]_p) :=
  forall (x: 'cV_p), x != 0 ->  0 <m: (x^T *m A *m x).

End posdefMat.

Section vectorMat.

Variable n : nat.
(* Notation n := n'.+1. *)

(* Lemma vec_prod_sum (R : ringType) (z : 'cV[R]_n) : *)
(*   z^T *m z = (\sum_k z^T ord0 k * z k ord0)%:M. *)
(* Proof. by apply/matrixP => i j; rewrite !mxE !ord1 eqxx. Qed. *)

(* Lemma sum_sqr_gt (R : numDomainType) : *)
(* forall x: 'cV[R]_n, x != 0 ->  0 < \sum_k x^T ord0 k * x k ord0. *)
(* move=> x Hx0. *)
(* have [i0 [j0 Hij0]] : exists i0, exists j0, x i0 j0 != 0. *)
(*   move/eqP/matrixP : Hx0 => Hx0. *)
(*   have : ~~[forall i', [forall j', x i' j' == 0]]. *)
(*     apply/forallP => ab; case: Hx0 => i j; move/forallP: (ab i) => ab'. *)
(*     by apply/eqP; rewrite mxE; apply: ab'. *)
(*   rewrite negb_forall => /existsP [u]; rewrite negb_forall => /existsP [v P]. *)
(*   by exists u, v. *)
(* rewrite (bigD1 i0) //=. *)
(* apply: ltr_le_trans (_ : x i0 j0 ^+2 + 0 <= _). *)
(* attempt to port stops here. *)
(*   Search (0 < _ ^+2). *)
(* apply Rlt_le_trans with (\sum_k Rsqr (x k ord0)). *)
(* apply Rlt_le_trans with (Rsqr (x i0 j0)). *)
(* by apply Rlt_0_sqr. *)
(* (*rewrite /Rsqr /GRing.exp /=; toR.*) *)
(* rewrite (ord1_0 j0). *)
(* rewrite (bigD1 i0) //=. *)
(* rewrite -{1}(addr0 (Rsqr (x i0 ord0))). *)
(* apply Rplus_le_compat_l. *)
(* elim: index_enum; [by rewrite big_nil; right|]. *)
(* move=> t s IH; rewrite big_cons; case C: (t!=i0). *)
(* rewrite -{1}(addr0 0); apply Rplus_le_compat. *)
(* rewrite /GRing.exp //=; toR; rewrite -/(Rsqr (x t ord0)); apply Rle_0_sqr. *)
(* done. *)
(* done. *)
(* by right; apply eq_bigr=> k _; rewrite /trmx mxE /GRing.exp /=. *)
(* Qed. *)


Lemma vec_prod_pos (R : rcfType) :
  forall (z : 'cV[R]_n), reflect (0 <m:  z^T *m z) (z != 0 ).
Proof.
case: n => [z | n' x].
  have -> : (z = 0) by apply/matrixP => [] [].
  rewrite eqxx mulmx0; apply: ReflectF.
  by move/(_ 0 0); rewrite ltrr.
case Hx0: (x != 0); [apply: ReflectT |apply: ReflectF].
  move => i j; rewrite !mxE !ord1.
  pose k := odflt ord0 [pick k | x k j != 0].
  rewrite (bigD1 k) //= ltr_spaddl // ?mxE ?ltr_def.
    rewrite sqr_ge0 andbT mulf_eq0 orbb.
    rewrite /k; case: pickP=> //= predx; first by rewrite ord1.
    rewrite -(ord1 j) predx -(negbTE Hx0).
    apply/eqP/matrixP=> h l; rewrite mxE; apply/eqP.
    by rewrite /is_true -[true]negbK [X in ~~ X]/= -(predx h) !ord1 negbK.
  by rewrite sumr_ge0 // => i' _; rewrite mxE sqr_ge0.
move/negbT:Hx0; rewrite negbK => /eqP => ->.
by rewrite mulmx0 => /(_ 0 0); rewrite ltrr.
Qed.

(* Lemma vec_prod_unit: *)
(* forall (z: 'cV[R]_n),  (z != 0) -> (z^T *m z \in unitmx). *)
(* Proof. *)
(* move=> x Hx0. *)
(* apply/eqP. *)
(* rewrite (mx11_scalar (x^T *m x)) det_scalar1. *)
(* apply Rgt_not_eq. *)
(* move/vec_prod_pos: Hx0 => Hx0. *)
(* have H := Hx0 ord0 ord0. *)
(* rewrite mxE in H; apply: H. *)
(* Qed. *)


(* Lemma vec_prod_abs:  *)
(* forall x: 'cV[R]_n,  x^T *m x = (Mabs x) ^T *m Mabs x. *)
(* move=> x. *)
(* apply/matrixP=> i j; rewrite !vec_prod_sum !mxE (ord1_0 i) (ord1_0 j) /= !GRing.mulr1n. *)
(* apply: eq_bigr => k _; rewrite /trmx !mxE; toR. *)
(* have Heqx:  (x k ord0 * x k ord0 = Rsqr (x k ord0))%Re by rewrite //. *)
(* have Heqxa:  (Rabs (x k ord0) * Rabs (x k ord0) = Rsqr (Rabs (x k ord0)))%Re  *)
(*   by rewrite //. *)
(* rewrite Heqx Heqxa; apply Rsqr_abs. *)
(* Qed. *)


(* Lemma det_scalar_invmx: *)
(* forall a:R, (@scalar_mx _ 1 a) \in unitmx -> *)
(*    (@scalar_mx _ 1 a)^-1 = (a^-1)%:M. *)
(* move=> a Ha. *)
(* set b:= a^-1. *)
(* rewrite   /GRing.inv /= /invmx . *)
(* rewrite Ha. *)
(* rewrite det_scalar1. *)
(* have Ha1: \adj a%:M = (1%:M: 'M_1). *)
(* rewrite /adjugate -matrixP => i j; rewrite (ord1_0 i) (ord1_0 j) !mxE /=. *)
(* rewrite expand_cofactor. *)
(* transitivity  (\sum_(s: 'S_1 | s ord0 == ord0) R1 ). *)
(* apply eq_bigr => s Hs. *)
(* have Heq: s = perm_one _. *)
(* rewrite /perm_one . *)
(* by rewrite -permP => k; rewrite (ord1_0 k) permE; apply/eqP. *)
(* rewrite Heq odd_perm1 /= expr0 mul1r. *)
(* rewrite big_pred0; first by done. *)
(* by move=> k; rewrite //= (ord1_0 k) /=. *)
(* rewrite  big_const_seq //=. *)
(* rewrite count_filter. *)
(* have H1 : *)
(* (size *)
(*      (filter (fun i1 : 'S_1 => i1 ord0 == ord0) *)
(*         (index_enum (perm_for_finType (ordinal_finType 1))))) = 1%Dn. *)
(* rewrite (@eq_filter _ _ (predT )); last first. *)
(* by move=> s; rewrite //= (ord1_0 (s ord0)). *)
(* rewrite filter_index_enum -cardT. *)
(*  rewrite /=. *)
(* rewrite -{10}(@cards1 (perm_for_finType (ordinal_finType (S O)))   *)
(*   (perm_one (ordinal_finType (S O)))). *)
(* apply: eq_card. *)
(* move=> s. *)
(* have Heq: s = perm_one _. *)
(* rewrite /perm_one . *)
(* by rewrite -permP => k; rewrite (ord1_0 k) permE (ord1_0 (s ord0)). *)
(* by rewrite Heq in_simpl /= set11. *)
(* by rewrite H1 //= addr0. *)
(* by rewrite Ha1 scalemx1 /b. *)
(* Qed. *)

End vectorMat.


 




(* 
*** Local Variables: ***
*** coq-prog-name: "/auto/sop-nas2a/u/sop-nas2a/vol/home_marelle/gcano/sandbox/coq-8.3pl2/bin/coqtop" ***
*** coq-load-path: ("/auto/sop-nas2a/u/sop-nas2a/vol/home_marelle/gcano/coqfinitgroup/ssreflect/ssreflect1.3_v8.3/src/" "/auto/sop-nas2a/u/sop-nas2a/vol/home_marelle/gcano/coqfinitgroup" "/auto/sop-nas2a/u/sop-nas2a/vol/home_marelle/gcano/interval/") ***
*** End: ***
 *)