Require Import Logic Classical_Prop Classical_Pred_Type IndefiniteDescription.
Require Import Reals Fourier.
Require Import ssreflect ssrbool ssrfun ssrnat ssralg eqtype ssrfun binomial.
Require Import bigop matrix ssrnum finfun fintype seq choice polyrcf.
Require Import struct_top mxorder set_theory outils.
Require Import Rstruct ssrcomplements. 
Require Import mxstructure jordan.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Section matrimetrie.

(******Espace metrique et topologique des marices réelles********************)
Variable p q : nat.
Variable R : archiFieldType.
Let N:= can_norm8 R.

Open Scope ring_scope.

Definition dist_mat (A B : 'M[R]_(p,q)) := 
  ||(A - B) : N|.

Lemma dist_mat_pos (A B : 'M[R]_(p,q)) : 0 <= dist_mat A B.
Proof. exact: pos_norm. Qed.

Lemma dist_mat_sym (A B : 'M[R]_(p,q)) : dist_mat A B = dist_mat B A.  
Proof. by rewrite /dist_mat -opprB cnormN. Qed.

Lemma dist_mat_refl (A B : 'M[R]_(p,q)) : dist_mat A B = 0 <-> A = B.
Proof.
rewrite /dist_mat; split=> [/eqP|->].
  by rewrite cnorm_eq0 subr_eq0=> /eqP.
by rewrite subrr cnorm0.
Qed.

Lemma dist_mat_tri (A B C : 'M[R]_(p,q)) :
  dist_mat A B <= dist_mat A C + dist_mat C B.
Proof. by rewrite /dist_mat -{1}(addrNK C A) -addrA trin_ineq_norm. Qed.

Canonical Structure EspMet_matMixin := 
   EspMetMixin dist_mat_pos dist_mat_sym dist_mat_refl dist_mat_tri.
Canonical Structure EspMet_mat :=
  EspMetType ('M[R]_(p,q)) EspMet_matMixin.
Canonical Structure EspTop_mat := 
  Eval hnf in TopologyType ('M[R]_(p,q)) (EspMet_TopMixin EspMet_mat).
(*   Eval hnf in [topologyType of ('M[R]_(p,q)) for EspMet_mat]. *)


(**************Espace métrique et topologique des réels*******************)

Variable T : archiFieldType.

Definition dist_real (x y : T) := `|x - y|.

Lemma dist_real_ge0 x y : 0 <= dist_real x y.
Proof. exact: normr_ge0. Qed.

Lemma dist_real_sym x y : dist_real x y = dist_real y x.
Proof. exact: distrC. Qed. 

Lemma dist_real_refl x y : dist_real x y = 0 <-> x = y.
Proof. 
rewrite /dist_real; split=> /eqP. 
  by rewrite normr_eq0 subr_eq0=> /eqP.
by rewrite -subr_eq0 -normr_eq0=> /eqP.
Qed.

Lemma dist_real_tri x y z : dist_real x y <= dist_real x z + dist_real z y.
Proof. by rewrite /dist_real -{1}(addrNK z x) -addrA ler_norm_add. Qed.
 

(* Lemma R_dist_pos : forall x y, 0 <= R_dist x y. *)
(* Proof. *)
(* move=> x y; apply: Rge_le. *)
(* by apply: R_dist_pos. *)
(* Qed. *)

Canonical Structure EspMet_archiMixin :=
  EspMetMixin dist_real_ge0 dist_real_sym dist_real_refl dist_real_tri.
Canonical Structure EspMet_archi :=
 Eval hnf in  EspMetType T EspMet_archiMixin.
Canonical Structure Esp_Top_archi := 
 Eval hnf in TopologyType T (EspMet_TopMixin EspMet_archi).

(* Print Canonical Projections. *)
(* Check topologyType. *)
(* Print EspMet_TopType. *)
(* Canonical Structure EspTop_R :=  *)
(*   Eval hnf in [topologyType of R for EspMet_TopType EspMet_R]. *)

(* Set Printing All. *)
(* Print EspTop_R. *)


End matrimetrie.

Section Theory.
Open Scope ring_scope.

Variable R : archiFieldType.


(**********déplacé dans "pos"*********)
Lemma lt_eps_le (a b : R) : (forall eps, 0 < eps -> a < b + eps) -> a <= b.
Proof.
move=> H.
rewrite lerNgt -subr_gt0; apply/negP=>/(H (a - b)).
by rewrite addrC addrNK ltrr.
Qed.
(***********************************)
Lemma le_cvg_le (a b : R) (U_ : nat -> R) : 
 (forall k, a <= U_ k) -> (U_ >->> b) -> a <= b.
Proof.
move=> Hle /Un_cvg2 Hcvg.
apply: lt_eps_le=> r Hpos.
have [N HN] := (Hcvg (PosD Hpos)).
rewrite (ler_lt_trans (Hle N)) // addrC -[X in X < _](addrNK b).
rewrite ltr_add2r -ltr_opp2 opprB.
have := HN N (leqnn N).
rewrite /open_ball ltr_norml.
by case/andP .
Qed.

Lemma lt_cvg_le (a b : R) (U_ : nat -> R) : 
 (forall k, a < U_ k) -> (U_ >->> b) -> a <= b.
Proof.
move=> Hlt /Un_cvg2 Hcvg.
apply: lt_eps_le=> r Hpos.
have [N HN] := (Hcvg (PosD Hpos)).
rewrite (ltr_trans (Hlt N)) // addrC -[X in X < _](addrNK b).
rewrite ltr_add2r -ltr_opp2 opprB.
have := HN N (leqnn N).
rewrite /open_ball ltr_norml.
by case/andP .
Qed.

(*******************)
(*********déplacé dans struct_top.v*****************************)
Lemma cvg_const (T : topologyType) (a : T) : (fun _ => a) >->> a.
Proof.
move=> S HS.
exists 0%N=> n _.
exact: in_neigh.
Qed.

Lemma cvg_ext  (T : topologyType) (a : T) (U V : nat -> T) :
  U =1 V -> U >->> a -> V >->> a.
Proof.
move=> HUV HU S HS.
have [N HN]:= HU S HS.
by exists N => n Hn; rewrite -HUV; apply: HN.
Qed.

Lemma cvg_shift (T : topologyType) (a : T) (U : nat -> T) k :
   U >->> a <-> (fun i => U (i + k)%N) >->> a.
Proof.
split=> Hcvg S HS; have[N HN]:= Hcvg S HS.
  by exists (N - k)%N=> n Hn; apply: HN; rewrite addnC -leq_subLR.
exists (N + k)%N=> n Hn.
have/(leq_trans (leq_addl N k)) Hn2 := Hn.
by rewrite -[n](subnK Hn2); apply: HN; rewrite -(leq_add2r k) subnK.
Qed.

(**********************************)

Lemma cvg_mulr (a b : R) (U : nat -> R) : U >->> a -> 
(fun k => b * U k) >->> (b * a).  
Proof.
move=> /Un_cvg2 HUa.
case: (altP (b =P 0))=> [->|Hb]; first rewrite mul0r.
  have H1: (fun _ => 0) =1 (fun k => 0 * U k) by move=> x /=; rewrite mul0r.
  by apply/(cvg_ext H1)/cvg_const.
apply/Un_cvg2=> r.
have Hpos : 0 < sort r / `|b|.
  by rewrite mulr_gt0 ?lt_pos // invr_gt0 normr_gt0.
have[N HN]:= HUa (PosD Hpos).
exists N=> n Hn; rewrite /open_ball /= /dist_real.
rewrite -mulrBr normrM mulrC -ltr_pdivl_mulr ?normr_gt0 //.
exact: HN.
Qed.

Lemma cvg_mull (a b : R) (U : nat -> R) : U >->> a -> 
(fun k => U k * b) >->> (a * b).  
Proof.
move=> HUa.
have H1 : (fun k : nat => b * U k) =1 (fun k : nat => U k * b) .
  by move=> x /=; rewrite mulrC.
by rewrite mulrC; apply/(cvg_ext H1)/cvg_mulr.
Qed.

Lemma cvg_normr0 (f : nat -> R) : (fun k => `|f k|) >->> 0 -> f >->> 0.
Proof.
move/Un_cvg2 => nf0; apply/Un_cvg2 => r; have [N Hn] := nf0 r.
exists N => n nN; have := Hn n nN.
by rewrite /open_ball /= /dist_real; rewrite !add0r !normrN normr_id.
Qed.

Lemma cvg_lt1 (x : R) : `|x| < 1 -> (fun k => x^+ k) >->> 0.
Proof.
move=> xsm. 
have := normr_ge0 x; rewrite ler_eqVlt.
case/orP.
  rewrite eq_sym normr_eq0 => /eqP ->.
  apply/(@cvg_shift (Esp_Top_archi R) 0 [eta GRing.exp 0] 1%N).
  have H:  (fun _ => (0 : R)) =1 (fun k => 0^+ (k + 1)).
    by move=> y /=; rewrite addn1 expr0n.
  by apply/(cvg_ext H)/cvg_const.
move=> xgt0; apply: cvg_normr0.
apply/Un_cvg3=> n.
set N := Num.bound (n.+1%:R * (`|x| ^-1 - 1) ^-1).
exists N => n' Nn'.
rewrite /open_ball /= /dist_real add0r normrN normr_id normrX.
have xXNp : 0 < `|x| ^+ N  by apply/exprn_gt0.
apply: (ler_lt_trans (_ : _ <= `|x | ^+ N)).
  rewrite -[N]addn0 -(subnKC Nn') ?exprD ?(ler_pmul2l xXNp).
  rewrite expr0; elim: (n' - N)%N; first by rewrite expr0.
  move => m ind; rewrite -[1]mulr1 exprS ler_pmul //; last exact: ltrW.
  by apply/ltrW/exprn_gt0.
rewrite {n' Nn'} -(ltr_pmul2l (ltr0Sn _ n)) mulfV; last first.
  by move: (ltr0Sn R n); rewrite lt0r; case/andP.
have xXNnz : `|x| ^+ N != 0 by move: xXNp; rewrite lt0r; case/andP.
rewrite -[X in _ < X](mulVf xXNnz) (ltr_pmul2r (xXNp)).
rewrite -exprVn (_ : `|x|^-1 = 1 + (`|x|^-1 - 1)); last first.
  by rewrite addrCA addrN addr0.
have invx1 : 0 < `|x| ^- 1 - 1 by rewrite subr_gt0 invf_gt1.
have bnd_exp: forall a : R, 0 < a -> 1 + N%:R * a <= (1 + a) ^+ N.
  clear; elim: N; clear n.
    by move => a ap; rewrite mul0r addr0 expr0.
  move => n Ih a ap.
  have -> : n.+1%:R = 1 + n%:R :>R.
    by clear; case: n; unfold GRing.natmul => //; rewrite addr0.
  rewrite mulrDl mul1r addrCA exprS mulrDl mul1r addrC.
  apply ler_add; first exact: Ih.
  rewrite -{1}[a]mulr1 (ler_pmul2l ap).
  by apply: (ler_trans _ (Ih _ ap)); rewrite cpr_add mulr_ge0 // ?ler0n ?ltrW.
have nn1 : n.+1%:R < 1 + n.+1%:R :> R by rewrite cpr_add ltr01.
apply: (ltr_trans nn1).
apply: (ltr_le_trans _ (bnd_exp _ invx1)).
have x1nz : `|x| ^-1 - 1 != 0 by move: invx1; rewrite lt0r; case/andP.
rewrite ltr_add2l -[_%:R]mulr1 -[X in _ * X < _](mulVf x1nz) mulrA.
rewrite (ltr_pmul2r invx1); apply/archi_boundP/ltrW/divr_gt0 => //.
apply ltr0Sn.
Qed.


End Theory.

Section mxTheory.


Variable R : archiFieldType.

Open Scope ring_scope.
Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Local Notation N8 := (can_norm8 R).

Lemma mx_cvgP m n (U_ : nat -> 'M_(m,n)) (A :'M[R]_(m,n)) :
  U_ >->> A <-> (forall i j, (fun n => (U_ n) i j) >->> (A i j)).
Proof.
split=> [/Un_cvg2 HUcvA i j|Hcv_coeff]; apply/Un_cvg2=> r.
  case: (HUcvA r)=> N HN.
  exists N => k Hk.
  have Hds: d (A i j) ((U_ k) i j) <= \sum_l (d (A i l) ((U_ k) i l)).
    by rewrite (bigD1 j) // ler_addl sumr_ge0 // => l _; rewrite dist_pos.
  have : \sum_l (d (A i l) ((U_ k) i l)) <= ||(A - U_ k) : N8|. 
    have H p :  0 <= \sum_l (d (A p l) ((U_ k) p l)).
      by rewrite sumr_ge0 // => l _; rewrite dist_pos.
    have Hsm:= bigmax_ger H (mem_index_enum i).
    rewrite (ler_trans Hsm) // (bigmax_ler H) // => l.
    by rewrite ler_sum // => p _; rewrite !mxE.
   move/(ler_trans Hds)=> H.
   exact: (ler_lt_trans _ (HN k Hk)).
have {Hcv_coeff} Hcvcf i j : Un_cvg_met (fun n0 : nat => (U_ n0) i j) (A i j).
 by apply/Un_cvg2.
rewrite /open_ball.
case: n U_ A r Hcvcf=> [U_ A r Hcvcf|n U_ A r Hcvcf].
  exists 0%N=> n _.
  have/dist_refl -> : A = U_ n by apply/matrixP=> i [][].
  by rewrite lt_pos.
case: m U_ A r Hcvcf=> [U_ A r Hcvcf|m U_ A r Hcvcf].
  exists 0%N=> m _.
  have/dist_refl -> : A = U_ m by apply/matrixP=> [][][].
  by rewrite lt_pos.
have Hpos: 0 < (sort r) / n.+1%:R by rewrite mulr_gt0 ?lt_pos // inv_INR_pos.
pose N := \max_i (\max_j pi1 (cid (Hcvcf i j (PosD Hpos)))).
exists N=> k Hk.
have Hs i :  0 <= \sum_j `|(A - (U_ k)) i j|.
  by rewrite sumr_ge0.
have Hnnul : index_enum (ordinal_finType m.+1) != [::].
 by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [i /andP [_ /eqP Hi]]:= bigmax_eq Hs Hnnul.
rewrite /open_ball /= /dist_mat /= /norm8 -Hi.
have Hu: (n.+1%:R : R) \is a GRing.unit by rewrite unitfE pnatr_eq0. 
rewrite -[sort r](mulrVK Hu) mulr_natr -[X in _*+ X]card_ord -sumr_const.
apply: (ltr_sum _ (mem_index_enum (ord0 : 'I_n.+1)))=> // j _; rewrite !mxE.
apply/(pi2 (cid (Hcvcf i j (PosD Hpos))))/(leq_trans _ Hk).
apply: (leq_trans (leq_bigmax (F := (fun j => pi1 (cid (Hcvcf i j _)))) j)).
exact: (leq_bigmax (F := (fun i => \max_j pi1 (cid (Hcvcf i j _)))) i).
Qed.

Lemma lemx_cvg_lemx m n (A B :'M[R]_(m,n)) (U_ : nat -> 'M_(m,n)) : 
 (forall k, A <=m: U_ k) -> (U_ >->> B) -> A <=m: B.
Proof.
move=> Hlemx /mx_cvgP Hcvg i j.
have/le_cvg_le Hle k: A i j <= (U_ k) i j by apply: Hlemx.
exact: Hle.
Qed.

Lemma ltmx_cvg_lemx m n (A B :'M[R]_(m,n)) (U_ : nat -> 'M_(m,n)) : 
 (forall k, A <m: U_ k) -> (U_ >->> B) -> A <=m: B.
Proof.
move=> Hltmx /mx_cvgP Hcvg i j.
have/lt_cvg_le Hlt k: A i j < (U_ k) i j by apply: Hltmx.
exact: Hlt.
Qed.



Lemma norm8_col_mx m1 m2 n (A : 'M_(m1,n)) (D : 'M_(m2,n)) :
  || col_mx A D : N8| = maxr ||A : N8| ||D : N8|.
Proof.
case: m1 A=>[A|m1 A].
  by rewrite col_flat_mx [X in maxr X _]/= /norm8 big_ord0 maxr_r // pos_norm. 
have Hspos1 i : 0 <= \sum_j `|A i j| by rewrite sumr_ge0.
have Hnv1: (index_enum (ordinal_finType m1.+1)) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k1 /andP [_ /eqP Hk1]] := bigmax_eq Hspos1 Hnv1.
case: m2 D=>[D|m2 D].
  rewrite [X in maxr _ X]/= /norm8 big_ord0 maxr_l ?pos_norm //= /norm8.
  have Ho p (k : 'I_p) : (k < p + 0)%N by rewrite addn0.
  have Ho2 p : forall k : 'I_(p + 0), (k < p)%N by rewrite addn0.
  have Heq_sum (k : 'I_(m1.+1 + 0)) (l : 'I_m1.+1) : (k = l :> nat) -> 
     \sum_i `|(col_mx A D) k i| = \sum_i `|A l i|. 
    move=> Heq; apply/eq_bigr=> i _; rewrite mxE; case: splitP=> [j Hj|[][] //].
    by have ->: j = l by apply: ord_inj; rewrite -Hj.
  apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
    exists (Ordinal (Ho _ k1)); rewrite -Hk1 mem_index_enum andbT eq_sym. 
    by apply/eqP/Heq_sum.
  rewrite (Heq_sum _ (Ordinal (Ho2 _ j))) //.
  by apply: (bigmax_ger Hspos1); rewrite mem_index_enum.
have Hspos2 i : 0 <= \sum_j `|D i j| by rewrite sumr_ge0.
have Hnv2: (index_enum (ordinal_finType (m2.+1))) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k2 /andP [_ /eqP Hk2]] := bigmax_eq Hspos2 Hnv2.
rewrite /= /norm8.
apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
    rewrite/Num.max; case: ifP => _.
    exists (lshift m2.+1 k1); rewrite mem_index_enum -Hk1 andbT.
    by apply/eqP/eq_bigr=> i _; rewrite col_mxEu.    
  exists (rshift m1.+1 k2); rewrite mem_index_enum -Hk2 andbT.
  by apply/eqP/eq_bigr=> i _; rewrite col_mxEd.
case: (ltnP j m1.+1)=> H.
  have -> : j = (lshift _ (Ordinal H)) by apply: ord_inj.
  rewrite (eq_bigr (fun l => `|A (Ordinal H) l|))=>[|i _]; last first.
    by rewrite col_mxEu.
  by rewrite lter_maxr (bigmax_ger Hspos1) // mem_index_enum.
have Ho : (j - m1.+1 < m2.+1)%N.
  by rewrite -(ltn_add2r m1.+1) subnK // [X in (_ < X)%N]addnC.
have ->: j = rshift _ (Ordinal Ho) by apply: ord_inj=> /=; rewrite subnKC.
rewrite (eq_bigr (fun l => `|D (Ordinal Ho) l|))=>[|i _]; last first.
  by rewrite col_mxEd.
by rewrite lter_maxr (bigmax_ger Hspos2) ?orbT // mem_index_enum.
Qed.

Lemma norm8_row_mx0l m n1 n2 (A : 'M_(m,n2)) :
  || row_mx (0 : 'M_(m,n1)) A : N8| = ||A : N8|.
Proof.
case: m A=>[A|m A].
  have ->: A = 0 by apply/matrixP=> [][].
  by rewrite row_mx_const !cnorm0.
have Hspos i : 0 <= \sum_j `|A i j| by rewrite sumr_ge0.
have Hnv: (index_enum (ordinal_finType m.+1)) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k /andP [_ /eqP Hk]] := bigmax_eq Hspos Hnv.
rewrite /= /norm8.
have Heq j : \sum_i `|(row_mx 0 A) j i| = \sum_i `|A j i|.
  rewrite big_split_ord /= big1 ?add0r=> [|i _]; last first.
    by rewrite row_mxEl mxE normr0.    
  rewrite (eq_bigr (fun l => `|A j l|)) // => i _.
  by rewrite row_mxEr.
apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
  by exists k; rewrite mem_index_enum -Hk Heq eqxx.
by rewrite Heq (bigmax_ger Hspos) // mem_index_enum.
Qed.

Lemma norm8_row_mx0r m n1 n2 (A : 'M_(m,n1)) :
  || row_mx A (0 : 'M_(m,n2)) : N8| = ||A : N8|.
Proof.
case: m A=>[A|m A].
  have ->: A = 0 by apply/matrixP=> [][].
  by rewrite row_mx_const !cnorm0.
have Hspos i : 0 <= \sum_j `|A i j| by rewrite sumr_ge0.
have Hnv: (index_enum (ordinal_finType m.+1)) != [::].
  by rewrite -size_eq0 /index_enum -enumT size_enum_ord.
have [k /andP [_ /eqP Hk]] := bigmax_eq Hspos Hnv.
rewrite /= /norm8.
have Heq j : \sum_i `|(row_mx A 0) j i| = \sum_i `|A j i|.
  rewrite big_split_ord /= [X in _ + X]big1 ?addr0=> [|i _]; last first.
    by rewrite row_mxEr mxE normr0.    
  rewrite (eq_bigr (fun l => `|A j l|)) // => i _.
  by rewrite row_mxEl.
apply: bigmax_def=>[i||j _]; first by rewrite sumr_ge0.
  by exists k; rewrite mem_index_enum -Hk Heq eqxx.
by rewrite Heq (bigmax_ger Hspos) // mem_index_enum.
Qed.

Lemma norm8_blockmxul m1 m2 n1 n2 (A : 'M_(m1,n1)) :
  || block_mx A 0 0 (0 : 'M_(m2,n2)) : N8| = ||A : N8|.
Proof. 
rewrite /block_mx norm8_col_mx !norm8_row_mx0r. 
by rewrite maxr_l // cnorm0 pos_norm.
Qed.

Lemma norm8_blockmxdr m1 m2 n1 n2 (A : 'M_(m2,n2)) :
  || block_mx (0 : 'M_(m1,n1)) 0 0 A : N8| = ||A : N8|.
Proof.
rewrite /block_mx norm8_col_mx !norm8_row_mx0l.
by rewrite maxr_r // cnorm0 pos_norm.
Qed.


Lemma norm8_blockmxuldr m1 m2 n1 n2 (A : 'M_(m1,n1)) (D : 'M_(m2,n2)) :
  || block_mx A 0 0 D : N8| = maxr ||A : N8| ||D : N8|.
Proof. by rewrite /block_mx norm8_col_mx norm8_row_mx0l norm8_row_mx0r. Qed.


Lemma blockmx_cvgul m1 m2 n1 n2 (U : nat -> 'M[R]_(m1,n1)) (A : 'M_(m1,n1))
 B C (D : 'M_(m2,n2)) :
 (U >->> A) -> (fun k => block_mx (U k) B C D) >->> (block_mx A B C D).
Proof.
move/Un_cvg2=> HUA; apply/Un_cvg2=> r.
have [N HN]:= HUA r.
exists N=> n Hn.
rewrite /open_ball /= /dist_mat opp_block_mx add_block_mx !subrr.
by rewrite norm8_blockmxul HN.
Qed.


Lemma blockmx_cvgdr m1 m2 n1 n2 (U : nat -> 'M[R]_(m2,n2)) (A : 'M_(m1,n1))
 B C (D : 'M_(m2,n2)) :
 (U >->> D) -> (fun k => block_mx A B C (U k)) >->> (block_mx A B C D).
Proof.
move/Un_cvg2=> HUA; apply/Un_cvg2=> r.
have [N HN]:= HUA r.
exists N=> n Hn.
rewrite /open_ball /= /dist_mat opp_block_mx add_block_mx !subrr.
by rewrite norm8_blockmxdr HN.
Qed.


Lemma blockmx_cvguldr m1 m2 n1 n2 (U : nat -> 'M[R]_(m1,n1))
 (V : nat -> 'M[R]_(m2,n2)) (A : 'M_(m1,n1)) B C (D : 'M_(m2,n2)) :
 (U >->> A) -> (V >->> D) -> 
 (fun k => block_mx (U k) B C (V k)) >->> (block_mx A B C D).
Proof.
move=>/Un_cvg2 HUA /Un_cvg2 HVD.
apply/Un_cvg2=> r.
have [N1 HN1]:= HUA r.
have [N2 HN2]:= HVD r.
exists (maxn N1 N2)=> n Hn.
rewrite /open_ball /= /dist_mat.
rewrite opp_block_mx add_block_mx !subrr norm8_blockmxuldr.
by rewrite ltr_maxl HN1 ?HN2 // (leq_trans _ Hn) // ?leq_maxl // leq_maxr.
Qed.



Lemma diag_block_mx_cvg s (UF : nat -> forall n, nat -> 'M[R]_n.+1)
  (F : forall n, nat -> 'M[R]_n.+1) :
 (forall i, (i < size s)%N -> (fun k => UF k (nth 0%N s i) i) >->> (F (nth 0%N s i) i))
 -> (fun k => diag_block_mx s (UF k)) >->> (diag_block_mx s F).
Proof.
case: s => [_| a l]; first exact: cvg_const.
elim: l a F UF=> [a F UF H|a l /= IHl b F UF H]; first exact: (H 0%N).
apply: (blockmx_cvguldr (U:= fun k => UF k b 0%N)); first exact: (H 0%N).
by apply: IHl => i Hi; exact: (H i.+1).
Qed.

(*
Lemma cvg_jordan_lt1 (lam : R) n : `|lam| < 1 -> 
  (fun k => (Jordan_block lam n.+1)^+ k) >->> 0.
Proof.
move=> Hlt1.
apply/mx_cvgP=> i j; rewrite mxE.
apply/Un_cvg2=> r.
exists 3%N.
  move=> m Hm.
  rewrite Jordan_expn mxE.
*)

End mxTheory.




(* 
*** Local Variables: ***
*** coq-prog-name: "/home/gcano/sandbox/coq-8.4pl1/bin/coqtop" ***
*** coq-prog-args: ("-emacs-U" "-I" "./lib" "-I" "/home/gcano/coqfinitgroup" "-I" "/home/gcano/PF2") ***
*** End: ***
 *)