Require Import Reals.
Require Import ssreflect ssrbool ssralg eqtype ssrfun fintype seq bigop matrix.
Require Import ssrnum.
Require Import Rstruct (* real_matrix min_max *). 

Import GRing.Theory.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


 
Reserved Notation "\maxr_ i F"
  (at level 41, F at level 41, i at level 0,
           format "'[' \maxr_ i '/ '  F ']'").
Reserved Notation "\maxr_ ( i <- r | P ) F"
  (at level 41, F at level 41, i, r at level 50,
           format "'[' \maxr_ ( i  <-  r  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i <- r ) F"
  (at level 41, F at level 41, i, r at level 50,
           format "'[' \maxr_ ( i  <-  r ) '/ '  F ']'").
Reserved Notation "\maxr_ ( m <= i < n | P ) F"
  (at level 41, F at level 41, i, m, n at level 50,
           format "'[' \maxr_ ( m  <=  i  <  n  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( m <= i < n ) F"
  (at level 41, F at level 41, i, m, n at level 50,
           format "'[' \maxr_ ( m  <=  i  <  n ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i | P ) F"
  (at level 41, F at level 41, i at level 50,
           format "'[' \maxr_ ( i  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i : t | P ) F"
  (at level 41, F at level 41, i at level 50,
           only parsing).
Reserved Notation "\maxr_ ( i : t ) F"
  (at level 41, F at level 41, i at level 50,
           only parsing).
Reserved Notation "\maxr_ ( i < n | P ) F"
  (at level 41, F at level 41, i, n at level 50,
           format "'[' \maxr_ ( i  <  n  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i < n ) F"
  (at level 41, F at level 41, i, n at level 50,
           format "'[' \maxr_ ( i  <  n )  F ']'").
Reserved Notation "\maxr_ ( i 'in' A | P ) F"
  (at level 41, F at level 41, i, A at level 50,
           format "'[' \maxr_ ( i  'in'  A  |  P ) '/ '  F ']'").
Reserved Notation "\maxr_ ( i 'in' A ) F"
  (at level 41, F at level 41, i, A at level 50,
           format "'[' \maxr_ ( i  'in'  A ) '/ '  F ']'").
 


Section bigmax.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Variable R : numDomainType.


Notation "\maxr_ ( i <- r | P ) F" :=
  (\big[(@maxr R)/0%R]_(i <- r | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i <- r ) F" :=
  (\big[(@maxr R)/0%R]_(i <- r) F%R) : ring_scope.
Notation "\maxr_ ( i | P ) F" :=
  (\big[(@maxr R)/0%R]_(i | P%B) F%R) : ring_scope.
Notation "\maxr_ i F" :=
  (\big[(@maxr R)/0%R]_i F%R) : ring_scope.
Notation "\maxr_ ( i : I | P ) F" :=
  (\big[(@maxr R)/0%R]_(i : I | P%B) F%R) (only parsing) : ring_scope.
Notation "\maxr_ ( i : I ) F" :=
  (\big[(@maxr R)/0%R]_(i : I) F%R) (only parsing) : ring_scope.
Notation "\maxr_ ( m <= i < n | P ) F" :=
 (\big[(@maxr R)/0%R]_(m <= i < n | P%B) F%R) : ring_scope.
Notation "\maxr_ ( m <= i < n ) F" :=
 (\big[(@maxr R)/0%R]_(m <= i < n) F%R) : ring_scope.
Notation "\maxr_ ( i < n | P ) F" :=
 (\big[(@maxr R)/0%R]_(i < n | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i < n ) F" :=
 (\big[(@maxr R)/0%R]_(i < n) F%R) : ring_scope.
Notation "\maxr_ ( i 'in' A | P ) F" :=
 (\big[(@maxr R)/0%R]_(i in A | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i 'in' A ) F" :=
 (\big[(@maxr R)/0%R]_(i in A) F%R) : ring_scope.

Open Scope ring_scope.

Lemma bigmax_pos (I : Type) (F : I -> R) (s : seq I) :
  (forall i, 0 <= F i) -> 0 <= \maxr_(i <- s) F i.
Proof.
move=> Hge0.
apply: (big_ind (fun x : R => 0 <= x))=> // x y Hx Hy.
by rewrite /maxr; case: (y <= x).
Qed.


Lemma bigmax_ger (I : eqType) (F : I -> R) (s : seq I) i :
(forall i, 0 <= F i) ->  i \in s -> F i <= \maxr_(i <- s) F i.
Proof.
move=> Hge0.
elim: s=> // a l IHl.
set x := F a; set y := \big[_/_]_(_ <- _) _.
have Hxr := ger0_real (Hge0 a).
have Hyr := ger0_real (bigmax_pos l Hge0).
rewrite in_cons /y big_cons /maxr=> /orP []=> [/eqP ->|/IHl Hi];
case: (real_lerP Hyr Hxr)=> //.
  by move/ltrW.
exact: ler_trans.
Qed.

Lemma bigmax_eq (I : eqType) (F : I -> R) (s : seq I) :
  (forall i, 0 <= F i) ->  s != [::] -> 
  {k | (k \in s) && (F k == \maxr_(i <- s) F i)}.
Proof.
move=> Hge0.
elim: s=> // a []=> [_ _|b l /(_ is_true_true) [k /andP [Hks Hk]] _].
  by exists a; rewrite big_cons big_nil /maxr Hge0 mem_seq1 !eqxx.
rewrite big_cons /maxr.
set y := \big[_/_]_(_ <- _) _.
case: (y <= F a); [exists a| exists k].
  by rewrite mem_head eqxx.
by rewrite Hk mem_behead // Hks.
Qed.

Lemma bigmax_def (I : eqType) (F : I -> R) (s : seq I) x :
  (forall i, 0 <= F i) -> {i | (x == F i) && (i \in s) } ->
  (forall j, j \in s -> F j <= x) ->  
  \maxr_(k <- s) F k = x.
Proof. 
move=> Hge0 [] i /andP [] /eqP ->.
elim: s=> // a []=> [_|b l IHl].
  by rewrite mem_seq1 big_cons big_nil /maxr Hge0=> /eqP ->.
rewrite in_cons big_cons.
have [k /andP [Hks /eqP Hk]] := bigmax_eq (s := b :: l) Hge0 is_true_true.
case/orP=> [/eqP Hi|Hi] Hge; rewrite /maxr -Hk.
  by rewrite -Hi Hge // mem_behead.
have := Hge a; rewrite -IHl // =>[|j Hj].
  rewrite mem_head -Hk ler_eqVlt=> /(_ is_true_true)/orP []=> [/eqP ->|].
    by rewrite lerr.
  by rewrite (real_lerNgt (ger0_real (Hge0 k)) (ger0_real (Hge0 a)))=> ->.
by rewrite Hge // mem_behead.
Qed.

Lemma bigmax_distrl (I : eqType) (F : I -> R) (s : seq I) (r : R) :
  0 <= r -> (r * \maxr_(i <- s) (F i)) = \maxr_(i <- s) (r * F i).
Proof.
rewrite ler_eqVlt=> le0r.
elim: s=> [|a l IHl]; first by rewrite !big_nil mulr0.
rewrite !big_cons -IHl /maxr.
set x := F a; set y := \big[_/_]_(_ <- _) _.
case/orP: le0r=> [/eqP <-|lt0r]; first by rewrite !mul0r lerr.
by rewrite ler_pmul2l //; case: (y <= x).
Qed.


Lemma bigmax_ler_seq (I : eqType) (F1 F2 : I -> R) (s : seq I) :
 (forall i, 0 <= F1 i) ->  (forall i, i\in s -> F1 i <= F2 i) -> 
 \maxr_(i <- s) F1 i <= \maxr_(i <- s) F2 i.
Proof.
move=> Hge0.
elim: s=> [|a l IHl Hle]; first by rewrite !big_nil.
rewrite !big_cons /maxr.
have IHle i: i \in l -> F1 i <= F2 i by move=> H; rewrite Hle // mem_behead.  
have x1r := ger0_real (Hge0 a).
have x2r := ger0_real (ler_trans (Hge0 a) (Hle a (mem_head a l))).
have y1r := ger0_real (bigmax_pos l Hge0).
have y2r := ger0_real (ler_trans (bigmax_pos l Hge0) (IHl IHle)). 
case: (real_lerP y1r x1r)=>[_|_]; case: (real_lerP y2r x2r)=>[Hxy|Hyx].
  by rewrite Hle // mem_head.
  by rewrite ltrW // (ler_lt_trans (Hle a (mem_head a l)) Hyx).
  by rewrite (ler_trans (IHl IHle) Hxy).
exact: IHl.
Qed.

Lemma bigmax_ler (I : finType) (F1 F2 : I -> R) :
 (forall i, 0 <= F1 i) ->  (forall i,  F1 i <= F2 i) -> 
 \maxr_i F1 i <= \maxr_i F2 i.
Proof.
move=> Hge0 Hle.
elim: (index_enum I) => [|a l IHl]; first by rewrite !big_nil.
rewrite !big_cons /maxr.
have x1r := ger0_real (Hge0 a).
have x2r := ger0_real (ler_trans (Hge0 a) (Hle a)).
have y1r := ger0_real (bigmax_pos l Hge0).
have y2r := ger0_real (ler_trans (bigmax_pos l Hge0) IHl). 
case: (real_lerP y1r x1r)=>[_|_]; case: (real_lerP y2r x2r)=>[Hxy|Hyx] //.
  by rewrite ltrW // (ler_lt_trans (Hle a) Hyx).
by rewrite (ler_trans IHl Hxy).
Qed.
 
Lemma bigmax_ler_sum (I : finType) (F1 F2 : I -> R) :
  (forall i, 0 <= F1 i) -> (forall i, 0 <= F2 i) ->
  \maxr_i (F1 i + F2 i) <=  (\maxr_i F1 i) + \maxr_i F2 i.
Proof.
move=> H1ge0 H2ge0.
elim: (index_enum I)=>[|a l IHl]; first by rewrite !big_nil addr0.
rewrite !big_cons /maxr.
set x1 := F1 a; set y1 := \big[_/_]_(_ <- _) _.
set x2 := F2 a; set y2 := \big[_/_]_(_ <- _) _.
have x1r := ger0_real (H1ge0 a).
have x2r := ger0_real (H2ge0 a).
have y2r := ger0_real (bigmax_pos l H1ge0).
have zr := ger0_real (bigmax_pos l H2ge0).
have y1r : y1 \is Num.real.
  by apply/ger0_real/bigmax_pos=> i; rewrite addr_ge0.
have x12r : x1 + x2 \is Num.real by rewrite realD.
case: (real_lerP y1r x12r)=> [_|_]; case: (real_lerP y2r x1r)=>[c|d]; 
case: (real_lerP zr x2r)=> [g|h] //.
  +by rewrite ler_add2l ltrW.
  -by rewrite ler_add2r ltrW.
  +by rewrite ltrW // ltr_add.
  -by rewrite (ler_trans IHl) // ler_add.
  +by rewrite (ler_trans IHl) // ler_add2r.
  -by rewrite (ler_trans IHl) // ler_add2l.
Qed.

Lemma bigmax_le_prod2 (I J : finType) (F1 : I -> R) (F2 : J -> R) :
  (forall i, 0 <= F1 i) -> (forall j, 0 <= F2 j) ->
  \maxr_i \maxr_j (F1 i * F2 j) <= (\maxr_i F1 i) * \maxr_j F2 j.
Proof.
move=> H1ge0 H2ge0.
elim: (index_enum I)=>[|a l IHl]; first by rewrite !big_nil mul0r.
rewrite !big_cons -bigmax_distrl //.
set x := F1 a; set y1 := \big[_/_]_(_ <- _) F2 _.
set y2 := \big[_/_]_(_ <- _) \big[_/_]_ _ _; set z := \big[_/_]_(_ <- _) _.
have xr := ger0_real (H1ge0 a).
have zr : z \is Num.real := ger0_real (bigmax_pos l H1ge0).
have y1r : y1 \is Num.real.
  by apply/ger0_real/bigmax_pos=> i.
have y2r : y2 \is Num.real.
  by apply/ger0_real/bigmax_pos=> i; apply: bigmax_pos=> j; rewrite mulr_ge0.
have xy1r : x * y1 \is Num.real by rewrite realM.
rewrite /maxr.
case: (real_lerP y2r xy1r)=> [c'|d']; case: (real_lerP zr xr)=>[c|d] //. 
  by rewrite ler_pmul // ?H1ge0 ?bigmax_pos // ltrW.
by rewrite (ler_trans IHl) // ler_pmul // bigmax_pos.
Qed.


Lemma bigmax_le_prod (I : finType) (F1 F2 : I -> R) :
  (forall i, 0 <= F1 i) -> (forall i, 0 <= F2 i) ->
  \maxr_i (F1 i * F2 i) <=  (\maxr_i F1 i) * \maxr_i F2 i.
Proof.
move=> H1ge0 H2ge0.
apply: ler_trans (bigmax_le_prod2 H1ge0 H2ge0).
have Hpos i j: 0 <= F1 i * F2 j by rewrite mulr_ge0.
by apply: bigmax_ler=> i //; apply: (bigmax_ger (Hpos i)).
Qed.

Lemma bigmax_eq0 (I : eqType) (F : I -> R) (s : seq I) :
  (forall i, i \in s -> F i = 0) -> \maxr_(i <- s) F i = 0.
Proof.
elim: s=>[|a l IHl Heq0]; first by rewrite big_nil.
rewrite big_cons IHl ?Heq0 /maxr ?lerr ?mem_head // => i Hi.
by rewrite Heq0 // mem_behead.
Qed.

End bigmax.


Notation "\maxr_ ( i <- r | P ) F" :=
  (\big[Num.max/0%R]_(i <- r | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i <- r ) F" :=
  (\big[Num.max/0%R]_(i <- r) F%R) : ring_scope.
Notation "\maxr_ ( i | P ) F" :=
  (\big[Num.max/0%R]_(i | P%B) F%R) : ring_scope.
Notation "\maxr_ i F" :=
  (\big[Num.max/0%R]_i F%R) : ring_scope.
Notation "\maxr_ ( i : I | P ) F" :=
  (\big[Num.max/0%R]_(i : I | P%B) F%R) (only parsing) : ring_scope.
Notation "\maxr_ ( i : I ) F" :=
  (\big[Num.max/0%R]_(i : I) F%R) (only parsing) : ring_scope.
Notation "\maxr_ ( m <= i < n | P ) F" :=
 (\big[Num.max/0%R]_(m <= i < n | P%B) F%R) : ring_scope.
Notation "\maxr_ ( m <= i < n ) F" :=
 (\big[Num.max/0%R]_(m <= i < n) F%R) : ring_scope.
Notation "\maxr_ ( i < n | P ) F" :=
 (\big[Num.max/0%R]_(i < n | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i < n ) F" :=
 (\big[Num.max/0%R]_(i < n) F%R) : ring_scope.
Notation "\maxr_ ( i 'in' A | P ) F" :=
 (\big[Num.max/0%R]_(i in A | P%B) F%R) : ring_scope.
Notation "\maxr_ ( i 'in' A ) F" :=
 (\big[Num.max/0%R]_(i in A) F%R) : ring_scope.


Section bigmin.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.


Variable R : numDomainType.

Open Scope ring_scope.

Fixpoint min_seq (s : seq R) := 
  match s with 
   | [::] => 0
   | [:: x]  => x 
   | x :: s' => minr (min_seq s') x
  end.

Definition min_fT (I : finType) (F : I -> R) := min_seq (map F (index_enum I)).

Lemma min_seq_real (s : seq R) : (forall x, x \in s -> x \is Num.real) ->
 min_seq s \is Num.real.
Proof.
case: s=>[|a s]; first by rewrite real0.
elim: s a=>[a -> //|a l IHl b Hreal /=]; first by rewrite mem_head.
set c := (X in Num.min X _).
rewrite  /minr; case: (c <= b).
  by rewrite IHl // => x Hx; rewrite Hreal // mem_behead.
by rewrite Hreal // mem_head.
Qed.

Lemma min_seq_le (s : seq R) x : (forall x, x \in s -> x \is Num.real) -> 
  x \in s ->  min_seq s <= x.
Proof.
elim: s=> //= a []; first by rewrite in_cons orbF=> _ _ /eqP ->.
move=> b l IHl Hreal Hx.
have ar: a \is Num.real by rewrite Hreal // mem_head.
have mr: min_seq (b :: l) \is Num.real.
  by apply: min_seq_real=> y Hy; rewrite Hreal // mem_behead.
move: Hx.
rewrite in_cons /minr=>/orP []=>[/eqP ->|H]; case: (real_lerP mr ar)=> // [|Ha].
  by rewrite IHl // => y Hy; rewrite Hreal // mem_behead.
rewrite ltrW // (ltr_le_trans Ha) // IHl // => y Hy.
by rewrite Hreal // mem_behead.
Qed.

Lemma min_fT_le (I : finType) (F : I -> R) (i : I) :
 (forall i, F i \is Num.real) -> min_fT F <= F i.
Proof.
move=> Hreal; apply: min_seq_le=>[x /mapP [y Hy] -> //|].
by apply/mapP; exists i.
Qed.

Lemma ex_min_fT (I : finType) (F : I -> R) (a : I) : {k | min_fT F = F k}.
Proof.
rewrite /min_fT.
have := mem_index_enum a.
case: (index_enum I)=> // b l _. 
elim: l b=>[b|b l /= IHl c]; first by exists b.
have [k ->] := IHl b.
by rewrite /minr; case: (F k <= F c); [exists k|exists c].
Qed.

(* Lemma bigmax_pos (I : Type) (F : I -> R) (s : seq I) : *)
(*   (forall i, 0 <= F i) -> 0 <= \maxr_(i <- s) F i. *)
(* Proof. *)
(* move=> Hge0. *)
(* apply: (big_ind (fun x : R => 0 <= x))=> // x y Hx Hy. *)
(* by rewrite /maxr; case: (y <= x). *)
(* Qed. *)

(* Lemma bigmin_ger (I : eqType) (F : I -> R) (s : seq I) i : *)
(*  i \in s -> (\big[(@minr R)/(last 0%R (map [eta F] s))]_(i <- s) F i) <= F i. *)
(* Proof. *)
(* elim: s=> // a l IHl. *)
(* (* set x := F a; set y := \big[_/_]_(_ <- _) _. *) *)
(* (* have Hxr := ger0_real (Hge0 a). *) *)
(* (* have Hyr := ger0_real (bigmax_pos l Hge0). *) *)
(* rewrite in_cons big_cons (* /maxr *)=> /orP []=> [/eqP ->|/IHl Hi]. *)
(* Search _ minr. *)
(* case: (real_lerP Hyr Hxr)=> //. *)
(*   by move/ltrW. *)
(* exact: ler_trans. *)
(* Qed. *)

(* Lemma bigmax_eq (I : eqType) (F : I -> R) (s : seq I) : *)
(*   (forall i, 0 <= F i) ->  s != [::] ->  *)
(*   {k | F k = \maxr_(i <- s) F i}. *)
(* Proof. *)
(* move=> Hge0. *)
(* elim: s=> // a []=> [_ _|b l /(_ is_true_true) [k Hk] _]. *)
(*   by exists a; rewrite big_cons big_nil /maxr Hge0. *)
(* rewrite big_cons /maxr. *)
(* set y := \big[_/_]_(_ <- _) _. *)
(* by case: (y <= F a); [exists a| exists k]. *)
(* Qed. *)


(* Lemma bigmax_distrl (I : finType) (F : I -> R) (r : R) : *)
(*   0 <= r -> (r * \maxr_i (F i)) = \maxr_i (r * F i). *)
(* Proof. *)
(* rewrite ler_eqVlt=> le0r. *)
(* elim: (index_enum I)=> [|a l IHl]; first by rewrite !big_nil mulr0. *)
(* rewrite !big_cons -IHl /maxr. *)
(* set x := F a; set y := \big[_/_]_(_ <- _) _. *)
(* case/orP: le0r=> [/eqP <-|lt0r]; first by rewrite !mul0r lerr. *)
(* by rewrite ler_pmul2l //; case: (y <= x). *)
(* Qed. *)

(* Lemma bigmax_le (I : finType) (F1 F2 : I -> R) : *)
(*  (forall i, 0 <= F1 i) ->  (forall i, F1 i <= F2 i) ->  *)
(*  \maxr_i F1 i <= \maxr_i F2 i. *)
(* Proof. *)
(* move=> Hge0 Hle. *)
(* elim: (index_enum I)=> [|a l IHl]; first by rewrite !big_nil. *)
(* rewrite !big_cons /maxr. *)
(* have x1r := ger0_real (Hge0 a). *)
(* have x2r := ger0_real (ler_trans (Hge0 a) (Hle a)). *)
(* have y1r := ger0_real (bigmax_pos l Hge0). *)
(* have y2r := ger0_real (ler_trans (bigmax_pos l Hge0) IHl).  *)
(* case: (real_lerP y1r x1r)=>[_|_]; case: (real_lerP y2r x2r)=>[Hxy|Hyx] //. *)
(*   by rewrite ltrW // (ler_lt_trans (Hle a) Hyx). *)
(* by rewrite (ler_trans IHl Hxy). *)
(* Qed. *)

(* Lemma bigmax_ler_sum (I : finType) (F1 F2 : I -> R) : *)
(*   (forall i, 0 <= F1 i) -> (forall i, 0 <= F2 i) -> *)
(*   \maxr_i (F1 i + F2 i) <=  (\maxr_i F1 i) + \maxr_i F2 i. *)
(* Proof. *)
(* move=> H1ge0 H2ge0. *)
(* elim: (index_enum I)=>[|a l IHl]; first by rewrite !big_nil addr0. *)
(* rewrite !big_cons /maxr. *)
(* set x1 := F1 a; set y1 := \big[_/_]_(_ <- _) _. *)
(* set x2 := F2 a; set y2 := \big[_/_]_(_ <- _) _. *)
(* have x1r := ger0_real (H1ge0 a). *)
(* have x2r := ger0_real (H2ge0 a). *)
(* have y2r := ger0_real (bigmax_pos l H1ge0). *)
(* have zr := ger0_real (bigmax_pos l H2ge0). *)
(* have y1r : y1 \is Num.real. *)
(*   by apply/ger0_real/bigmax_pos=> i; rewrite addr_ge0. *)
(* have x12r : x1 + x2 \is Num.real by rewrite realD. *)
(* case: (real_lerP y1r x12r)=> [_|_]; case: (real_lerP y2r x1r)=>[c|d];  *)
(* case: (real_lerP zr x2r)=> [g|h] //. *)
(*   +by rewrite ler_add2l ltrW. *)
(*   -by rewrite ler_add2r ltrW. *)
(*   +by rewrite ltrW // ltr_add. *)
(*   -by rewrite (ler_trans IHl) // ler_add. *)
(*   +by rewrite (ler_trans IHl) // ler_add2r. *)
(*   -by rewrite (ler_trans IHl) // ler_add2l. *)
(* Qed. *)

(* Lemma bigmax_le_prod2 (I J : finType) (F1 : I -> R) (F2 : J -> R) : *)
(*   (forall i, 0 <= F1 i) -> (forall j, 0 <= F2 j) -> *)
(*   \maxr_i \maxr_j (F1 i * F2 j) <= (\maxr_i F1 i) * \maxr_j F2 j. *)
(* Proof. *)
(* move=> H1ge0 H2ge0. *)
(* elim: (index_enum I)=>[|a l IHl]; first by rewrite !big_nil mul0r. *)
(* rewrite !big_cons -bigmax_distrl //. *)
(* set x := F1 a; set y1 := \big[_/_]_(_ <- _) _. *)
(* set y2 := \big[_/_]_(_ <- _) _; set z := \big[_/_]_(_ <- _) _. *)
(* have xr := ger0_real (H1ge0 a). *)
(* have y2r := ger0_real (bigmax_pos l H1ge0). *)
(* have zr := ger0_real (bigmax_pos (index_enum _) H2ge0). *)
(* have y1r : y1 \is Num.real. *)
(*   by apply/ger0_real/bigmax_pos=> i; apply: bigmax_pos=> j; rewrite mulr_ge0. *)
(* have xzr : x * z \is Num.real by rewrite realM. *)
(* rewrite /maxr. *)
(* case: (real_lerP y1r xzr)=> [_|_]; case: (real_lerP y2r xr)=>[c|d] //.  *)
(*   by rewrite ler_pmul // ?H1ge0 ?bigmax_pos // ltrW. *)
(*   by rewrite (ler_trans IHl) // ler_pmul // bigmax_pos. *)
(* Qed. *)


(* Lemma bigmax_le_prod (I : finType) (F1 F2 : I -> R) : *)
(*   (forall i, 0 <= F1 i) -> (forall i, 0 <= F2 i) -> *)
(*   \maxr_i (F1 i * F2 i) <=  (\maxr_i F1 i) * \maxr_i F2 i. *)
(* Proof. *)
(* move=> H1ge0 H2ge0. *)
(* apply: ler_trans (bigmax_le_prod2 H1ge0 H2ge0). *)
(* have Hpos i j: 0 <= F1 i * F2 j by rewrite mulr_ge0. *)
(* by apply: bigmax_le=> i //; apply: (bigmax_ger (Hpos i)). *)
(* Qed. *)

End bigmin.


Section outil.
 
Open Scope R_scope.


(* Lemma utile : forall r,  0 <= r -> r<>0 -> 0 < r. *)
(* Proof. *)
(* move=> r Hpos Hnnul. *)
(* case: (Rdichotomy _ _ Hnnul) => Habs //. *)
(* by have : ~ 0 <= r by apply: Rlt_not_le. *)
(* Qed. *)
  

Lemma big_max_exists_2:
  forall (T : eqType) (r : seq T) (F: T -> R), r != [::] -> 
  (forall i, 0 <= F i)-> 
  {i0 |i0 \in r /\ \big[Rmax/0]_(i <- r) F i = F i0}.
Proof.
move=> T r F Hr H.
move: Hr; elim: r => // i0 s Ih _.
case Hs: (s != [::]).
  elim: (Ih Hs) => i1 [Hi0 Hi1].
  elim: (Rle_dec (F i0) (F i1)) => Hdec.
    exists i1; split.
      by rewrite in_cons; rewrite Hi0; apply: orbT.
    by rewrite big_cons Hi1 Rmax_right.
  exists i0; split; 
   [apply: mem_head 
   | rewrite big_cons Hi1 Rmax_left => //].
  by left; apply: Rnot_le_gt.
exists i0; rewrite big_cons.
by move/eqP: Hs => ->; split; 
 [apply: mem_head 
 | rewrite big_nil; apply: Rmax_left; apply: (H i0)].
Qed.
 
 
Lemma big_max_le2:
  forall (T : eqType) (r : seq T) (F : T -> R) i0, 
  i0 \in r -> F i0 <= \big[Rmax/0]_(i <- r) F i.
Proof.
move=> T r F i0 Hi0.
have Hnul: r != [::] 
          by apply/eqP=> Hr; rewrite Hr in Hi0; have Hr2:= @in_nil _ i0;
           rewrite Hi0 in Hr2.
move: Hnul Hi0; elim: r => // i s IHs _ Hri0.
rewrite big_cons; case Hi: (i0 == i).
  by move/eqP: Hi => ->; apply RmaxLess1.
apply: (Rle_trans _ (\big[Rmax/0]_(j <- s) F j)); last by apply RmaxLess2.
rewrite in_cons Hi orFb in Hri0.
have Hnul: s != [::] 
          by apply/eqP=> Hr; rewrite Hr in Hri0; have Hr2:= @in_nil _ i0;
           rewrite Hri0 in Hr2.
by apply IHs.
Qed.

End outil.


Section can_norm.

Open Scope ring_scope.
Delimit Scope R_scope with Re.


Definition sum_abs m n (A : 'M_(m, n)) i : R := \sum_j (Rabs (A i j)).
Definition norm_m m n (A : 'M_(m, n)) : R:= \big[Rmax/R0]_i (sum_abs A i).

Lemma sum_abs_pos: forall m n (A : 'M_(m,n)) i, (0 <= sum_abs A i)%Re.
Proof.
move =>A i; rewrite /sum_abs; elim: index_enum.
  by move=> A0 i0; rewrite big_nil; right.
by move => x s IH A0 i0; rewrite big_cons; apply: Rplus_le_le_0_compat; 
 [apply: Rabs_pos
 |].
Qed.

Lemma eq_sum_abs:
forall m n (A :'M_(m,n)) i, sum_abs A i = Rabs (sum_abs A i).
Proof.
by move => *; rewrite Rabs_right //; apply: Rle_ge; apply: sum_abs_pos.
Qed.

Lemma abs_sum_le_f:
forall p (f : 'I_p ->R) , (Rabs (\sum_i (f i)) <= \sum_i (Rabs (f i)))%Re. 
Proof.
move => p v; elim: index_enum.
  by rewrite !big_nil Rabs_R0; right.
by move => x s IH; rewrite !big_cons; 
 apply: (Rle_trans _ _ _ (Rabs_triang _ _)); apply: Rplus_le_compat_l.
Qed.


(* Lemma normM_pos: *)
(* forall m n (A : 'M_(m,n)), (0 <= norm_m A)%Re. *)
(* Proof. *)
(* move => A; rewrite /norm_m; elim: index_enum. *)
(*   by move=> A0 i0; rewrite big_nil; right. *)
(* by move=> t s IH A0 i0; rewrite big_cons; apply: Rmax_ge;  *)
(*  [apply: sum_abs_pos *)
(*  |]. *)
(* Qed. *)

(* Lemma desc_norm_m:forall m n (A : 'M_(m,n)) r, norm_m A = Rmax (sum_abs A r) *)
(*     (\big[Rmax/R0]_(k<m|k!=r) (sum_abs A k)). *)
(* Proof. *)
(* by move=>m n A r; rewrite /norm_m; have H:= sum_abs_pos A; apply: sxbigD1.  *)
(* Qed. *)

(* Lemma normM_max: *)
(* forall m n (A: 'M_(m,n)) i, sum_abs A i <= norm_m A.  *)
(* Proof. *)
(* by move => m n A i; rewrite (desc_norm_m A i); apply: RmaxLess1. *)
(* Qed. *)

(* Lemma ineq_norm_m: *)
(* forall m n p (A : 'M_(m, n))  (B : 'M_(n, p)), *)
(* norm_m (A *m B) <= norm_m A * norm_m B. *)
(* Proof. *)
(* move => m n p A B; rewrite {1}/norm_m /mulmx /sum_abs. *)
(* have ->: \big[Rmax/0]_i (\sum_j Rabs  *)
(*            ((\matrix[mulmx_key]_(i0, k) (\sum_j0 A i0 j0 * B j0 k)) i j)) =  *)
(*            \big[Rmax/0]_i (\sum_j Rabs (\sum_s A i s*B s j)) *)
(*         by apply eq_bigr => i _; apply eq_bigr => j _; rewrite mxE. *)
(* apply: (Rle_trans _ (\big[Rmax/0]_i (\sum_j (\sum_s (Rabs (A i s * B s j)))))). *)
(*   by apply: (Rmax_le_f_le) => i; apply: (sum_le_f_le) => j; apply: abs_sum_le_f. *)
(* have ->: \big[Rmax/0]_i (\sum_j \sum_s Rabs (A i s * B s j))  =  *)
(*                \big[Rmax/0]_i (\sum_j \sum_s Rabs (A i s )*Rabs( B s j)) *)
(*         by apply: eq_bigr =>i _; apply eq_bigr => j _;apply eq_bigr => s _; *)
(*          apply Rabs_mult. *)
(* have ->: \big[Rmax/0]_i (\sum_j \sum_s Rabs (A i s) * Rabs (B s j)) = *)
(*             \big[Rmax/0]_i (\sum_s Rabs (A i s)*\sum_j  Rabs (B s j)) . *)
(*   have ->: \big[Rmax/0]_i (\sum_j \sum_s Rabs (A i s) * Rabs (B s j)) = *)
(*              \big[Rmax/0]_i (\sum_s \sum_j Rabs (A i s) * Rabs (B s j)) *)
(*           by apply eq_bigr => i _; apply: exchange_big. *)
(*   by apply eq_bigr => i _; apply eq_bigr => s _; rewrite -big_distrr //=. *)
(* apply: (Rle_trans _ (\big[Rmax/0]_i (\sum_s Rabs (A i s) *norm_m B))). *)
(*   apply: (Rmax_le_f_le ) => i; apply: (sum_le_f_le ) => s. *)
(*   apply: Rmult_le_compat_l; first by apply: Rabs_pos. *)
(*   by apply: normM_max. *)
(* have ->: \big[Rmax/0]_i (\sum_s Rabs (A i s) * norm_m B)  =  *)
(*            (\big[Rmax/0]_i (\sum_s Rabs (A i s) ))* norm_m B. *)
(*   rewrite -> Rmult_comm. *)
(*   have ->: \big[Rmax/0]_i (\sum_s Rabs (A i s) * norm_m B) = *)
(*               \big[Rmax/0]_i (\sum_s norm_m B *Rabs (A i s) ) *)
(*           by apply eq_bigr => i _; apply eq_bigr => s _;  *)
(*            rewrite /(@GRing.mul) /= Rmult_comm. *)
(*   rewrite distr_mult_max. *)
(*     by apply eq_bigr => i _; rewrite -big_distrr. *)
(*   by apply normM_pos. *)
(* have ->: \big[Rmax/0]_i (\sum_s Rabs (A i s)) = norm_m A  *)
(*         by rewrite /norm_m /sum_abs. *)
(* by apply Rle_refl. *)
(* Qed. *)

(* Lemma normM_plus_le: *)
(* forall m n (A B : 'M_(m, n)), norm_m (A + B) <= norm_m A+ norm_m B. *)
(* Proof. *)
(* move => m n A B; rewrite {1}/norm_m /addmx /sum_abs. *)
(* have ->: \big[Rmax/0]_i (\sum_j Rabs ((A + B)%Ri i j)) *)
(*            = \big[Rmax/0]_i (\sum_j Rabs (A i j + B i j)) *)
(*         by apply:eq_bigr => i _; apply:eq_bigr => j _; rewrite mxE. *)
(* apply: (Rle_trans _ (\big[Rmax/0]_i (\sum_j (Rabs (A i j) +Rabs( B i j))))). *)
(*   by apply:(Rmax_le_f_le )=> i; apply:(sum_le_f_le ) => s; apply:Rabs_triang. *)
(* have ->: \big[Rmax/0]_i (\sum_j (Rabs (A i j) + Rabs (B i j))) = *)
(*            \big[Rmax/0]_i (\sum_j Rabs (A i j) +\sum_j Rabs (B i j)) *)
(*         by apply eq_bigr => i _; rewrite big_split //=. *)
(* apply: (Rle_trans _ (\big[Rmax/0]_i (\sum_j Rabs (A i j)) + *)
(*                                       \big[Rmax/0]_i ( \sum_j Rabs (B i j)) )). *)
(*   by apply: (distr_plus_max ) => i; apply: sum_abs_pos. *)
(* by rewrite /norm_m /sum_abs; apply: Rle_refl. *)
(* Qed. *)

(* Lemma hom_pos_normM: *)
(* forall m n (M : 'M_(m,n)) a, norm_m (a *: M) = Rabs a * norm_m M. *)
(* Proof. *)
(* move => m n v a; rewrite /norm_m; elim: index_enum. *)
(*   by rewrite !big_nil mulr0. *)
(* move=> t s IH; rewrite !big_cons IH. *)
(* rewrite <- RmaxRmult; [| apply: Rabs_pos]. *)
(* set f:= (fun x => Rmax x ((Rabs a * \big[Rmax/0]_(i <- s) sum_abs v i))). *)
(* apply: (f_equal f). *)
(* have -> : ((Rabs a * sum_abs v t)%Re) = ((Rabs a * sum_abs v t)%Ri) by toR. *)
(* by rewrite /sum_abs big_distrr /=; apply: eq_bigr => i _; *)
(*  rewrite mxE Rabs_mult; toR. *)
(* Qed. *)


(* Lemma canonical1: *)
(*  forall m n (A:'M_(m, n)) i j, Rabs (A i j) <= norm_m A. *)
(* Proof. *)
(* move => m n A i j; rewrite /norm_m. *)
(* apply: (Rle_trans _ (sum_abs A i)); [|apply: max_max; apply: sum_abs_pos]. *)
(* rewrite /sum_abs (bigD1 j ) //= -{1}(Rplus_0_r (Rabs (A i j))). *)
(* apply: Rplus_le_compat_l; elim: index_enum. *)
(*   by rewrite big_nil; right. *)
(* move=> x s IH; rewrite !big_cons. *)
(* case C: (x!=j) => //. *)
(* by rewrite -{1}(Rplus_0_r R0); apply: Rplus_le_compat; *)
(*  [apply Rabs_pos *)
(*  |]. *)
(* Qed. *)


(* Lemma canonical2: forall m n (A B:'M_(m ,n)),  *)
(* (forall i j, Rabs (A i j) <= Rabs (B i j)) ->norm_m A <= norm_m B. *)
(* Proof. *)
(* by move=> m n A B H;rewrite /norm_m; apply: (Rmax_le_f_le ) => i;  *)
(*  rewrite /sum_abs; apply: (sum_le_f_le ) => j. *)
(* Qed. *)


(* Canonical Structure the_norm := *)
(* MNorm  normM_pos hom_pos_normM normM_plus_le ineq_norm_m canonical1 canonical2. *)


 
(* Lemma Rabs_0 : forall x, Rabs x = 0 <-> x = 0. *)
(* Proof. *)
(* move=> x; split => H. *)
(*   case (Rcase_abs x)=> H0. *)
(*     have H1 : Rabs x = - x by apply: Rabs_left. *)
(*     by rewrite H in H1; apply/eqP; rewrite -oppr_eq0; apply/eqP. *)
(*   by rewrite -H; apply: sym_eq; apply: Rabs_right. *)
(* by rewrite H; apply: Rabs_R0. *)
(* Qed. *)


(* Lemma norm_0 : forall m n (A : 'M_(m ,n)) (mnorm : can_norm),  *)
(*   anorm mnorm A = 0 <-> A = 0. *)
(* Proof. *)
(* move=> m n A mnorm. *)
(* split=> H. *)
(*   apply/matrixP=> i j; rewrite mxE. *)
(*   have Hr : Rabs (A i j) = 0  *)
(*           by apply: Rle_antisym; *)
(*            [rewrite -H; apply: canonical_norm *)
(*            | apply: Rabs_pos]. *)
(*   by apply/Rabs_0. *)
(* by rewrite H mynorm0. *)
(* Qed. *)



End can_norm.



(* 
*** Local Variables: ***
*** coq-prog-name: "coqtop" ***
*** coq-prog-args: ("-emacs-U" "-I" "../../coqfinitgroup" "-I" "../../coqfinitgroup/ssreflect/ssreflect1.3_v8.4/src/") ***
*** End: ***
 *)