Require Import Reals.
Require Import ssrmatching ssreflect ssrfun ssrbool eqtype ssrnat seq.
Require Import fintype finfun bigop ssralg poly matrix mxpoly complex.
Require Import ssrnum mxalgebra polyrcf polyorder realalg zmodp. 
Require Import closed_poly mxorder mxtop struct_top outils.

Require Import Rstruct.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.


Section preuve.

Variable R : rcfType.
Local Notation C := (complex R).

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).

 
(* Definition eigenvalueC n (M : 'M[R]_n) := eigenvalue M^f. *)

Definition eigen_seq n (M : 'M[C]_n) := root_seq (char_poly M).

Definition spectral_radius n (M : 'M[R]_n) :=
  Re (\big[(@maxr _)/0]_(lam <- eigen_seq M^%:C) `|lam|).

Notation "\rho A" := (spectral_radius A) (at level 0).

Lemma eigenE n (M : 'M[R]_n) lam : 
 (lam \in eigenvalue M^%:C) = (lam \in eigen_seq M^%:C).
Proof.
rewrite -topredE /= eigenvalue_root_char.
by rewrite -root_root_seq // -size_poly_eq0 size_char_poly.
Qed.

Lemma eigen_seq_nnil n (M : 'M[C]_n.+1) : eigen_seq M != [::].
Proof. by rewrite -size_eq0 size_root_seq size_char_poly. Qed.

Lemma Im0 (a : C) : a \is Num.real -> Im a = 0.
Proof.
rewrite realE=> /orP []; first exact: ger0_Im.
by rewrite lecE=> /andP [] /eqP <-.
Qed.

Lemma real_Re_eq0 (x : C) : x \is Num.real -> (Re x == 0) = (x == 0).
Proof. by rewrite eq_complex=> /Im0 ->; rewrite eqxx andbT. Qed.


Lemma rho0 n : \rho (0 : 'M_n) = 0.
Proof.
rewrite /spectral_radius; apply/eqP; rewrite real_Re_eq0.
  apply/eqP/bigmax_eq0=> i.
  rewrite -eigenE=> /eigenvalueP [x] /eqP.
  rewrite map_mx0 mulmx0 eq_sym scaler_eq0=> /orP []=>[/eqP ->|->] //.
  by rewrite normr0.
by apply/ger0_real/bigmax_pos.
Qed.

Lemma Re_ge0 (a : C) : a \is Num.real -> (0 <= Re a) = (0 <= a).
Proof. by move/Im0=> Ha; rewrite lecE Ha eqxx. Qed.

Lemma rho_ge0 n (M : 'M[R]_n) : 0 <= \rho (M).
Proof.
rewrite /spectral_radius /eigen_seq.
case: n M=>[M|n M].
  have/eqP ->: root_seq (char_poly M^%:C) == [::].
    by rewrite -root_seq_nil size_char_poly.
  by rewrite big_nil.
have Hnv: root_seq (char_poly M^%:C) != [::].
    by rewrite -root_seq_nil size_char_poly.
have [k /andP [_ /eqP <-]]:= bigmax_eq (@normr_ge0 _) Hnv.
by rewrite Re_ge0.
Qed.



Lemma ReM (a b : C) : Re (a * b) = Re a * Re b - Im a * Im b.
Proof. by case: a b=> Ra Ia [] Rb Ib. Qed.

(* Lemma ReK (a : C) : a \is Num.real -> (Re a)%:C = a. *)
(* Proof. by move=>/Im0 Ha; apply/eqP; rewrite eq_complex Ha !eqxx. Qed. *)

(* Lemma RecJ (a : C) : Re a^* = Re a. *)
(* Proof. by case: a. Qed. *)
 (*************************************)
Lemma normc_real (x : R) : `|x%:C| = `|x|%:C.
Proof.
by rewrite normcE sqrtc_sqrtr ?mulcJ_ge0 //= mul0r subr0 -expr2 sqrtr_sqr.
Qed.

Lemma rho_scalemx n a (A :'M[R]_n) :  \rho (a *: A) = `|a| * \rho A.
Proof.
case: n A => [A |n A].
  have ->: A = 0 by apply/matrixP=> [][].
  by rewrite scaler0 rho0 mulr0.
case: (altP (a=P0))=> [->|Hanz].
  by rewrite scale0r normr0 rho0 mul0r.
have Hanzc : a%:C != 0 by rewrite fmorph_eq0.
rewrite /spectral_radius.
have [k /andP[]]:= bigmax_eq (@normr_ge0 _) (eigen_seq_nnil A^%:C).
rewrite -eigenE=> /eigenvalueP [v Hv Hvneq0] /eqP Hk.
have ->:  \maxr_(lam <- eigen_seq (a *: A)^%:C) `|lam| =
        `|a|%:C * \maxr_(lam <- eigen_seq A^%:C) `|lam|.
  apply/eqP. rewrite eqr_le ; apply/andP; split; last first.
    have: a%:C * k \in eigenvalue (map_mx (real_complex R) (a *: A)).
    by apply/eigenvalueP; exists v=> //; rewrite map_mxZ -scalemxAr Hv scalerA.
    rewrite eigenE=> /(bigmax_ger (@normr_ge0 _)).
    by rewrite normrM normc_real -Hk.  
  have [k' /andP[]]:= bigmax_eq (@normr_ge0 _) (eigen_seq_nnil (a *: A)^%:C).
  rewrite -eigenE=> /eigenvalueP [v' Hv' Hv'neq0] /eqP <-.
  have: a^-1%:C * k' \in eigenvalue (map_mx (real_complex R) A).
    apply/eigenvalueP; exists v'=> //;  apply: (scalerI Hanzc). 
    rewrite scalemxAr -map_mxZ Hv' scalerA mulrA -rmorphM mulrV ?mul1r //.
    by rewrite unitfE. 
  rewrite eigenE => /(bigmax_ger (@normr_ge0 _)). 
  move/(ler_pmul (normr_ge0 _) (normr_ge0 _) (lerr `|a%:C|)) .
  rewrite normrM mulrA -normrM -[X in `|X|]rmorphM mulrV ?unitfE //.
  by rewrite normr1 mul1r normc_real.
by rewrite -Hk ReM /= mulr0 subr0.
Qed.

Lemma mxreal_gt0 m n (A : 'M[R]_(m,n)) : 0 <m: A <-> 0 <m: A^%:C.
Proof. by split=> Hpos i j; have/(_ i j) := Hpos; rewrite !mxE ltcR. Qed.

Lemma mxreal_ge0 m n (A : 'M[R]_(m,n)) : 0 <=m: A <-> 0 <=m: A^%:C.
Proof. by split=> Hpos i j; have/(_ i j) := Hpos; rewrite !mxE lecR. Qed.

End preuve.

Notation "\rho A" := (spectral_radius A) (at level 0).

Section perron_thm.

(* Variable R : rcfType. *)
Local Notation C := (complex R).

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Local Notation "A ^%:C" := (map_mx (real_complex_def (Phant R)) A) (at level 2).

Lemma mx_cvg0P n (A :'M[R]_n.+1) : 
  reflect ((fun k => A^+k) >->> 0) (\rho A < 1).
Proof. admit. Qed.

Local Notation N := (can_norm8 realArchiFieldType) .
Local Notation posD := (posDomain realArchiFieldType).

Lemma pos_const_norm m n (A : 'M[R]_(m,n)) (r : posD) :
  0 < (sort r) / ( ||A : N| + 1).
Proof.
by rewrite mulr_gt0 ?lt_pos // invr_gt0 -[0]addr0 ler_lt_add // pos_norm.
Qed.

Lemma continu_mulr : forall m n p (A : 'M[R]_(m,n)), 
   continue (fun (B : 'M[R]_(n,p)) => A *m B).
Proof.
move=> m n p A; apply/continu_eq=> B r.
pose r2 := (PosD (pos_const_norm A r)).
exists r2=> C HC.
rewrite /open_ball /= /dist_mat -mulmxBr.
apply: (ler_lt_trans (prod_ineq_norm _ A (B - C))).
apply: (ler_lt_trans (y:= ||A : N| * (sort r2))). 
  by rewrite ler_wpmul2l ?pos_norm // ltrW.
rewrite /r2 /=.
rewrite mulrC -mulrA gtr_pmulr ?lt_pos //.
have Hgt0 : 0< (norm8 A + 1).
  by rewrite -[0]addr0 ler_lt_add ?ltr01 // pos_norm.  
rewrite -(ltr_pmul2l Hgt0) mulrA mulrV ?mul1r ?mulr1 ?unitf_gt0 //.
by rewrite -[X in X < _]addr0 ltr_add2l ltr01.
Qed.

Lemma continu_mull : forall m n p (A : 'M[R]_(n,p)),
    continue (fun (B : 'M[R]_(m,n)) => B *m A ). 
Proof.
move=> m n p A; apply/continu_eq=> B r.
pose r2 := (PosD (pos_const_norm A r)).
exists r2=> C HC.
rewrite /open_ball /= /dist_mat -mulmxBl.
apply: (ler_lt_trans (prod_ineq_norm _ (B - C) A)).
apply: (ler_lt_trans (y:= (sort r2) * ||A : N|)). 
  by rewrite ler_wpmul2r ?pos_norm // ltrW.
rewrite /r2 /=.
rewrite -mulrA gtr_pmulr ?lt_pos //.
have Hgt0 : 0< (norm8 A + 1).
  by rewrite -[0]addr0 ler_lt_add ?ltr01 // pos_norm.  
rewrite -(ltr_pmul2l Hgt0) mulrA mulrV ?mul1r ?mulr1 ?unitf_gt0 //.
by rewrite -[X in X < _]addr0 ltr_add2l ltr01.
Qed.

Lemma Perron_gt0 n (A : 'M[R]_n) (x : 'cV_n) (lam : C) : 0 <m: A ->
  x != 0 -> (A *m x)^%:C = lam *: x^%:C -> `|lam| = (\rho (A))%:C -> 
  [/\ A *m Mabs x = \rho (A) *: Mabs x, 0 <m: Mabs x & 0 < \rho A].
Proof.
case: n A x => [A x|n A x HAgt0 Hxneq0 Hegv Hlam].
  have ->: x = 0 by apply/matrixP=> [][].
  by rewrite eqxx.
have/subr_Mge0 := MabsM A^%:C x^%:C. 
rewrite !Mabs_real_complex -!map_mxM Hegv Mabs_scalemx Hlam.
rewrite [Mabs A]eq_Mabs_id; last exact: MltW.
rewrite Mabs_real_complex -map_mxZ -map_mx_sub.
set y := (X in _ <=m: X^%:C)=> /mxreal_ge0 Hge0.
set z := _ *m Mabs _.
have Hzgt0 : 0 <m: z.
  by apply (mulmx_gt0 HAgt0 (Mabs_ge0 _)); rewrite Mabs_eq0.
case: (altP (y =P 0))=> [/eqP Hy|Hy].
 have:= rho_ge0 A; rewrite ler_eqVlt /y.
  case/orP=> [/eqP Hrho0|rhogt0].
    rewrite /y -Hrho0 scale0r subr0 in Hy. 
    rewrite /z (eqP Hy) in Hzgt0.
    by have:= Hzgt0 0 0; rewrite ltrr.
  rewrite /y subr_eq0 in Hy. 
  split=> // [|i j]; first by apply/eqP.
  rewrite !mxE ltr_neqAle normr_ge0 andbT.
  apply/eqP=> Habs.
  have/eqP/matrixP/(_ i j) := Hy.
  rewrite ![in X in _ = X]mxE -Habs mulr0=> H0.
  by have/(_ i j):= Hzgt0; rewrite mxE /z H0 ltrr.
have := mulmx_gt0 HAgt0 Hge0 Hy.
rewrite /y mulmxBr -scalemxAr=> Hpos. 
have/subr_Mgt := Mle_eps Hpos Hzgt0.
set eps := min_fT _.
have Hepsgt0 : 0 < eps.
  rewrite /eps;   set F := (fun i => _).
  have [k ->] := ex_min_fT F ord0.
  have/(_ k 0):= Hzgt0; rewrite mxE=> Hzk0.
  have/(_ k 0):= Hpos; rewrite mxE=> HAz0.
  by rewrite !mulr_gt0 // invr_gt0 // ltr0n.
have Hrgt0 : 0 < (\rho (A) + eps)^-1.
  by rewrite invr_gt0 -[0]addr0 ler_lt_add // rho_ge0.
rewrite -scalerDl=> /(Mlt_scaler Hrgt0).
have:= Hrgt0; rewrite lt0r -unitfE unitrV=> /andP [Hru _].
rewrite scalerA mulVr // scale1r scalemxAl.
set B := _ *: _=> HB.
have := Hrgt0; rewrite invr_gt0=> Hrgt02.
have/mx_cvg0P rBlt1 : \rho (B) < 1.
  rewrite rho_scalemx ?ltrW // gtr0_norm //. 
  rewrite -(ltr_pmul2l Hrgt02) ?mulrA mulrV // ?mulr0 ?ltr01 //.
  by rewrite mulr1 mul1r ltr_addl.
have HBltz k : z <m: B^+k.+1 *m z.
  have Htr m : B^+m *m z <m: B^+m.+1 *m z.
    elim: m=>[|m IHm]; first by rewrite mul1mx.
    rewrite 2!exprS -!mulmxA; apply: (Mlt_pmul2l _ IHm).
    move=> i j; rewrite !mxE mulr_gt0 ?invr_gt0 //.
    by have/(_ i j):= HAgt0; rewrite !mxE.
  by elim: k=> // k IHk; apply/(Mlt_trans IHk).
have/continue_cvg Hcvgm := (continu_mull (m:= n.+1) (A:= z)).
have/Hcvgm : (fun k : nat => B ^+ k.+1) >->> 0.
  have Hg: forall m n, (m < n)%N -> (m.+1 < n.+1)%N by [].
  pose g := mkss Hg.
  exact: (cvg_ss_cvg g rBlt1).
rewrite mul0mx=> Hcv0.
have Habs := ltmx_cvg_lemx HBltz Hcv0.
have: 0 < z 0 0 <= 0.
  have/(_ 0 0):= Hzgt0; rewrite mxE=> ->.
  by have/(_ 0 0):= Habs; rewrite [X in _ <= X]mxE=> ->.
by rewrite lter_anti.
Qed.

End perron_thm.