Require Import Classical_Prop IndefiniteDescription Classical_Pred_Type Reals.
Require Import ssreflect ssrbool eqtype ssrfun ssrnat seq fintype bigop ssralg.
Require Import ssrnum.
Require Import Rstruct outils set_theory.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.



Module Topology.

 Structure mixin_of (T_Space : Type) : Type := Mixin
  {
    open : (T_Space -> Prop) -> Prop;
    _ : open {ø : T_Space};
    _ : open {full : T_Space};
    _ : forall (E : (T_Space -> Prop) -> Prop), E <s open -> 
          open (union_s E id);
    _ : forall A B, open A -> open B -> open (inter A B);
    _ : forall A B, A =s B -> (open A <-> open B)
  }.

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure type := Pack {sort; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
Definition pack c := @Pack T c T.
Definition clone := fun c  & cT -> T & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.

Coercion sort : type >-> Sortclass.

Notation topologyType := type.
Notation TopologyMixin := Mixin.
Notation TopologyType T m := (@pack T m).
Notation " [ 'topologyType' 'of' T ] " := (@clone T _ _ id idfun)
  (at level 0, format "[ 'topologyType' 'of' T ]") : form_scope.
Notation "[ 'topologyType' 'of' T 'for' C ]" := (@clone T C _  id idfun)
  (at level 0, format "[ 'topologyType' 'of' T 'for' C ]") : form_scope.


End Exports.
End Topology.
Export Topology.Exports.

Section TopologyTheory.

Variable T : topologyType.

Canonical Structure Top_eqType := Eval hnf in EqType T (T_eqMixin T).


Definition open := Topology.open (Topology.class T).

(* Definition id_s := @idfun (T ->Prop). *)

Add Morphism open with signature (@eqset T) ==> iff as open_mor.
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma empty_in_open : open {ø : T}.
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma full_in_open : open {full : T}.
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma union_open : forall (E : (T -> Prop) -> Prop), E <s open -> 
      open (union_s E (@id (T ->Prop))).
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma inter_open : forall A B, open A -> open B -> open (inter A B).
Proof. by rewrite /open; case: T => ? []. Qed.

Lemma union_open2 (E: (T -> Prop) -> Prop) F : 
  (forall A, A \ins E -> open (F A)) -> open (union_s E F). 
Proof.
move=> H; rewrite eq_union; apply: union_open=> S.
by case=> A [/ H HA1 HA2]; rewrite -HA2. 
Qed.

Definition interior (A : T -> Prop) := 
  union_s (fun B => open B /\ B <s A) id. 

Notation "A °" := (interior A) (at level 0, no associativity).

Lemma int_go A B : open A -> A <s B -> A <s B°.
Proof. by move=> Ho HAB x Hx; exists A. Qed.

Lemma open_int A : open A°.
Proof. by apply: union_open=> X []. Qed.

Lemma int_in A : A° <s A.
Proof. by move=> x [] E [[H1 H2] /H2 H3]. Qed.

Add Morphism interior with signature (@eqset T) ==> (@eqset T)
  as interior_mor.
Proof.
move=> A B HAB; apply/double_i; apply/int_go; try exact: open_int.
  by rewrite -HAB; apply: int_in.
by rewrite HAB; apply: int_in.
Qed.

Lemma int_open_eq A : open A <-> A <s A°.  
Proof.
split=> H; first exact: int_go. 
have -> : A =s A° by apply: double_i => //; apply: int_in.
exact: open_int.
Qed.

Lemma int_inter A B : (inter A B)° =s inter A° B°.
Proof.
split; case; [|case]=> S [[HS1 HS2] HS3]. 
  by split; exists S; do 2! split=> //; apply/(include_trans HS2)=> y [].
case=> U [[HU1 HU2] HU3].
exists (inter S U); split; split=> //.
  exact: inter_open.
exact: in_inter.
Qed.

Lemma int_union A B : union A° B° <s (union A B)° .
Proof.
move=> x [][] S [[HS1 HS2] HS3]; exists S; do 2! split=>//;
 apply/(include_trans HS2).
  exact: union_left.
exact: union_right.
Qed.

Definition close (A : T -> Prop) : Prop := open (A^c).

Add Morphism close with signature (@eqset T) ==> iff as close_mor. 
Proof. by move=> A B HAB; rewrite /close HAB. Qed.

Lemma empty_close : close {ø : T}.
Proof. by rewrite /close empty_c; apply: full_in_open. Qed.

Lemma full_close : close {full : T}.
Proof. by rewrite /close full_c; apply: empty_in_open. Qed.

Lemma inter_close (E : (T -> Prop) -> Prop) : 
  (forall A, A \ins E -> close A) -> close (inter_s E id).
Proof. by rewrite /close inter_compl_union; apply: union_open2. Qed.  

Lemma union_close A B : close A -> close B -> close (union A B). 
Proof. by move=> HA HB; rewrite /close compl_union; apply: inter_open. Qed.

Definition closure (A : T -> Prop) := inter_s (fun B => close B /\ A <s B) id.

Notation "/ A" := (closure A).

Lemma closure_lc A B : close A -> B <s A -> /B <s A.
Proof. by move=> Hf HAB x; apply. Qed.

Lemma in_closure A : A <s /A. 
Proof. by move=> x Hx S [_ HS]; apply: HS. Qed.

Lemma close_closure A : close (/A).
Proof. by apply: inter_close => S []. Qed.

Add Morphism closure with signature (@eqset T) ==> (@eqset T) as closure_mor.
Proof.
move=> A B HAB; apply: double_i; apply: closure_lc; try exact: close_closure.
  by rewrite HAB; apply: in_closure.
by rewrite -HAB; apply: in_closure.
Qed.

Lemma closure_close_eq A : close A <-> /A <s A.
Proof.
split=> H; first exact: closure_lc.
have ->: A =s /A by apply: double_i => //; apply: in_closure.
exact: close_closure.
Qed.

Lemma closure_union A B : /(union A B) =s union (/A) (/B).
Proof.
split=>[Hx|].
  case: (classic ((/A) x)); first by left.
  case/not_all_ex_not=> S; move/not_imply_to_and=> [[H1 H2] H3].
  right=> U [HU1 HU2]; move: (Hx (union S U))=> Hf.
  have H: close (union S U) /\ union A B <s union S U. 
    split; first exact: union_close.
      exact: in_union.
  by case: (Hf H).
case=> H S [H1 H2]; apply: H; split=> //; apply: (include_trans _ H2).
  exact: union_left.
exact: union_right.
Qed.

Lemma closure_inter A B : /(inter A B) <s inter (/A) (/B).
Proof.
move=> x Hx. 
by split=>S [H1 H2]; apply: Hx; split=>//; apply: (include_trans _ H2)=>y [] //.
Qed.

Lemma clsr_cmpl_int A : /(A^c) =s (A°)^c .
Proof.
split=> [Hx|/not_ex_all_not Hx S [HS1 HS2]].
  apply: all_not_not_ex=> S.
  apply/andPC/not_and_to_imply=> [][H1 /compl_include H2].
  by apply: (Hx S^c); split=> //; rewrite /close compl_invo.
have/andPC/not_and_to_imply H := (Hx S^c).
apply/(compl_invo S)/H.
by split=> //; apply/compl_include; rewrite compl_invo.
Qed.

Lemma int_cmpl_clsr A : (A^c)° =s (/A)^c.
Proof.
split.
  case=> S [[HS1 HS2] HS3]; apply: ex_not_not_all.
  exists S^c; apply/not_imply_to_and.
  split; last exact/(compl_invo S).
   split=> //.  
     by rewrite /close  compl_invo.
   by apply/compl_include; rewrite compl_invo.
case/not_all_ex_not=> S /not_imply_to_and [[HS1 /compl_include HS2] HS3].
by exists S^c.
Qed.

Definition boundary (A : T -> Prop) := (/A) \ (A°).

Add Morphism boundary with signature (@eqset T) ==> (@eqset T) as boundary_mor.
Proof. by move=> A B HAB; rewrite /boundary HAB. Qed.

Lemma partition1 A x : x \ins A  -> x \ins A° \/ boundary A x. 
Proof.
move/in_closure=> H; case: (classic (boundary A x))=> HA; first by right.
by left; case/not_and_or: HA=> // Ho; apply/(compl_invo A°).
Qed.

Lemma partition2 A x : x \ins A^c \/ x \ins A° \/ boundary A x.
Proof.
case: (classic (x \ins A))=> HA; last by left.
by right; apply: partition1.
Qed.

Lemma boundary_eq_compl A : boundary A =s boundary A^c.
Proof.
by rewrite /boundary clsr_cmpl_int int_cmpl_clsr /without compl_invo inter_sym.
Qed.

Definition neighbourhood (V : T -> Prop) (x : T) := 
  exists A, [/\ x \ins A, open A & A <s V].

Definition neigh_open (V : T -> Prop) (x : T) := 
  x \ins (neighbourhood V) /\ open V.

Lemma in_neigh V x : x \ins (neighbourhood V) -> x \ins V.
Proof. by case=> S [H1 _ H2]; apply: H2. Qed.

Lemma open_neigh A : 
  open A <-> (forall x, x \ins A -> x \ins (neighbourhood A)).
Proof.
split=> [H x Hx|H]; first by exists A; split.
by apply/int_open_eq=> x Hx; case: (H x Hx)=> S [HS1 HS2]; exists S.
Qed.

Lemma int_m A : A° =s fun x => exists V, (x \ins (neighbourhood V) /\ V <s A).
Proof.
split=> [[] S [[H1 H2] H3]|[] S [[V [HV HVo HVS]]] H].
  by do 2! (exists S; split=> //).
by exists V; do 2! (split=> //); apply: (include_trans HVS H).
Qed.

Lemma closure_m A : 
 (/A) =s 
 fun x => (forall V, x \ins (neighbourhood V) -> exists y, y \ins (inter V A)).
Proof.
move=> x.
have HC : x \ins (/A)^c <-> x \ins (neighbourhood A^c).
  split=> H.
    case: (not_all_ex_not _ _ H)=> F.
    move/not_imply_to_and=> [[H1 H2] H3].
    exists (F^c); split=>//.
    by move/compl_include: H2.
  case: H=> F [HF1 HF2 HF3].
  apply: ex_not_not_all; exists (F^c).
  apply/not_imply_to_and.
  split; last by move=> Hfin.
  split; rewrite /close ?compl_invo //. 
  by apply/compl_include; rewrite compl_invo.
split=> H S.
  case=> V [HV HV1 HV2]; case: (classic (exists y, y \ins (inter S A)))=> // Hy.
  suff: x \ins (/A)^c => //.
  apply/HC; exists V; split=> // y /HV2 Hy2.
  have Hab:= ((not_ex_all_not _ _ Hy) y).
  by case: (not_and_or _ _ Hab)=> //.
case: (classic ((/A) x))=> [|/HC Hab]; first by apply. 
by case: (H A^c Hab)=> y [].
Qed.

Definition apart := forall (x y : T), x <> y -> exists V1, exists V2,         
 [/\ x \ins (neighbourhood V1), y \ins (neighbourhood V2) &
     (inter V1 V2) <s {ø : T}].

Lemma apart_eq : apart <-> 
  (forall (x y : T), x <> y -> exists V1, exists V2,         
 [/\ x \ins (neigh_open V1), y \ins (neigh_open V2) &
     (inter V1 V2) <s {ø : T}]).
Proof.
split=> H x y Hxy; case: (H x y Hxy)=> V1; case=> V2 [H1 H2 H3].
  case: H1=> S [HS1 HS2 HS3]; case: H2=> U [HU1 HU2 HU3]; exists S; exists U.
  split; try split=> //.
      by exists S; split.
    by exists U; split.  
  by move=> z [/HS3 Hz1 /HU3 Hz2]; apply: H3. 
by exists V1; exists V2; case: H1; case: H2. 
Qed.

Lemma fin_inter : forall (s : seq (T -> Prop)), 
   (forall g, g \in s -> open g) -> open (inter_seq s).
Proof.
elim=> [H|A s HR H] /=; first exact: full_in_open. 
apply: inter_open; first by apply/H/mem_head.
by apply: HR=> g He; apply/H/mem_behead.
Qed.

Definition open_family Ti (f : family T Ti) := 
  forall i, i \ins (ind f) -> open (f i).

Lemma open_inter_fini (Ti : eqType) (f : finite_family T Ti) : 
  open_family f ->  open (inter_fam f).
Proof.
move=> Ho; rewrite (inter_fin_eq f); apply: fin_inter. 
move=> g; case H: (ind_seq f)=>[|t s]; rewrite /fam_fin_to_seq H //.
rewrite in_cons; case/orP=>[/eqP ->|].
  by apply: Ho; rewrite ind_seqP H;apply: mem_head.
rewrite -/(map _ _).
case/(nthP (f t))=> i; rewrite size_map=> Hi <-.
rewrite (nth_map t) //; apply: Ho.
by rewrite ind_seqP H; apply/mem_behead/mem_nth.
Qed.

Definition quasi_compact (K : T -> Prop) :=
   forall Ti (f : family T Ti),
   open_family f -> cover K f -> 
   exists g : finite_family T Ti, sub_family f g /\ cover K g.

Add Morphism quasi_compact with signature (@eqset T) ==> iff
  as quasi_compact_mor.
Proof.
move=> K1 K2 H12; split=> H Ti f Hof Hrf.
  have Hr : cover K1 f by rewrite /cover H12. 
  case: (H Ti f Hof Hr)=> g [Hg1 Hg2]; exists g.
  by split=> //; rewrite /cover -H12.
have Hr : cover K2 f by rewrite /cover -H12. 
case (H Ti f Hof Hr)=> g [Hg1 Hg2]; exists g.
by split=> //; rewrite /cover H12.
Qed.

Definition compact (K : T -> Prop) := apart /\ quasi_compact K.

Add Morphism compact with signature (@eqset T) ==> iff as compact_mor.
Proof. by move=> A B; rewrite /compact=> ->. Qed.

Lemma compct_close K : compact K -> close K.
Proof.
case; move/apart_eq=> SP HK; apply/int_open_eq.
rewrite int_m=> x Hx.
pose V := mkfamily K (fun i y : T => (forall Hi : K i,
             pi1 (cid (pi2 (cid (SP x i (neq_compl Hx Hi))))) y)).
have Hr : cover K V.
  move=> z Hz; exists z => /=; split=> //.
  move=> H2; apply: in_neigh.
  by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx H2))))))=> _ [H _ _].
have Ho : open_family V.
  move=> z Hz; apply/open_neigh => y Hy /=.
  exists (pi1 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hz)))))).
  split; first exact: Hy.
  by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hz))))))=> _ [_ H _].
  move=> w Hw /= Hz2. 
  by have -> : Hz2 = Hz by apply: proof_irrelevance.  
case: (HK _ V Ho Hr)=> U [[H1 H2] H3].  
pose V' := mkfamily (ind U) (fun i y => (forall Hi : K i, 
                          pi1 (cid (SP x i (neq_compl Hx Hi))) y)).
have HfV': ffam V' by apply: finfam U.
pose fV' := Ff HfV'.
exists (inter_fam fV'); split.
  exists (inter_fam fV').
  split=> //.
    move=> y Hy /= Hi; apply: in_neigh.  
    by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hi))))))=> [][H _] _.
  apply: open_inter_fini.
  move=> z Hz; apply/open_neigh => y Hy /=.
  have Hy2 : K z by apply H1.
  exists (pi1 (cid (SP _ _ (neq_compl Hx Hy2)))).
  split; first exact: Hy.
  by case: (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hy2))))))=> [][_ H] _.
  move=> w Hw /= Hz2. 
  by have -> : Hz2 = Hy2 by apply: proof_irrelevance. 
move=> y Hy Habs.
case: (H3 y Habs)=> z [Hz1 Hz2].
have Hz : K z by apply: H1.
have [_ _ Hv]:= ( (pi2 (cid (pi2 (cid (SP _ _ (neq_compl Hx Hz))))))).
suff: False=> //.
apply: (Hv y); split.
  have Hbz : ind V' z by[].
  by apply: (inter_fam_in Hbz).
by move/H2: Hz2; apply. 
Qed.

Fixpoint ss_opt (T1 : Type) (s : seq (option T1)) := 
    match s with
     | [::]=> [::]
     | x :: s'=> match x with
                     |Some i => i :: ss_opt s'
                     |None => ss_opt s'
                   end
    end.

Lemma in_ss_opt : forall (T1 : eqType) (s : seq (option T1)) (x : T1),
   (Some x) \in s  <-> x \in ss_opt s. 
Proof.
move=> U; elim=> // t s HR x; case: t.
  move=> s0 /=; rewrite !in_cons; split; case/orP.
     -by move=> H ; apply/orP; left.
     -by move=> H; apply/orP; right; apply/HR.
     -by move/eqP=> ->; apply: mem_head.
     -by move=> H; apply/orP; right; apply/HR.
rewrite /= in_cons; split.
  -by case/orP=> H //; apply/HR.
  -by move=> H; apply/orP; right; apply/HR.
Qed.


Lemma close_in_comp K A : compact K -> close A -> A <s K -> compact A.
Proof.
move=> [HS HK] HA HKA; split=> // Ti V Ho Hr.
pose V' := mkfamily (oapp (ind V) true) (oapp V A^c).
have HrK : cover K V'. 
  move=> x HKx; case: (classic (x \ins A))=> HAx.
    by case: (Hr x HAx)=> y HBy; exists (Some y).
  by exists None.
have HoK: open_family V' by case => //=; apply: compl_f_o.
case: (HK _ V' HoK HrK)=> U [[H1 H2] H3].
pose W := mkfamily (fun i => ind U (Some i)) (fun i => U (Some i)).
have HfW: ffam W.
  pose s := ss_opt (ind_seq U).
  exists s => x /=; split=> H.
    by apply/in_ss_opt/ind_seqP.
  by apply/ind_seqP/in_ss_opt.
pose fW := Ff HfW.
exists fW; split=> [|x Hx].
  split=> x Hx.
    by have Hi: (Some x) \ins (ind V') by apply: H1.
  have -> : W x =s U (Some x) by [].
  by rewrite -H2.
have HKx : K x by apply: HKA.
case: (H3 x HKx); case.
  move=> t [Ht1 Ht2].
  by exists t.
by case=> HnA; move/H2=> Hn2; move: (Hn2 HnA).
Qed.

End TopologyTheory.

Existing Instance open_mor_Proper. 
Existing Instance interior_mor_Proper.
Existing Instance close_mor_Proper.
Existing Instance closure_mor_Proper.
Existing Instance boundary_mor_Proper.
Existing Instance quasi_compact_mor_Proper.
Existing Instance compact_mor_Proper.

Notation "A °" := (interior A) (at level 0, no associativity).
Notation "/ A" := (closure A).

Section Topology_sequence.

Variable (T : topologyType).

Definition Un_cvg (Un : nat -> T) (l : T) := 
  forall V, l \ins (neighbourhood V) -> 
  exists N : nat, forall n, N <= n -> (Un n) \ins V.
Notation "Un >->> l" := (Un_cvg Un l) (at level 0, no associativity).

Definition cluster_point (Un : nat -> T) (l : T) := 
  forall V, l \ins (neighbourhood V) -> 
  forall N : nat, exists n, N <= n /\ (Un n) \ins V.

Definition cluster_point2 (Un : nat -> T) (l : T) := 
  forall V, l \ins (neigh_open V) -> 
  forall N : nat, exists n, N <= n /\ (Un n) \ins V.

Lemma cluster_point2_eq (Un : nat-> T) : 
  cluster_point Un =s cluster_point2 Un.
Proof.
move=> l; split=> H V.
  by move=> [H1 H2]; apply: H.
case=> S [HS HS1 HS2]=> N.
have HVS : neigh_open S l by split=> //; move/open_neigh: HS1; apply.
by case: (H S HVS N)=> n [H1 H2]; exists n; split=> //; apply: HS2. 
Qed.

Definition fam_clstr (Un : nat -> T) (n : nat) (x : T) :=
  exists k, n <= k /\ x = Un k.

Definition cluster_point3 (Un : nat -> T) := 
    inter_fam (mkfamily {full : nat} (fun n x => (/(fam_clstr Un n)) x)). 

Lemma cluster_point3_eq Un : cluster_point Un =s cluster_point3 Un.
Proof.
apply: double_i => x Hx.
  move=> N _ ; apply/closure_m => V HV.
  case: (Hx V HV N)=> n [H1 H2] .
  by exists (Un n); split=> //; exists n.
rewrite /cluster_point2 /inter_fam /= in Hx.
move=> V HV N.
have t : True by [].
move/closure_m: (Hx N t)=> Hx2.
case: (Hx2 V HV)=> y [Hy1 [n [Hy3 Hy4]]].
by exists n; split=> //; rewrite -Hy4.
Qed.

Lemma lim_unicity Un l1 l2 : (apart T) ->
   Un >->> l1 -> Un >->> l2 -> l1 = l2.
Proof.
move=> HS H1 H2.
case: (classic (l1 = l2))=> // H.
case: (HS _ _ H)=> V1; case=> V2 [HV1 HV2 HV].
case: (H1 _ HV1)=> N1 HN1; case: (H2 _ HV2)=> N2 HN2.
pose N := maxn N1 N2.
have H1N : N1 <= N by rewrite leq_max; apply/orP; left.
have H2N : N2 <= N by rewrite leq_max; apply/orP; right.
have/HV Habs : (Un N) \ins (inter V1 V2).  
  split; first exact: HN1.
  exact: HN2.
done.
Qed.

Structure sub_seq : Type := mkss
   {
     phi :> nat -> nat;
     grow : forall m n, (m < n)%N -> (phi m < phi n)%N
   }.

Lemma grow_contra : forall (g : sub_seq) m n, g n <= g m -> n <= m.
Proof.
move=> g m n; rewrite leqNgt (leqNgt n m).
by apply: contra; apply: grow.
Qed.

Lemma grow_eq : forall g : nat ->nat, 
   (forall k, (g k < g k.+1)%N) <-> (forall m n, (m<n)%N -> (g m < g n)%N ).
Proof.
move=> g; split => H; last by move=> k; apply: (H k k.+1).
by move=> m n; elim: n m => // n HR m; rewrite ltnS leq_eqVlt => h; case/orP: h;
 [move/eqP => -> 
 | move=> h; apply: (ltn_trans (HR m h))].
Qed.

Lemma ss_min_id : forall (g : sub_seq) n, n <= g n.
Proof.
move=> g; elim=> // n HR.
apply: (leq_ltn_trans HR); clear HR; move: n; apply/grow_eq.
exact (grow g).
Qed.

Lemma cvg_ss_cvg : forall (Un : nat -> T) (g : sub_seq) l,
   Un >->> l -> (Un \o g) >->> l.
Proof.
move=> Un g l H V HV.
case: (H V HV)=> N1 HN1.
exists N1=> n Hn.
by apply: HN1; apply: (leq_trans Hn (ss_min_id g n)).
Qed.

Lemma ss_cvg (Un : nat -> T) l : 
  (exists g : sub_seq, (Un \o g) >->> l) -> l \ins (cluster_point Un).
Proof.
case=> g Hg V HV N.
case: (Hg V HV)=> N2 HN2.
exists (maxn (g N) (g N2)).
have : N <= (maxn (g N) (g N2)) /\ N2 <= (maxn (g N) (g N2)).
  by rewrite !leq_max; split; apply/orP; [left | right]; apply: ss_min_id.
case=> [HN HN22]; split=> //.
rewrite /maxn; case H: ((g N) < (g N2)); apply: HN2 => //.
by apply: (@grow_contra g); move/negbT: H; rewrite -leqNgt.
Qed.

(* Definition untangle (s : seq T)(P : T -> nat -> Prop)  *)
(*   (f : forall x, x\in s -> exists n, P x n) x  : nat := *)
(*  match x \in s as b return x\in s = b -> nat with  *)
(*    | true => fun h => pi1 (cid ((f x h)))  *)
(*    | false => fun _=> 0%N  *)
(*   end (erefl _). *)

(* Lemma untangle_prop : forall (s : seq T) P  *)
(*    (f : forall x, x \in s -> exists n, P x n) x, x \in s -> P x (untangle f x). *)
(* Proof. *)
(* move => s P f x h; rewrite /untangle. *)
(* move: (erefl (x \in s)). *)
(* rewrite {2 3}[x \in s] h. *)
(* move => e; apply pi2. *)
(* Qed. *)

Lemma Bolzano_Weierstrass : forall (Un : nat -> T) K , compact K ->
   (forall n, (Un n) \ins K) -> exists l, l \ins (cluster_point Un).
Proof.
move=> Un K [_ HK] HUn.
case: (classic (exists l : T, l \ins (cluster_point Un))) => //.
move/not_ex_all_not=> Habs; suff : False => //.
have H n : n \ins (cluster_point2 Un)^c by apply/cluster_point2_eq.
have H1 l : exists V,  ~(l \ins (neigh_open V) ->
              forall N : nat, exists n, N <= n /\ (Un n) \ins V).
  by apply: not_all_ex_not; apply: H.
have H2 l : exists V, l \ins (neigh_open V) /\ 
             ~(forall N : nat, exists n, N <= n /\ (Un n) \ins V).
  by case: (H1 l)=> V HV; exists V; apply/not_imply_to_and.
have H3 l : exists V, l \ins (neigh_open V) /\ 
              exists N, forall n, ~ (N <= n) \/ (Un n) \ins V^c.
  case: (H2 l)=> V [HV1 HV2].
  exists V; split=> //; case/not_all_ex_not: HV2=> N. 
  by move/not_ex_all_not=> HN; exists N=> n; apply: not_and_or; apply: HN.
clear H Habs H1 H2.
pose f := (pi1 (cid (H3 _))).
pose W := mkfamily K f.
have Hr : cover K W.
  move=> x Hx /=; exists x; split => //; apply: in_neigh.
  by apply: (proj1 (proj1 (pi2 (cid (H3 x))))).
have Ho :  open_family W.
  by move=> x Hx; apply: (proj2 (proj1 (pi2 (cid (H3 x))))).
case : (HK _ W Ho Hr)=> U [[HU1 HU2] HU3].
have:= ind_seqP U.
move: (ind_seq U); case=> [HU4 | t s HU4].
  have Hkf : forall x, K x -> False.
    by move=> x Hx; case: (HU3 x Hx)=> y [Hbs _]; move/HU4: Hbs.
  by move: (Hkf (Un 0%N) (HUn 0%N)).
pose P := fun N => forall n, N <= n -> ~ U _ (Un n).
pose g := (pi1 (cid (proj2 (pi2 (cid (H3 _)))))).
have Habs : forall x, x \in t :: s -> P x (g x).
  move=> x; move/HU4=> Hx n Hn.
  apply/(HU2 x Hx).
  move/or_to_imply: ((pi2 (cid (proj2 (pi2 (cid (H3 x)))))) n).  
  by apply.
pose m := \max_(x <- t :: s) g x.
  rewrite (big_nth t) big_mkord in m.
have: { i :  'I_(size (t :: s)) | m = g (nth t (t :: s) i)}.
   by apply: eq_bigmax; rewrite card_ord.
case=> i Hi.
pose x := (nth t (t :: s) i).
pose n := g x.
have Hm : forall j :'I_(size (t :: s)), g (nth t (t :: s) j) <= n.
  by move=> j; rewrite /n -Hi; apply: (leq_bigmax j).
have H : forall y, y \in t :: s -> index y (t :: s) < size (t :: s).
   by move=> y Hy; rewrite index_mem. 
have HP : forall y, y \in t :: s ->
  P y (g (nth t (t :: s) (index y (t :: s)))).
  by move=> y Hy; rewrite nth_index //; apply: Habs.
have Hfin : forall x, ind U x -> ~ U x (Un n).
  move=> y; move/HU4=> Hy; apply: (HP y Hy); rewrite insE -index_mem in Hy.  
  exact (Hm (Ordinal Hy)).
by case: (HU3 (Un n) (HUn n))=> y [Hy1 Hy2]; apply: (Hfin y Hy1).
Qed.

Lemma close_lim A : close A -> 
 (forall (Un : nat -> T) l, (forall n, (Un n) \ins A) -> Un >->> l -> l \ins A).
Proof.
move/closure_close_eq=> H Un l HA Hcv; apply/H/closure_m=> V HV.
by case: (Hcv V HV)=> n Hn; exists (Un n); split=> //; apply: (Hn n).
Qed.

End Topology_sequence.

Notation "Un >->> l" := (Un_cvg Un l) (at level 0, no associativity).

Section Topology_function.

Variables T1 T2 : topologyType.

Definition continue_pt (f : T1 -> T2) (x : T1) :=
  forall V', (f x) \ins (neighbourhood V') ->
  exists V, x \ins (neighbourhood V) /\ f^(V) <s V'.

Definition continue (f : T1 -> T2) := forall x, continue_pt f x.

Lemma continue_open : forall (f : T1 -> T2),
  (continue f <-> forall U, open U -> open f@^-1(U)).
Proof.
move=> f; split => H.
  move=> U HU.
  apply/int_open_eq=> x Hx; case: (H x U). 
    by exists U; split; case: Hx => y [Hy1 Hy2] //; rewrite Hy2 in Hy1.
  move=> S [] [A [HA1 HA2 HA3]] HSU.
  exists A; split=> //; split=> // y Hy.
  exists (f y); split=> //; apply: HSU.
  by exists y; split=> //; apply: HA3.
move=> x V' [] A [HA1 HA2 HA3]. 
exists f@^-1(A); split.
  exists f@^-1(A); split=> //; first by exists (f x).
  exact: H.
by apply: (include_trans _ HA3); apply: f_finv_in.
Qed.

Lemma continue_close : forall (f : T1 -> T2),
  (continue f <-> forall U, close U -> close f@^-1(U)).
Proof.
move=> f; rewrite /close; split=> H.
  by move=> U; rewrite -compl_finv; case: (continue_open f)=> Hf _; apply: Hf.
have H2 U: open (U)^c -> open (f) @^-1(U^c).
  by rewrite compl_finv; apply: H.
by apply/continue_open=> U; rewrite -(compl_invo U); apply: H2.
Qed.

Lemma cont_cmpct f K : apart T2 -> continue f -> compact K -> compact (f^(K)).
Proof.
move=> HS Hf [H1 H2]; split=> // Ti U Ho Hr.
pose V := mkfamily (ind U) (fun i : Ti => (f) @^-1( U i)).
have HoV : open_family V.
   by move=> x Hx; case:(continue_open f)=> H _; apply: H=> //; apply: Ho.
have HrV : cover K V.
   move=> x Hx; case: (Hr (f x)); first by exists x.
   by move=> y [Hy1 Hy2]; exists y; split=> //; exists (f x).
case: (H2 _ V HoV HrV)=> W [[HW1 HW2] HW3].
pose V' := mkfamily (ind W) (fun i x => ind W i /\ U i x).
have HfV': ffam V' by apply: (finfam W).
pose fV' := Ff HfV'.
exists fV' ; split.
  split=> //.
  by move=> i Hi x; split=> //= [][_ Hx]. 
move=> x; case=> y [Hy1 ->]; case: (HW3 y Hy1)=> z [Hz1 Hz2].
exists z; split => //.
by move/(HW2 _ Hz1): Hz2=> /=; case=> w [H <-].
Qed.

Lemma continue_cvg f : continue f -> 
  (forall (Un : nat -> T1) l, Un >->> l -> (f \o Un) >->> (f l)).
Proof.
move=> Hf Un l H V HV; case: (Hf l V HV)=> V' [H1 H2]; case: (H V' H1)=> N HN.
by exists N=> n Hn; apply: H2; exists (Un n); split=>//; apply: HN.
Qed.

End Topology_function.    
 
Section Topology_function2.
  
Variables T1 T2 T3 : topologyType.

Lemma continue_comp : forall (f : T1 -> T2) (g : T3 -> T1),
 continue f -> continue g -> continue (f \o g). 
Proof.
move=> f g; move/continue_open=> Hf; move/continue_open=> Hg.
by apply/continue_open=> U HU; rewrite comp_inv; apply: Hg; apply: Hf.
Qed.

End Topology_function2.   


Module EspMet.

Open Scope ring_scope.

Structure mixin_of (T : Type) := Mixin {
     R : archiFieldType;
   dist : T -> T -> R;
    _ : forall x y,  0 <= dist x y ;
    _ : forall x y, dist x y = dist y x;
    _ : forall x y, dist x y = 0 <-> x = y;
    _ : forall x y z, dist x y <= dist x z + dist z y
   }. 

Section ClassDef.

Notation class_of := mixin_of (only parsing).

Structure type := Pack {sort; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.

Variable (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
Definition pack c := @Pack T c T.
Definition clone := fun c & cT -> T & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.

Coercion sort : type >-> Sortclass.
Notation espMetType := type.
Notation EspMetMixin := Mixin.
Notation EspMetType T m := (@pack T m).
Notation "[ 'espMetMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'espMetMixin' 'of' T ]") : form_scope.
Notation "[ 'espMetType' 'of' T 'for' C ]" := (@clone T C _ idfun id)
  (at level 0, format "[ 'espMetType' 'of' T 'for' C ]") : form_scope.
Notation "[ 'espMetType' 'of' T ]" := (@clone T _ _ id id)
  (at level 0, format "[ 'espMetType' 'of' T ]") : form_scope.

End Exports.
End EspMet.
Export EspMet.Exports.


Section positive.

Variable R : numDomainType.
Open Scope ring_scope.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.

 
Lemma inv_INR_pos k : (0 : R) < k.+1%:R^-1.
Proof. by rewrite invr_gt0 ltr0n. Qed.

Structure posDomain := PosD { sort :>  R ; _ : 0 < sort }. 

Canonical Structure natmulV_pos k := PosD (inv_INR_pos k). 

Fact lt_pos (r : posDomain) : 0 < r. Proof. by case: r. Qed.

Lemma ler_minl_pos (x y : posDomain) : minr x y <= x.
Proof.
rewrite /minr.
case: (real_lerP (gtr0_real (lt_pos x)) (gtr0_real (lt_pos y)))=> //.
exact: ltrW.
Qed.

Lemma ler_minr_pos (x y : posDomain) : minr x y <= y.
Proof.
rewrite /minr.
by case: (real_lerP (gtr0_real (lt_pos x)) (gtr0_real (lt_pos y))).
Qed.

Lemma minr_gt0 (x y : R) : 0 < x -> 0 < y -> 0 < minr x y.
Proof. by move=> xlt0 ylt0; rewrite /minr; case: (x <= y). Qed.

Lemma inv_pos (r : posDomain) : 0 < (sort r)^-1.
Proof. by rewrite invr_gt0 lt_pos. Qed.

Canonical Structure posD_inv r := PosD (inv_pos r).

Lemma ltn_ltrV (F : numFieldType) k n : 
 (k < n)%N = (n.+1%:R^-1 < k.+1%:R^-1 :> F).
Proof.
rewrite /= -(ltr_pmul2l (ltr0Sn _ k)) -(ltr_pmul2r (ltr0Sn _ n)).
by rewrite -mulrA mulrV ?mulVr ?unitf_gt0 ?ltr0Sn // mulr1 mul1r ltr_nat.
Qed.

End positive.

Section positive_archi.

Variable R : archiFieldType.
Open Scope ring_scope.

Import Num.Theory.
Import Num.Def.

Lemma ex_sub_proof (x : posDomain R) : exists n : nat, x < n%:R.
Proof. by exists (Num.bound x); rewrite archi_boundP // ltrW // lt_pos. Qed.

Definition upbound (x : posDomain R) := ex_minn (ex_sub_proof x). 

Lemma floor_bound (x : posDomain R) : (upbound x).-1%:R <= x < (upbound x)%:R.
Proof.
rewrite /upbound; case: ex_minnP=> [][] =>[|m -> /(_ m)/contra].
  by rewrite ltrNge ler_eqVlt lt_pos orbT.
by rewrite -lerNgt ltnn => ->.
Qed.

Lemma upbound_gt0 x : (0 : R) < (upbound x)%:R.
Proof. by have/andP [_]:= floor_bound x; apply/ltr_trans/lt_pos. Qed.

End positive_archi.


Section EspMet_Struct.

Variable E : espMetType.

Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.

Notation d := (EspMet.dist (EspMet.class E)).

Definition metricDomain := (EspMet.R (EspMet.class E)).

Lemma dist_pos :  forall x y : E,  0 <= d x y.
Proof. by case: E => ? []. Qed.

Lemma dist_sym : forall x y : E, d x y = d y x.
Proof. by case: E=> ? []. Qed.

Lemma dist_refl :  forall x y : E, d x y = 0 <-> x = y.
Proof. by case: E=> ? []. Qed.

Lemma dist_tri :  forall x y z : E, d x y <= d x z + d z y.
Proof. by case: E=> ? []. Qed.
 
Definition eqmet (x y : E) : bool := (d x y) == 0.

Lemma eqmetP : Equality.axiom eqmet.
Proof.
by move=> x y; apply: (iffP idP)=>[/eqP/dist_refl H|/dist_refl/eqP H].
Qed.

Canonical Structure EspMet_eqMixin := Eval hnf in EqMixin eqmetP.
Canonical Structure EspMet_eqType := Eval hnf in EqType E EspMet_eqMixin.

Notation posD := (posDomain metricDomain).

Definition open_ball (x : E) (r : posD) (y : E) : Prop := d x y < r.
Notation B x r := (open_ball x r).

Definition open_met (A : E -> Prop) := 
 forall x, x \ins A -> exists r, B x r <s A.

Lemma empty_in_open_met : open_met {ø : E}.
Proof. by []. Qed.

Lemma full_in_open_met : open_met {full : E}.
Proof. by move=> x _; exists (PosD ltr01). Qed.

Lemma union_open_met : forall (S : (E -> Prop) -> Prop), S <s open_met ->
  open_met (union_s S id).
Proof.
move=> S HS x; case=> A [HA1 HA2].
case: (HS A HA1 x HA2)=> r Hr.
exists r; apply: (include_trans Hr).
by apply: (@in_unions _ _ id).
Qed.

Lemma inter_open_met : forall A B, open_met A -> open_met B ->
   open_met (inter A B).
Proof.
move=> A B HA HB x [Hx1 Hx2].
case: (HA x Hx1)=> r1 Hr1; case: (HB x Hx2)=> r2 Hr2.
exists (PosD (minr_gt0 (lt_pos r1) (lt_pos r2))).
move=> y Hy; split.
  by apply/Hr1/(ltr_le_trans Hy)/ler_minl_pos.
by apply/Hr2/(ltr_le_trans Hy)/ler_minr_pos.
Qed.

Lemma ext_open_met : forall A B, A =s B -> (open_met A <-> open_met B).
Proof.
move=> A B H; split=> Ho x Hx.
  elim: (Ho x ((right_i H) x Hx))=> r Hr; exists r.
  exact (include_trans Hr (left_i H)).
elim: (Ho x ((left_i H) x Hx))=> r Hr; exists r.
exact (include_trans Hr (right_i H)).
Qed.

Definition EspMet_TopMixin := Eval hnf in TopologyMixin 
  empty_in_open_met full_in_open_met union_open_met inter_open_met ext_open_met.
Definition EspMet_TopType := Eval hnf in
   TopologyType E EspMet_TopMixin.

Canonical EspMet_TopMixin.
Canonical EspMet_TopType.

End EspMet_Struct.

Notation d := (EspMet.dist (EspMet.class _)).
Coercion EspMet_TopType : espMetType >-> topologyType.

Section EspMet_Theory.

Variable E : espMetType.

Notation d := (EspMet.dist (EspMet.class E)).
Notation B x r := (open_ball x r).

Notation posD := (posDomain (metricDomain E)).

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.
(* Delimit Scope R_scope with Re. *)
Open Scope ring_scope.

Lemma dist_nul x : d x x = 0.
Proof. exact/dist_refl. Qed.

Lemma bxrx (x : E) r : x \ins (B x r).
Proof. by rewrite /open_ball insE dist_nul lt_pos. Qed.

Lemma cont_to_discr (r : posD) :
   (PosD (inv_INR_pos _ (upbound (posD_inv r))) < r).
Proof.
have/andP [_]:= floor_bound (posD_inv r).
set n := upbound _.
have Hlt0 : (0 : (metricDomain E)) < n%:R^-1 by rewrite invr_gt0 /n upbound_gt0.
rewrite /= -(ltr_pmul2l (lt_pos r)) -(ltr_pmul2r Hlt0).
rewrite -[_*_%:R*_]mulrA !mulrV ?unitf_gt0 ?lt_pos -?[0 < _%:R]invr_gt0 //.
rewrite mulr1 mul1r; apply: ltr_trans.
rewrite invr_gt0 in Hlt0.
by rewrite -{2}[n]prednK // -?ltn_ltrV ?prednK // -(ltr_nat (metricDomain E)).
Qed.

Lemma ball_open : forall (x :  E) r, open (B x r).
Proof.
move=> x r a H.
have p : 0 < (sort r) - d x a by rewrite subr_gt0.
pose s:= PosD p.
exists s => y h.
have t : (sort r) = (d x a + (sort s)) by rewrite addrC addrNK.
apply: (@ler_lt_trans _ (d x a + d a y)); first exact: dist_tri.
by rewrite t ltr_add2l. 
Qed.

Lemma ball_neigh : forall (x : E) r, neighbourhood (B x r) x.
Proof.
move=> x r; exists (B x r).
split=> //; first by apply: bxrx.
exact: ball_open.
Qed.
  
Lemma interior_met : forall (A : E -> Prop), 
   A° =s (fun x=>exists r, (B x r) <s A). 
Proof.
move=> A; apply: double_i=> x [] => [S [[HS1 HS2] HS3]|r Hr].
  case: (HS1 x HS3)=> r Hr; exists r.
  exact: (include_trans Hr HS2).
exists (B x r); split; last exact: bxrx.
by split => //; apply: ball_open.
Qed.

Lemma int_nat (A :  E -> Prop) : 
A° =s (fun x=> exists n : nat, (B x (PosD (inv_INR_pos _ n))) <s A).
Proof.
rewrite interior_met=> x; split; elim=> r Hr.
  exists (upbound (posD_inv r)).
  apply: (include_trans _ Hr)=> y Hy; apply: (ltr_trans Hy).
  exact: (cont_to_discr r).
by exists (PosD (inv_INR_pos _ r)).
Qed.

Lemma closure_met : forall (A : E -> Prop) ,
  /A =s (fun x=> forall r, exists y, inter (B x r) A y).
Proof.
move=> A; apply: double_i=> x.
  by move/closure_m=> Hx r; apply: (Hx (B x r)); apply: ball_neigh.
move=> Hr; apply/closure_m => V; case=> B [HB1 HB2 HB3].
case: (HB2 x HB1)=> r HrB; case: (Hr r)=> y [Hy1 Hy2].
exists y; split=> //.
by apply/HB3/HrB.
Qed.

Lemma closure_nat : forall (A : E -> Prop), /A =s 
(fun x=> forall n : nat, exists y, inter (B x (PosD (inv_INR_pos _ n))) A y). 
Proof.
move=> A; rewrite closure_met; split => H r.
  by elim: (H (PosD (inv_INR_pos _ r)))=> y Hy; exists y.
elim: (H (upbound (posD_inv r)))=> y [Hy1 Hy2]; exists y.
by split=> //; apply/(ltr_trans Hy1)/(cont_to_discr r).
Qed.

Definition borne (A : E -> Prop) :=
  exists r, forall x y, A x -> A y -> B x r y.

Definition borne2 (A : E -> Prop) : Prop := 
   exists r, exists x, A <s B x r. 

Lemma born_eq : forall A, borne2 A -> borne A.
Proof.
move=> A; case=> r; case=> x H.
have Hp : 0 < (sort r) + (sort r) by rewrite addr_gt0 // lt_pos.
exists (PosD Hp)=> y z Hy Hz. 
have Hd : d y x + d x z < (sort r) + (sort r).
  by rewrite ltr_add // ?(dist_sym y) H. 
by apply: (ler_lt_trans _ Hd); apply: dist_tri.
Qed.

Lemma born_nat A : borne A <-> 
  exists n, forall x y, A x -> A y -> (B x (PosD (ltr0Sn _ n)) y).
Proof.
split; case=> r H. 
  exists (upbound r).-1=> x y Hx Hy.
  apply: (ltr_trans (H _ _ Hx Hy))=> /=.
  have/andP [_] := floor_bound r.
  by rewrite prednK // -(ltr_nat (metricDomain E)) upbound_gt0.
by exists (PosD (ltr0Sn _ r)).
Qed.

Lemma pos : forall (x y : E), x<>y -> 0 < d x y.
Proof.    
move=> x y Hxy.
rewrite ltr_def dist_pos andbT. 
by apply/eqP=> /dist_refl.
Qed.

Lemma pos_2 : forall (x y : E), x <> y -> 0 < (d x y)/2%:R.
Proof.
move=> x y Hxy.
by rewrite mulr_gt0 // ?pos // invr_gt0 ltr0n.
Qed.

Lemma apart_esp_met : apart E.
Proof.
move=> x y Hxy.
exists (B x (PosD (pos_2 Hxy))).
exists (B y (PosD(pos_2 Hxy))).
split; try by apply: ball_neigh.
move=> z [Hz1 Hz2].
have H : d x y = 2%:R * (d x y / 2%:R). 
  by rewrite mulrA mulrC mulKr ?unitf_gt0 ?ltr0n.
have: d x z + d y z < d x y.
 by rewrite H mulrDl !mul1r ltr_add.
by rewrite ltrNge (dist_sym y z) dist_tri.
Qed.
  

Lemma comp_born : forall (K : E -> Prop), compact K -> borne K.
Proof.
move=> K [HS H].
pose V := mkfamily K (fun x y => B x (PosD ltr01) y). 
have Hr : cover K V.
  by move=> x HKx; exists x; split=> //; apply: bxrx.
have Ho : open_family V by move=> x Hx; apply: ball_open.
case: (H _ V Ho Hr)=> g [[Hg1 Hg2] Hg3].
have:= ind_seqP g.
case He: (ind_seq g)=> [|t s0].
  move=> Hg4; exists (PosD ltr01)=> x y HKx.
  case: (Hg3 x HKx)=> z [Habs _].
  by move/Hg4: Habs.
move=> Hs0 .
pose r:= \maxr_(i <- (ind_seq g)) d t i.
have [x /andP [_ /eqP Hx]]: {x |  (x \in (ind_seq g)) &&  (d t x == r)}.
  apply: bigmax_eq=>[i|]; first by rewrite dist_pos.
  by rewrite He.
have Hp1 : 0 < r + 1 by rewrite -[0]addr0 ler_lt_add // ?ltr01 // -Hx dist_pos.
have Hp : 0 < (r +1) + (r + 1) by rewrite addr_gt0. 
exists (PosD Hp)=> y z Hy Hz.
have Hi : forall w, K w -> exists v, v \in (ind_seq g) /\ B v (PosD ltr01) w.
  move=> w Hw; case: (Hg3 w Hw) => a [Ha1 Ha2].
  exists a; split; first by rewrite He; apply/Hs0.
  by apply/Hg2.
case: (Hi y Hy)=> v [Hiv Hv].
case: (Hi z Hz)=> w [Hiw Hw].
apply: (ler_lt_trans (dist_tri _ _ t)).
rewrite ler_lt_add // ?(dist_sym y t) ?ltrW //.
  rewrite (ler_lt_trans (dist_tri _ _ v)) // ler_lt_add // bigmax_ger //.
  by move=> i; rewrite dist_pos.
rewrite (ler_lt_trans (dist_tri _ _ w)) // ler_lt_add // bigmax_ger //.
by move=> i; rewrite dist_pos.
Qed.

End EspMet_Theory.

Notation B x r := (open_ball x r).


Notation posD := (posDomain (metricDomain _)).


Section EspMet_sequence.

Import GRing.Theory.
Import Num.Theory.
Import Num.Def.
Open Scope ring_scope.
Variables E E': espMetType.

Definition Un_cvg_met (Un : nat -> E) (l : E) :=
  forall (eps : posD), exists N : nat, 
  forall n, (N <= n)%N -> B l eps (Un n). 

Lemma Un_cvg2 : forall (Un : nat -> E) l, Un_cvg_met Un l <-> 
  Un >->> l.
Proof.
move=> Un l; split => H; [move => V  [S [HS1 HS2 HS3]] | move=> e].
  case: (HS2 l) => // e hB; case: (H e) => N Hf; exists N => n Hn; apply: HS3.
  exact ((hB (Un n)) (Hf n Hn)).
exact (H _ (ball_neigh l e)).
Qed.

Lemma Un_cvg3 : forall (Un : nat -> E) l, Un >->> l <-> 
       (forall (k : nat), exists N : nat, forall n, (N <= n)%N -> 
        B l (PosD (inv_INR_pos _ k) ) (Un n)). 
Proof.
move=> Un l; split.
  by move/Un_cvg2=> H k; case (H (PosD (inv_INR_pos _ k)))=> N H2; exists N.
move=> H; apply/Un_cvg2=> e; elim (H (upbound (posD_inv e))) => N H2.
exists N => n h; move: (H2 n h) => H1; apply: (ltr_trans H1).
exact: (cont_to_discr e).
Qed.


Lemma ex_ss_cvg : forall (Un : nat -> E), 
  (exists x, exists g: sub_seq, forall (k:nat), 
   B x (PosD (inv_INR_pos _ k))  (Un (g k))) -> 
  (exists g : sub_seq, exists x,  (Un \o g) >->> x). 
Proof.
move=> Un; case => x; case => g H; exists g; exists x.
apply/Un_cvg3 => k; exists k.+1 => n h.
by apply: (ltr_trans (H n)); rewrite -ltn_ltrV.
Qed.


(* Fixpoint g_rec (k:nat) (F:nat ->nat->nat) : nat :=  *)
(*       match k with *)
(*         | 0 => F O O *)
(*         | k'.+1 => F (g_rec k' F).+1 k *)
(*       end. *)


Lemma ex_ss_cvg2 : forall (Un : nat -> E) l, 
 (forall N: nat, forall (k:nat), exists n: nat, 
      (N <= n)%N /\ B l (PosD (inv_INR_pos _ k)) (Un n.+1)) -> 
             (exists g: sub_seq, (Un \o g) >->> l). 
Proof.
move=> Un l H.
have: exists G : nat -> (nat -> nat), forall N: nat, 
      forall k : nat , (N <= (G N) k)%N /\ 
      B l (PosD (inv_INR_pos _ k)) (Un ((G N) k).+1) 
    by apply: (functional_choice (fun (N: nat) (f: nat -> nat) => forall k:nat,
        (N <= f k)%N /\ B l (PosD (inv_INR_pos _ k)) (Un (f k).+1))) => N; 
       apply: (functional_choice (fun (k:nat) (n:nat) => 
        (N <= n)%N /\ B l (PosD (inv_INR_pos _ k)) (Un n.+1))).
elim => G H1.
pose g := fix t m : nat := if m is m'.+1 then (G (t m') m').+1 else G 0%N 0%N.
have H2: forall m n, (m < n)%N -> (g m < g n)%N.
  by apply/grow_eq=> k; case: (H1 (g k) k).
exists (mkss H2) => /=; apply/Un_cvg3 => k; exists k.+1 => n h.
set r:= (PosD _).
rewrite -(ltn_predK h). 
case: (H1 (g n.-1) n.-1)=> _ H3.
apply/(ltr_le_trans H3).
rewrite ler_eqVlt -ltn_ltrV (inj_eq invr_inj) eqr_nat eqSS.
by rewrite  eq_sym -leq_eqVlt -ltnS (ltn_predK h).
Qed.

Lemma clstr_ss_cvg : forall (Un : nat -> E) l, cluster_point Un l ->
  (exists g : sub_seq, (Un \o g) >->> l).
Proof.
move=> Un l H; apply: ex_ss_cvg2=> N k.
have HV : neighbourhood (B l (PosD (inv_INR_pos _ k))) l.
  by apply: ball_neigh.
case: (H _ HV N.+1)=> n [Hn H2].
exists n.-1; rewrite (ltn_predK Hn); split=> //.
by rewrite -ltnS (ltn_predK Hn).
Qed.

End EspMet_sequence. 

Section EspMet_function.

Import Num.Theory.
Import Num.Def.

Open Scope ring_scope.
Variables E E': espMetType.

Definition continu_pt (f : E -> E') (x : E) : Prop := 
  forall (r : posD), exists r' : posD, forall (y : E),
  B x r' y -> B (f x) r (f y).

Definition continu (f : E -> E') : Prop := forall x, continu_pt f x.

Lemma continu_met : forall (f : E -> E') (x : E), 
  continu_pt f x <-> continue_pt f x.
Proof.
move=> f x; split=> H.  
  move=> V [S [H1 H2 H3]].
  case: (H2 (f x))=> // r Hr.
  case: (H r)=> r2 Hr2.
  exists (B x r2).
  split; first by apply: ball_neigh.
  move=> y [z [Hz1 Hz2]].
  by apply: H3; apply: Hr; rewrite Hz2; apply: Hr2.  
move=> r.
case: (H _ (ball_neigh (f x) r))=> V [[S [H1 H2 H3]] H4].
case: (H2 x)=> // r2 Hr2.
exists r2=> y Hy.
apply: H4; exists y; split=> //.
by apply: H3; apply: Hr2.
Qed.

Lemma continu_eq : forall (f : E -> E'), continu f <-> continue f.
Proof.
by move=> f; split=> H x; apply/continu_met.
Qed.

Lemma lim_close : forall (A : E -> Prop), 
  (forall (Un : nat -> E) l, (forall n, A (Un n)) -> Un >->> l -> A l) 
  -> close A.
Proof.
move=> A H; apply/closure_close_eq=> x.
move/closure_nat => Hx.
have: exists Un : nat -> E, forall n,
          inter (B x (PosD (inv_INR_pos _ n))) A (Un n)
        by apply: (functional_choice (fun (n : nat) (y : E) =>
         inter (B x (PosD (inv_INR_pos _ n))) A y)).
elim=> Un Hun.
apply: (H Un x); first by move=> n; move: (Hun n)=> [_ Ha].
apply/Un_cvg3=> k; exists k.+1 => n Hn.
have [Hu _]:= (Hun n).
by apply: (ltr_trans Hu); rewrite -ltn_ltrV.
Qed.

End EspMet_function.

(* Section Metric_of_NumDomain. *)
(* Import GRing.Theory. *)
(* Import Num.Theory. *)
(* Import Num.Def. *)
(* Open Scope ring_scope. *)

(* Variable T : numIdomainType. *)


(* Definition d_num (x y : T) := `|x - y|. *)

(* Lemma d_num_pos x y :  0 <= d_num x y. *)
(* Proof. exact: normr_ge0. Qed. *)

(* Lemma d_num_sym x y : d_num x y = d_num y x. *)
(* Proof. by rewrite /d_num -{1}[x]opprK -opprD addrC normrN. Qed. *)

(* Lemma d_num_refl x y : d_num x y = 0 <-> x = y. *)
(* Proof. *)
(* split=>[/normr0P|->]. *)
(*   by rewrite subr_eq0=> /eqP. *)
(* by rewrite /d_num subrr normr0. *)
(* Qed. *)

(* Lemma d_num_tri x y z : d_num x y <= d_num x z + d_num z y. *)
(* Proof. *)
(* apply: (ler_trans _ (ler_norm_add (x - z) (z - y))). *)
(* by rewrite addrA addrNK lerr. *)
(* Qed. *)

(* Canonical Structure numIdomainMetMixin :=  *)
(*   EspMetMixin d_num_pos d_num_sym d_num_refl d_num_tri  *)

(* End Metric_of_NumDomain. *)



(* 
*** Local Variables: ***
*** coq-prog-name: "/home/gcano/sandbox/coq-8.4pl1/bin/coqtop" ***
*** coq-prog-args: ("-emacs-U" "-I" "./lib" "-I" "../coqfinitgroup" "-I" "../coqfinitgroup/ssreflect/ssreflect1.3_v8.4/src/") ***
*** End: ***
 *)